:- module(eq_solver,
	[ solve_recs/11
	, solve_rec/10
	, solve_general_form_rec_rel/5
	, equations_to_general_form/5
	, simplify_min_max_expr/2
	, min_/3
	, max_/3
	, add_/3
	, subtract_/3
	, solve_non_determinism_remove/3
	],
	[assertions,
	 nativeprops,
	 regtypes,
	 dcg,
	 datafacts]).

:- doc(title, "Extended recurrence equations solver").
:- doc(author, "Alejandro Serrano").
:- doc(author, "Maximiliano Klemen (Since 1/1/2014)").
:- doc(module, "Allows to solve recurrence equations
	by deleting unneeded arguments, and knowing when
        to minimize or maximize results.").

% :- doc(bug, " When the backend solver mathematica throws an exception
% 	that is catched in this module (for example in eq_solver_simplify),
% 	the  MathKernel process keeps alive. But in the followings calls
%       to mathematica, a new MathKernel process is created.").
%SOLVED

%:- use_module(domain(resources/eq_solver/solver/relevant_vars)).

:- use_module(engine(hiord_rt), [call/1]). % TODO: review uses here

:- use_module(domain(resources/eq_solver/solver/solver_utils_trans)).
:- use_module(domain(resources/plai2infer), [replace_term/3]).
:- use_module(library(lists)).
:- use_module(library(sort)).
:- use_module(library(sets)).
:- use_module(library(terms_check)).
:- use_module(library(terms_vars)).
:- use_module(library(iso_misc)).
:- use_module(ciaopp(preprocess_flags), [current_pp_flag/2] ).
:- use_module(library(system),[time/1]).
:- use_module(library(hiordlib),[maplist/2]).
:- use_module(library(messages)).
:- use_module(domain(resources/eq_solver/solver/solver_messages),[show_user_message/1]).
:- use_module(domain(resources/eq_solver/solver/solver_utils_builtin),[trivial_tautologies/1]).

:- import(new, [
normal_form/2 ,     
diff_equ_type/7 ,   
solve_diff_equ/7 ,  
general_form/2 ,    
normal_form/2 ,     
normal_form/2
	]).

:- import(algebraic_res, []).
:- import(normal_form_res,[]).
:- import(general_form_res, []).
:- import(diff_equ_res, []).

:- use_module(eq_solver_graphs, [varmember/2]).
:- use_module(eq_solver_term_position).
:- use_module(eq_solver_types).

:- use_module(domain(resources/eq_solver/solver/solver),
	[simplify/2,
	 rsolve/5,
	 greater_than/2,
	 greater_or_equal_than/2,
	 equal_than/2
	]
).

:- use_module(domain(resources/recurrence_processing), [closed_form/1, slice_args/3, remove_elements/3, normalize_eqs_/2, unify_vars/1, denormalize_tests/2,non_recursive_eq/1]).

:- use_module(typeslib(typeslib), [dz_equivalent_types/2]).


:- pred solve_recs(Name,RFlag,RecRels,IArgs,OArgs,AlreadySolved,NbResources,IsDet,CanFail,TrySolve,Resolution)
	: atom * atom*list(recurrence_relation) * list(var) * list(var) * list(resolution) * num * term * term *atom* var
       => atom * atom * list(recurrence_relation) * list(var) * list(var) * list(resolution) * num * term * term *atom* resolution.
:- doc(solve_recs/10,
	"Solves the set of recurrence relations
         in @var{RecRels} into @var{Resolution}.
         It supposes it must minimize and
         maximize solutions alternatively.
         The solutions to @var{AlreadySolved} are
         known or set as equations.").
solve_recs(Name,RFlag,RecRels,IArgs,OArgs,AlreadySolved,NbResources,IsDet,CanFail,TrySolve,Resolution) :-
	% Split between argument and resource equations
	NbResEqs is NbResources * 2,
	length(RecRelsRes,NbResEqs),
	append(RecRelsArgs,RecRelsRes,RecRels),
	solve_recs_(RecRelsArgs,RFlag,IArgs,AlreadySolved,lower,IsDet,CanFail,false,TrySolve,SolutionsArgs),
	solve_recs_(RecRelsRes,RFlag,IArgs,AlreadySolved,lower,IsDet,CanFail,true,TrySolve,SolutionsRes),
	append(SolutionsArgs,SolutionsRes,Solutions),
	convert_to_resolution(Name,IArgs,OArgs,Solutions,Resolution).

convert_to_resolution(Name,IArgs,OArgs,Solutions,Resolution) :-
	vars_position(OArgs,OPos),
	terms_at_positions(OArgs,OPos,OVars),
	zip_list(OVars,Solutions,VarSols),
	Resolution = resolution(Name,IArgs,OArgs,VarSols).

:- pred zip_list(R,S,Zipped)
	: list * list * var
       => list * list * list.
:- doc(zip_list/3,
	"Creates a new list @var{Zipped}
         by interleaving elements from
         @var{R} and @var{S}.").
zip_list([],_,[]) :- !.
zip_list(_,[],[]) :- !.
zip_list([A|R],[B|S],[(A,B)|Zipped]) :-
	zip_list(R,S,Zipped).

solve_recs_([],_,_IArgs,_AlreadySolved,_,_,_,_,_,[]).
solve_recs_([RecRel|R],RFlag,IArgs,AlreadySolved,Bound,IsDet,CanFail,AreSolEqs,TrySolve,[Sol|SolR]) :-
	solve_rec(RecRel,RFlag,IArgs,AlreadySolved,Bound,IsDet,CanFail,AreSolEqs,TrySolve,Sol),
%	((Bound = lower,instance(Sol,closed_form(GG)),Sol = closed_form(GG),ground(GG)) -> error_message("solve_rec(~q, ~q, ~q, ~q, ~q,~q,~q,~q)",[RecRel,IArgs,AlreadySolved,Bound,IsDet,AreSolEqs,TrySolve,Sol]);true),
	opposite_bound(Bound,OpBound),
	solve_recs_(R,RFlag,IArgs,AlreadySolved,OpBound,IsDet,CanFail,AreSolEqs,TrySolve,SolR).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%TEMPORARY: for debugging purposes
:- data profile_info/2.
%:- export(profile_info/2).
%:- export(clean_timers/0).
clean_timers :-
	retractall_fact(profile_info(_,_)).

update_timer(T1,N) :-
	time(T2),
	current_fact(profile_info(N,Prev)),!,
	retract_fact(profile_info(N,Prev)),
	New is Prev + T2 - T1,
	asserta_fact(profile_info(N,New)).
update_timer(T1,N) :-
	time(T2),
	New is T2 - T1,
	asserta_fact(profile_info(N,New)).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


:- pred solve_rec(RecurrenceRelation,RFlag,InputArgs,AlreadySolved,Bound,IsDet,CanFail,AreSolutionEqs,TrySolve,Solution)
	: recurrence_relation * atom * list * list(resolution) * bound * term * term  * atom  * atom * var
       => recurrence_relation * atom *list * list(resolution) * bound * term * term  * atom * atom* resolution.
:- doc(solve_rec/8,
	"Solves the @var{RecurrenceRelation} for
         getting the specified @var{Bound} and
         puts it into @var{Solution}.").


%%  TODO: Use det information, or take into account constraints, in order to obtain more precise bounds. By the moment, IsDet argument is not used.
solve_rec(RecRel,RFlag,InputArgs,AlreadySolved,Bound,IsDet,CanFail,AreSolutionEqs,TrySolve,Solution):-
	setup_equations(RecRel, RFlag,InputArgs,Bound,IsDet,CanFail,AreSolutionEqs,AlreadySolved,FinalEqs,AllEqs,_ChosenV,ChosenA,Incomplete_eq),!,
	(Incomplete_eq == true -> (Solution = unsolved_equations(ChosenA,FinalEqs));
	    FinalEqs = [equation(F,_,_,_)|_],
	    (already_solved(F,FinalEqs,AllEqs,Bound,Expr)
	    -> Solution = closed_form(Expr)
	    ;
		(get_input_positions(FinalEqs,FinalPos),
		 get_mappings(FinalEqs,1,FinalPos,FinalVars,_),
		 (((TrySolve == just_equations);current_pp_flag(postpone_solver,on)) ->
		   Solution = unsolved_equations(ChosenA,FinalEqs)
		 ;		 
		     (solve_special_equation(FinalEqs,Bound,SolutionExpr)
		     -> simplify_min_max_expr(SolutionExpr,SolutionExprS),
		     debug_message("EQ SOLVED>>> ~q",[SolutionExprS]),
		     Solution = closed_form(SolutionExprS)
		     ; (
			   debug_message("EQ READY FOR SOLVER>>> ~q",[AllEqs]),
			   solve_rec_with_solver(AllEqs,FinalVars,Bound,SolutionExpr),
			   debug_message("EQ SOLUTION FROM SOLVER>>> ~q",[SolutionExpr])
		       -> Solution = closed_form(SolutionExpr);Solution = unsolved_equations(ChosenA,FinalEqs)
		       )
		     )
		 ))
	    )).


:- pred setup_equations(RR,RFlag,IArgs,Bound,IsDet,CanFail,AreSols,Already,Eqs,AllEqs,ChosenV,ChosenA,Incomplete_eq):
	(term(RR),atom(RFlag),list(IArgs),bound(Bound),boolean(IsDet),boolean(CanFail),yesno(AreSols),
	 list(Already,resolution),var(Eqs),
	 var(AllEqs),var(ChosenV),var(ChosenA),var(Incomplete_eq))
 => (recurrence_relation(Eqs), recurrence_relation(AllEqs), 
 list(ChosenV,list(num)),list(ChosenA),boolean(Incomplete_eq))
# 
	"Setup the recurrence relation with the information received
	from pred_solver. Returns the recurrence relation in
	@var{Eqs}; the recurrence relation including those equations
	called by @var{Eqs} in @var{AllEqs}; a list of argument
	positions corresponding to the positions of the arguments in
	the received pseudo-recurrence relation @var{RR}, in
	@var{ChosenV}; a list of input arguments corresponding to the
	input arguments of the linearized version of @var{RR}, in
	@var{ChosenA}; and a value in @var{Incomplete_eq} indicating
	if the resulting recurrence relation correspong to a special
	class that needs to be returned unsolved.".

setup_equations(RecRel,RFlag ,InputArgs,Bound,IsDet,CanFail,AreSolutionEqs,AlreadySolved,Eqs,AllEqs,ChosenV,ChosenA,Incomplete_eq):-
	get_linearized_equations(RecRel, Lin),
	get_linIArgs(RecRel,InputArgs,LinArgs),
	keep_only_needed_variables(Lin,RFlag,LinArgs,Bound,Eqs0,ChosenV,ChosenA,Incomplete_eq),
	simplify_equations_(Eqs0,Eqs1),!,
	solve_nondet(AreSolutionEqs,Eqs1,Bound,IsDet,NEqs0),
	inline_ground_calls(NEqs0,NEqs00),
	add_base_if_fail(NEqs00,CanFail,NEqs000),
	add_equations(NEqs000,ChosenV,AlreadySolved,NAllEqs),
	NEqs000 = Eqs,NAllEqs=AllEqs.
%	sanitize_equations(NEqs,Eqs),
%	sanitize_equations(NAllEqs,AllEqs).



sanitize_equations([],[]).
sanitize_equations([equation(N,A,Exp,Con0)|T],[equation(N,A,Exp,Con)|TT]):-
	remove_ground_tautologies(Con0,Con),!,
	sanitize_equations(T,TT).

remove_ground_tautologies([],[]).
remove_ground_tautologies([Comp|R],T):-
	Comp =.. [_Op,A,B],
	ground(A),ground(B),
	call(Comp),!,
	remove_ground_tautologies(R,T).
remove_ground_tautologies([Comp|R],[Comp|T]):-
	remove_ground_tautologies(R,T).



:- pred inline_ground_calls(Eqs,REqs):
	(recurrence_relation(Eqs),var(REqs)) =>
	 (recurrence_relation(REqs))
# "Reduce a recurrence relation, by inlining ground calls to a
function that match a base case of that function.  
Example: (f(1) = 1,f(X) = f(1) + C) ==> (f(1) = 1, f(X) = 1 + C)".

inline_ground_calls(Eqs,REqs):-
	obtain_call_solution_pairs(Eqs,Map),
	replace_term(Eqs,Map,RREqs),
	simplify_equations(RREqs,REqs).

obtain_call_solution_pairs([],[]).
obtain_call_solution_pairs([equation(F,Args,Exp,_)|T],[(fun(F,Args),Exp)|R]):-
	\+ \+ maplist(is_num,Args),
	\+ \+ num(Exp),!,
	obtain_call_solution_pairs(T,R).
obtain_call_solution_pairs([_|T],R):-
	obtain_call_solution_pairs(T,R).
	
is_num(X):-
	num(X),!.
is_num(_):-
	fail.


:- pred get_linearized_equations(term(RR),var(Lin)) =>
	(recurrence_relation(Lin))
# "Generates the recurrence relation by putting the variables
	appearing in @var{RR} at first level, instead of in a type
	functor. 

Example: f(num(A,B)) = f(num(A-1,B)) ==> f(A,B) = f(A-1,B)".

get_linearized_equations(RecRel,Lin):-
	get_input_positions(RecRel,IPos),
	linearize_equations(RecRel,IPos,Lin).


get_linIArgs(RecRel,InputArgs,LinIArgs):-
	get_input_positions(RecRel,IPos),
	%linearize_equations(RecRel,IPos,Lin),
	terms_at_positions(InputArgs,IPos,LinIArgs).


:- pred keep_only_needed_variables(Eqs,RFlag,IArgs,Bound,REqs,CV,CA,Incomplete_eq):
	(recurrence_relation(LinEqs),list(IArgs),bound(Bound),var(REqs),var(CV),var(CA),
	var(Incomplete_eq)) =>
	(recurrence_relation(REqs),list(CV,list(num)),list(CA),yesno(Incomplete_eq))
# "Basically returns a recurrence relation in @var{REqs} equivalent to
	@var{LinEqs}, without unecessary arguments. It also returns
	additional information that is used in following steps of
	setup_equation".
keep_only_needed_variables(LinEqs,RFlag,LinIArgs,Bound,Eqs,ChosenV,ChosenA, Incomplete_eq):-
%	terms_at_positions(InputArgs,IPos,LinIArgs),
	get_input_positions(LinEqs,LinIPos),
	analyze_eq_vars(LinIPos,LinEqs,Bound,Info),
	length(Info,NumberOfVars),
	decide_eq_vars(Info,Bound,ChosenVars),
	(initial_incomplete_equation(LinEqs,RFlag,ChosenVars) ->
	 (
	     Incomplete_eq=true,
	    split_in_two_lists(LinIPos,VarsMin,VarsMax),
	    (Bound = lower -> OnlyVars = VarsMin ; OnlyVars = VarsMax),
	     terms_at_positions(LinIArgs,OnlyVars,ChosenOnlyIArgs),
	     keep_only_needed_vars_(LinEqs,OnlyVars,NumberOfVars,ChosenOnlyIArgs,Eqs),
	     ChosenV = OnlyVars, ChosenA = ChosenOnlyIArgs
	);
	    (
		Incomplete_eq = false,
		decide_not_ground_vars(ChosenVars,Info,ChosenVars2),
		terms_at_positions(LinIArgs,ChosenVars2,ChosenIArgs2),
		keep_only_needed_vars_(LinEqs,ChosenVars2,NumberOfVars,ChosenIArgs2,Eqs),
		ChosenV = ChosenVars2, ChosenA = ChosenIArgs2
	    )
	).
%	filter_out_irrelevant_vars(FEqs,Eqs).
%	remove_uneeded_vars(Eqs,FEqs,_ChosenVars),
%	(Incomplete_eq == false -> error_message("ARGS EQS >>> ~q ",[FEqs]);true).



% :- export(remove_uneeded_vars/3).
% remove_uneeded_vars(Eqs,NEqs,Vars):-
% 	remove_uneeded_vars_(Eqs,Res),
% 	select_vars(Eqs,Res,Vars),
% 	remove_vars(Res,Eqs,NEqs),!.

% :- export(select_vars/3).
% select_vars([equation(_F,Vars,_,_)|_],Res,Chosen):-
% 	select_vars_(Vars,Res,Chosen).

% select_vars_([],_,[]).
% select_vars_([X|T],[true|R],[X|Rest]):-
% 	select_vars_(T,R,Rest).
% select_vars_([_X|T],[false|R],Rest):-
% 	select_vars_(T,R,Rest).

% :- export(remove_vars/3).
% remove_vars(_,[],[]).
% remove_vars([],E,E).
% remove_vars(Res,[equation(F,Args,Expr,Cons)|T],[equation(F,NArgs,NExpr,Cons)|TT]):-
% 	select_vars_(Args,Res,NArgs),
% 	remove_vars_in_exp(Res,F,Expr,NExpr),
% 	remove_vars(Res,T,TT).
% :- export(remove_vars_in_exp/4).
% remove_vars_in_exp(_Res,_F,X,X):-
% 	var(X),!.
% remove_vars_in_exp(Res,F,fun(F,Call),fun(F,NCall)):-
% 	select_vars_(Call,Res,NCall).
% remove_vars_in_exp(Res,F,X,NX):-
% 	X =.. [Op|Args],
% 	remove_vars_in_exp_args(Res,F,Args,NArgs),
% 	NX =.. [Op|NArgs].

% remove_vars_in_exp_args(_Res,_F,[],[]).
% remove_vars_in_exp_args(Res,F,[A|T],[B|TT]):-
% 	remove_vars_in_exp(Res,F,A,B),
% 	remove_vars_in_exp_args(Res,F,T,TT).




% :- export(remove_uneeded_vars_/2).
% remove_uneeded_vars_([],[]).
% remove_uneeded_vars_([equation(F,Args,Expr,C)|T],Result):-
% 	remove_uneeded_vars_one_eq(Args,F,Expr,C,(Sol,_IsRec)),!,
% 	remove_uneeded_vars_(T,RR),
% 	or_list(Sol,RR,Result).
% remove_uneeded_vars_([_|T],RR):-
% 	%remove_uneeded_vars_one_eq(Args,F,Expr,C,),
% 	remove_uneeded_vars_(T,RR).

% or_list([],A,A).
% or_list(A,[],A).
% or_list([A|A1],[B|B1],[AB|AB1]):-
% 	or_(A,B,AB),
% 	or_list(A1,B1,AB1).

% or_(false,V,V).
% or_(true,_,true).


% :- export(remove_uneeded_vars_one_eq/5).
% remove_uneeded_vars_one_eq(Args,F,Expr,Cons,(Ut,R)):-
% 	is_recursive(F,Expr,R),
% 	label_args(F,Args,Expr,Cons,Ut),!.

% :- export(is_recursive/3).
% is_recursive(F,Expr,T):-
% 	contains_function(F,Expr),!,T=true.
% is_recursive(_F,_Expr,false).


% contains_function(_F,V):-
% 	var(V),!,
% 	fail.
% contains_function(F,fun(F,_)).
% contains_function(F,Exp):-
% 	Exp =.. [_|ExpArgs],
% 	contains_function_args(F,ExpArgs).

% contains_function_args(F,[E|_T]):-
% 	contains_function(F,E),!.
% contains_function_args(F,[_|T]):-
% 	contains_function_args(F,T).

% :- export(label_args/5).
% label_args(_F,[],_,_,[]).
% label_args(F,[X|T],E,Cons,[true|TT]):-
% 	nonvar(X),!,
% 	label_args(F,T,E,Cons,TT).
% label_args(F,[X|T],Expr,Cons,[true|TT]):-
% 	take_off_recursive_calls(F,Expr,NRExpr,_Recs),
% 	varset(NRExpr,Args0),
% 	varset(Cons,ArgsCons),
% 	append(Args0,ArgsCons,Args),
% 	varmember(X,Args),!,
% 	label_args(F,T,Expr,Cons,TT).
% label_args(F,[X|T],Expr,Cons,[false|TT]):-
% 	take_off_recursive_calls(F,Expr,_NRExpr,Recs),
% 	used_in_rec_calls(X,Recs,true),!,
% 	label_args(F,T,Expr,Cons,TT).	
% label_args(F,[_X|T],Expr,Cons,[false|TT]):-
% 	label_args(F,T,Expr,Cons,TT).

% take_off_recursive_calls(F,Exp,[],[Exp]):-
% 	nonvar(Exp),
% 	Exp = fun(F,_),!.
% take_off_recursive_calls(_F,Exp,[Exp],[]):-
% 	var(Exp),!.
% take_off_recursive_calls(_F,Exp,[Exp],[]):-
% 	num(Exp),!.
% take_off_recursive_calls(F,Exp,NRExp,Recs):-
% 	nonvar(Exp),
% 	Exp =.. [_|Args],
% 	take_off_recursive_calls_args(F,Args,NRExp,Recs).

% take_off_recursive_calls_args(_,[],[],[]).
% take_off_recursive_calls_args(F,[A|Args],NRExp,Recs):-
% 	take_off_recursive_calls(F,A,NRA,RecA),
% 	take_off_recursive_calls_args(F,Args,NRR,RR),
% 	append(NRR,NRA,NRExp),
% 	append(RR,RecA,Recs).




% :- export(used_in_rec_calls/3).
% used_in_rec_calls(X,[fun(_,Args)|_T],true):-
% 	used_in_rec_call(X,Args,true),!.
% used_in_rec_calls(X,[fun(_,_Args)|T],V):-
% 	used_in_rec_calls(X,T,V).
% used_in_rec_calls(_X,[],false).

% used_in_rec_call(_X,[],false).
% used_in_rec_call(X,[D|_T],true):-
% 	nonvar(D),
% 	X \== D,
% 	varset(D,Vars),
% 	varmember(X,Vars),!.
% used_in_rec_call(X,[_D|T],V):-
% 	used_in_rec_call(X,T,V).


:- pred simplify_equations_(Eqs,SEqs):(recurrence_relation(Eqs),var(SEqs)) =>
	(recurrence_relation(SEqs))
# "@var{SEqs} is a recurrence relation equivalent to @var{Eqs}, with
	all the expressions simplified and does not contain duplicated
	equations.".

simplify_equations_(Eqs, EqsSimplified):-
	simplify_min_max(Eqs,MinMaxSimpEqs),
	simplify_equations(MinMaxSimpEqs, SEqs0),
	simplify_tests(SEqs0, SEqs1),
	remove_duplicate_eqs(SEqs1,EqsSimplified).


simplify_tests([],[]).
simplify_tests([equation(F,Args,Exp,T)|R], [equation(F,Args,Exp,[])|RR]):-
	trivial_tautologies(T),!,
	simplify_tests(R,RR).
simplify_tests([equation(F,Args,Exp,T)|R], [equation(F,Args,Exp,T)|RR]):-
	simplify_tests(R,RR).



%First argument: AreSolutionEquations.
% TODO: Now, in order to solve non determinism, the constraints of the
% equations have to be taken into account, instead that just
% the information that the original predicate is deterministic.
% The following call is commented out until the predicate being corrected.

% solve_nondet(Sol,Eqs,Bound,IsDet,FinalEqs):-
% 	split_rec_and_nonrec(Eqs,NonRecSet,RecSet),
% 	solve_nondet_(Sol,NonRecSet,Bound,IsDet,FinalEqsNonRec),
% 	solve_nondet_(Sol,RecSet,Bound,IsDet,FinalEqsRec),
% 	append(FinalEqsNonRec,FinalEqsRec,FinalEqs).

% split_rec_and_nonrec([],[],[]).
% split_rec_and_nonrec([Eq|T],[Eq|TT],TTT):-
% 	non_recursive_eq(Eq),!,
% 	split_rec_and_nonrec(T,TT,TTT).
% split_rec_and_nonrec([Eq|T],TT,[Eq|TTT]):-
% 	split_rec_and_nonrec(T,TT,TTT).



solve_nondet(true,Eqs,Bound,IsDet,FinalEqs):-
	solve_non_determinism_for_solutions(Eqs,Bound,IsDet,FinalEqs).
% solve_nondet(true,Eqs,_Bound,_IsDet,Eqs).
solve_nondet(false,Eqs,Bound,_IsDet,FinalEqs):-
	solve_non_determinism(Eqs,Bound,FinalEqs).


add_equations(Eqs, ChosenIArgs, AlreadySolved,FEqs):-
	[equation(F,_,_,_)|_] = Eqs, % Get function name
	 add_extra_equations(Eqs,F,ChosenIArgs,AlreadySolved,FEqs).

add_base_if_fail(Eqs,IsFail,NEqs):-
	no_basecase(Eqs),!,
	add_base_if_fail_(Eqs,IsFail,NEqs).
add_base_if_fail(F,_,F).

no_basecase([]).
no_basecase([equation(F,_,Expr,_)|Eqs]):-
	is_recursive(F,Expr,true),!,
	no_basecase(Eqs).

is_recursive(F,Expr,T):-
	contains_function(F,Expr),!,T=true.
is_recursive(_F,_Expr,false).

contains_function(_F,V):-
	var(V),!,
	fail.
contains_function(F,fun(F,_)).
contains_function(F,Exp):-
	Exp =.. [_|ExpArgs],
	contains_function_args(F,ExpArgs).

contains_function_args(F,[E|_T]):-
	contains_function(F,E),!.
contains_function_args(F,[_|T]):-
	contains_function_args(F,T).


add_base_if_fail_(F,false,F).
add_base_if_fail_([equation(F,Args,B,T)|R],true,[equation(F,Args0,0,[])|[equation(F,Args,B,T)|R]]):-
	length(Args,D),
	zero_list(D,Args0).

zero_list(0,[]):-!.
zero_list(X,[0|Y]):-
	X1 is X - 1,
	zero_list(X1,Y).

% We have an equation of the form f(x) = f(x)
% or one of the form f(cte, x) = cte
% We can safely just depend in each case of min or max variables
% initial_incomplete_equation(Lin,RFlag,ChosenVars):-
% 	( ChosenVars = [], Lin = [equation(MF,_,MExpr,_)]% , ground(MExpr)
% 	,MExpr = fun(MF,_)  %functor(MExpr,MF,_)
% 	; Lin = [equation(MF,G,Ex,_)], has_ground_element(G),!
% 	; Lin = [equation(MF,G,_,Csts)], Csts \= []
% 	).
initial_incomplete_equation(Lin,_RFlag,ChosenVars):-
	ChosenVars = [], Lin = [equation(MF,_,MExpr,_)],MExpr = fun(MF,_),!.
initial_incomplete_equation(Lin,RFlag,_ChosenVars):-
	Lin = [equation(_MF,G,_Ex,_)], has_ground_element(G),RFlag=nr,
	(\+ closed_form(Lin)),!.
initial_incomplete_equation(Lin,RFlag,_ChosenVars):-
	Lin = [equation(_MF,G,_Ex,_)], has_ground_element(G),RFlag=r,!.
initial_incomplete_equation(Lin,_RFlag,_ChosenVars):-
	Lin = [equation(_MF,_G,_,Csts)], Csts \= [],!.

 
	
% contains_non_linearized_eqs(FinalEqs):-
% 	debug_message("CHECKING EQ: >>> ~q",[FinalEqs]),
% 	\+ recurrence_relation(FinalEqs).

%has_ground_element([]).
has_ground_element([G|_]) :- ground(G),!.
has_ground_element([_|R]) :- has_ground_element(R).


% TODO: This predicate should be implemented inside the solver component.
% This should be solved by taking into account the constraints in the analysis
% of a possible solution.
already_solved(F,Eqs,_Eqs,Bound,OutExpr) :-
	contains_only_ground_results(F,Eqs,GResults),!,
	make_min_max_expression(GResults,Bound,OutExpr).
already_solved(F,Eqs,Eqs,Bound,OutExpr) :-
	find_possible_solved(F,Eqs,Eqs,Possible0),	
	(contains_bad(Possible0) -> Possible = [];Possible = Possible0), 
	Possible \= [],
	sort(Possible,PossibleS),
	make_min_max_expression(PossibleS,Bound,OutExpr).
% already_solved(F,Eqs,Eqs,Bound,OutExpr) :-
% 	find_possible_solved(F,Eqs,Eqs,Possible0),
% 	(member('$bad',Possible0) -> Possible = [];Possible = Possible0), 
% 	Possible \= [],
% 	sort(Possible,PossibleS),
% 	make_min_max_expression(PossibleS,Bound,OutExpr).

:- export(contains_only_ground_results/3).
contains_only_ground_results(_F,[],[]).
contains_only_ground_results(F,[equation(F,_Args,Exp,[])|Eqs],[Exp|RR]):-
	ground(Exp),
	contains_only_ground_results(F,Eqs,RR).

	
find_possible_solved(_,[],_,[]).
find_possible_solved(F,[equation(F,[],Expr,_T)|R],AllEqs,[Expr|PosR]) :-
	find_possible_solved(F,R,AllEqs,PosR).
find_possible_solved(F,[equation(F,Args,Expr,_T)|R],AllEqs,Exprs) :-
	% (arithexpression(Expr)), !, % No argument is instantiannted
	maplist(var,Args), !, % No argument is instantiated
	find_possible_solved(F,R,AllEqs,PosR),
	( has_bad_inner_functor(Expr,AllEqs)
	-> Exprs = ['$bad']
	 ; (is_unknown(Expr) -> Exprs = PosR; Exprs = [Expr|PosR])
	).
find_possible_solved(F,[_|R],AllEqs,PosR) :-
	find_possible_solved(F,R,AllEqs,PosR).

make_min_max_expression([A],_,A).
make_min_max_expression([A,B|R],lower,min(A,MaxR)) :-
	!, make_min_max_expression([B|R],lower,MaxR).
make_min_max_expression([A,B|R],upper,max(A,MaxR)) :-
	!, make_min_max_expression([B|R],upper,MaxR).


contains_bad([S|_]):-
	S == '$bad',!.

contains_bad([_X|Y]):-
	contains_bad(Y).

is_unknown(X):- X == '?',!.

% already_solved(F,[equation(F,[],Expr,_T)|R],AllEqs,OutExpr) :-
% 	!,
% 	( has_bad_inner_functor(Expr,AllEqs)
% 	-> already_solved(F,R,AllEqs,OutExpr)
% 	 ; \+ already_solved(F,R,AllEqs,OutExpr),
% 	   OutExpr = Expr
% 	).
% already_solved(F,[equation(F,Args,Expr,_T)|R],AllEqs,OutExpr) :-
% 	maplist(var,Args), !, % No argument is instantiated
% 	( has_bad_inner_functor(Expr,AllEqs)
% 	-> already_solved(F,R,AllEqs,OutExpr)
% 	 ; \+ already_solved(F,R,AllEqs,OutExpr),
% 	   OutExpr = Expr
% 	).
% already_solved(F,[_|R],AllEqs,Expr) :-
% 	already_solved(F,R,AllEqs,Expr).


% has_bad_inner_functor(Expr,_) :-
% 	Expr == '?'.
has_bad_inner_functor(Expr,AllEqs) :-
	nonvar(Expr),
	Expr = fun(F,Args),!,
	%Expr =.. [F|Args], !,
	( find_bad_inner_functor(AllEqs,F)
	; has_bad_inner_functor_list(Args,AllEqs)
	).
has_bad_inner_functor(Expr,AllEqs) :-
	nonvar(Expr),
	% Expr = fun(F,Args),!,
	Expr =.. [F|Args], !,
	( find_bad_inner_functor(AllEqs,F)
	; has_bad_inner_functor_list(Args,AllEqs)
	).

has_bad_inner_functor_list([A|_],AllEqs) :-
	has_bad_inner_functor(A,AllEqs), !.
has_bad_inner_functor_list([_|R],AllEqs) :-
	has_bad_inner_functor_list(R,AllEqs).

find_bad_inner_functor([equation(F,_,_,_)|_],F) :- !.
find_bad_inner_functor([_|R],F) :-
	find_bad_inner_functor(R,F).

:- pred get_input_positions(Equations,Positions)
	: list(equation) * var
       => list(equation) * list(list(num)).
:- doc(get_input_positions/2,
	"Finds the @var{Positions} of the variables that
         are used as input for the @var{Equations}.").
get_input_positions([],[]).
get_input_positions([equation(_,IArgs,_,_T)|EqR],Pos) :-
	leaves_position(IArgs,PosEq),
	% varfilter(PosEqPlusEmpty,[],PosEq),
	get_input_positions(EqR,PosR),
	ord_union(PosEq,PosR,Pos).

%varfilter([],_,[]).
%varfilter([A|R],B,F) :- A == B, !, varfilter(R,B,F).
%varfilter([A|R],B,[A|F]) :- A \== B, !, varfilter(R,B,F).
%:- export(linearize_equations/3).

% :- success linearize_equations(InputPos,RecRel,LinRecRel)
% 	%: recurrence_relation * list(list(num)) * var
%         => recurrence_relation(LinRecRel).
:- pred linearize_equations(InputPos,RecRel,LinRecRel)
	: recurrence_relation * list(list(num)) * var
       => recurrence_relation * list(list(num)) * recurrence_relation.
:- doc(linearize_equations/3,
	"Converts @var{RecRel} recurrence relation into
         a form where variables refered in @var{InputPos}
         are at first level instead of in a functor.").
linearize_equations([],_,[]).
linearize_equations([equation(F,IArgs,Expr,T)|EqR],IPos,[equation(F,NewArgs,NewExpr,T)|LinR]) :-
	terms_at_positions(IArgs,IPos,NewArgs),
%	collect_equation_names([equation(F,IArgs,Expr,T)|EqR],FList),
	linearize_expression(Expr,IPos,IArgs,F,NewExpr),
	linearize_equations(EqR,IPos,LinR).

linearize_expression(Var,_,_,_,Var) :-
	var(Var), !.
linearize_expression(Ground,_,_,_,Ground) :-
	ground(Ground), !.
linearize_expression(Expr,IPos,IArgs,F,NewExpr) :-
	nonvar(Expr),
	Expr = fun(F,CallArgs), !,
	( special_unifiable(IArgs,CallArgs)
	-> terms_at_positions(CallArgs,IPos,NewArgs)
	 ; NewArgs = CallArgs
	),
	%( terms_at_positions(CallArgs,IPos,NewArgs) ; NewArgs = CallArgs ),
	linearize_expression_each(NewArgs,IPos,IArgs,F,NewLinArgs),
	NewExpr = fun(F,NewLinArgs),!.
linearize_expression(Expr,IPos,IArgs,F,NewExpr) :-
	nonvar(Expr),
	Expr = fun(G,Args), G \= F,
	linearize_expression_each(Args,IPos,IArgs,F,LinArgs),
	NewExpr0 = fun(G,LinArgs),
	((\+ algebraic_expression(NewExpr0)) ->
	  %error_message("BEF LIN>>> ~q",[NewExpr0]),
	  extract_expressions(LinArgs,LLinArgs),
	  NewExpr=fun(G,LLinArgs)
	  %error_message("AFT LIN>>> ~q",[NewExpr])
	;
	    NewExpr=NewExpr0
	  ).
linearize_expression(Expr,IPos,IArgs,F,NewExpr) :-
	nonvar(Expr),
	Expr =.. [G|Args],
	linearize_expression_each(Args,IPos,IArgs,F,LinArgs),
	NewExpr =.. [G|LinArgs].

linearize_expression_each([],_,_,_,[]).
linearize_expression_each([E|R],IPos,IArgs,F,[Lin|LinR]) :-
	linearize_expression(E,IPos,IArgs,F,Lin),
	linearize_expression_each(R,IPos,IArgs,F,LinR).

% collect_equation_names([],[]).
% collect_equation_names([equation(F,_IArgs,_Expr,_T)|EqR],[F|RR]):-
% 	collect_equation_names(EqR,RR).


%:- export(extract_expressions/2).
:- pred extract_expressions(Args,LArgs): (list(Args),var(LArgs)) =>
	(list(LArgs,algebraic_expression)) #
"Given a list of size types, returns a list of all the algebraic_expressions
that are subexpressions of @var{Args}. It is used for finishing the linearization of
equations".

extract_expressions([],[]).
extract_expressions([Exp|T],NExps):-
	extract_expressions_(Exp,LExp),
	extract_expressions(T,TExp),
	append(LExp,TExp,NExps).

extract_expressions_(V,[V]):-
	algebraic_expression(V),!.

extract_expressions_(F,NF):-
	F =.. [_|Args],
	extract_expressions(Args,NF),!.
	
special_unifiable([],[]).
special_unifiable([F|R],[G|S]) :-
	functor(F,FFun,_),
	functor(G,GFun,_),
	dz_equivalent_types(FFun,GFun),
	%unifiable(FArgs,GArgs,_),
	special_unifiable(R,S).

% TODO: Bound not used.
:- pred analyze_eq_vars(Positions,Equations,Bound,Information)
	: list(list(num)) * list(equation) *atom * var
       => list(list(num)) * list(equation) *atom * list.
:- doc(analyze_eq_vars/3,
	"Analyzes the variable @var{Positions} in the
         @var{Equations} returning a list of triples
         (corresponding variable, the variable was used
	 in the expression, the variable was used as
	 recurrence argument).").
analyze_eq_vars([],_,_,[]).
analyze_eq_vars([Pos|PosR],Eqs,Bound,[(Pos,UsedExpr,UsedArgInRec,UsedUnknown,UsedAllGround,UsedInTests)|AnR]) :-
	analyze_eq_vars_pos(Eqs,Pos,Bound,UsedExpr,UsedArgInRec,UsedUnknown,UsedAllGround,UsedInTests),
	analyze_eq_vars(PosR,Eqs,Bound,AnR).

analyze_eq_vars_pos([],_,_,false,false,false,true,false).
analyze_eq_vars_pos([equation(Function,IArgs,Expr,Tests)|EqR],Pos,Bound,UsedExpr,UsedArgInRec,UsedUnknown,UsedAllGround,UsedInTests) :-
	analyze_eq_vars_pos(EqR,Pos,Bound,UExprR,UArgR,UUnkR,UAllGR,UInTests), !,
	% Analyze our equation
	term_at_position(IArgs,Pos,TheArg),
	( ground(TheArg)
	-> ThisUsedExpr = false, 
	ThisUsedArgInRec = true, 
	ThisUsedAllGround = true,
	ThisUsedInTests = false,
	   ( TheArg = '?'
	   -> ThisUsedUnknown = true
	    ; ThisUsedUnknown = false
	   )
	 ; analyze_expr(Expr,Function,TheArg,ThisUsedExpr,ThisUsedArgInRec),
	   ThisUsedUnknown = false, ThisUsedAllGround = false,
	   varset(Tests,Vars), (varmember(TheArg,Vars) -> ThisUsedInTests = true; ThisUsedInTests = false)
	),
	% Combine both
	or(UExprR,ThisUsedExpr,UsedExpr),
	or(UArgR,ThisUsedArgInRec,UsedArgInRec),
	or(UUnkR,ThisUsedUnknown,UsedUnknown),
	or(UInTests,ThisUsedInTests,UsedInTests),
	and(UAllGR,ThisUsedAllGround,UsedAllGround).

analyze_expr(Expr,_,Var,true,false) :-
	var(Expr),
	Expr == Var, !. % Our var is used
analyze_expr(Expr,_,Var,false,false) :-
	var(Expr),
	Expr \== Var, !. % Another var
analyze_expr(Expr,_,_,false,false) :-
	ground(Expr), !. % Ground terms has no vars
analyze_expr(Expr,Function,Var,false,UsedArgInRec) :-
	Expr = fun(Function,CallArgs), !,
	varset(Expr,CallVars),
	% We need it to be in the variables inside
	% but not in first level
	( varmember(Var,CallArgs)
	-> UsedArgInRec = false
	 ; ( varmember(Var,CallVars)
	   -> UsedArgInRec = true
	    ; UsedArgInRec = false
	   )
	).
analyze_expr(Expr,Function,Var,UsedExpr,UsedArgInRec) :-
	Expr = fun(G,CallArgs),
	G \== Function,
	TExp =.. [G|CallArgs],
	length(CallArgs,N),
	analyze_expr_arg_each(Function,Var,TExp,1,N,UsedExpr,UsedArgInRec).
analyze_expr(Expr,Function,Var,UsedExpr,UsedArgInRec) :-
	functor(Expr,_G,N),
	analyze_expr_arg_each(Function,Var,Expr,1,N,UsedExpr,UsedArgInRec).

analyze_expr_arg_each(_,_,_,I,N,false,false) :-
	I > N, !.
analyze_expr_arg_each(Function,Var,Expr,I,N,UsedExpr,UsedArgInRec) :-
	I =< N, !,
	% The rest
	I1 is I + 1,
	analyze_expr_arg_each(Function,Var,Expr,I1,N,UsedExprR,UsedArgInRecR),
	% Analyze this argument
	arg(I,Expr,Arg),
	analyze_expr(Arg,Function,Var,UsedExprArg,UsedArgInRecArg),
	% Combine both
	or(UsedExprR,UsedExprArg,UsedExpr),
	or(UsedArgInRecR,UsedArgInRecArg,UsedArgInRec).

% choose_bound_vars(Tests,Bound,Vars):-
% 	varset()

% choose_bound_vars_([],_,[]).
% choose_bound_vars_(Y,B,Y):-
% 	var(B),!.


% choose_bound_vars_([A,_B|Y],lower,[A|YY]):-
% 	choose_bound_vars_(Y,lower,YY).
% choose_bound_vars_([_A,B|Y],upper,[B|YY]):-
% 	choose_bound_vars_(Y,upper,YY).




or(false,false,false) :- !.
or(true,_,true) :- !.
or(_,true,true) :- !.

and(true,true,true) :- !.
and(false,_,false) :- !.
and(_,false,false) :- !.

%and(true,true,true).
%and(false,_,false) :- !.
%and(_,false,false).

not(true,false).
not(false,true).

eval_bool(P,true):-
	call(P),!.
eval_bool(_,false).

%andnot(X,Y,Z) :-
%	not(Y,YNeg),
%	and(X,YNeg,Z).

:- pred decide_eq_vars(Bound,Info,ChosenVars)
	: list * bound * var
       => list * bound * list(list(num)).
:- doc(decide_eq_vars/3,
	"Decides, taking into account the @var{Info}
         in the analyze_eq_vars phase, which variable
         positions should be included in the
         final equation to be solved.").
decide_eq_vars([],_,[]).
decide_eq_vars([(V1,UEx1,UArg1,UUnk1,_UAllG1,T1),(V2,UEx2,UArg2,UUnk2,_UAllG2,T2)|R],Bound,ChosenVars) :-
	decide_eq_vars(R,Bound,ChosenVarsR), !,
	decide_table(UEx1,UArg1,UUnk1,UEx2,UArg2,UUnk2,Bound,PInclude1,PInclude2),
	eval_bool(Bound=lower,Lb),
	and(Lb,T1,IsLow),
	eval_bool(Bound=upper,Ub),
	and(Ub,T2,IsUpp),
	or(PInclude1,IsLow,Include1),
	or(PInclude2,IsUpp,Include2),
	( Include2 = true
	-> ChosenStep1 = [V2|ChosenVarsR]
	 ; ChosenStep1 = ChosenVarsR
	),
	( Include1 = true
	-> ChosenVars = [V1|ChosenStep1]
	 ; ChosenVars = ChosenStep1
	).

%:- pred decide_table(UEx1,UArg1,UUnk1,UEx2,UArg2,UUnk2,Bound,Include1,Include2).
%%%%%%%%%%%%%%%
%Not Safe
% This two cases might cause an error un nubl.pl, but without them prefixSums_linear doesn't work
% decide_table(false, true, true,  false, true, true,  upper, false, true) :- !. % 32 - 32
% decide_table(false, true, true,  false, true, true,  lower, true, false) :- !.

decide_table(true, true, false,  true, true, false,  upper, false, true) :- !. 
decide_table(true, true, false,  true, true, false,  lower, true, false) :- !.
%%%%%%%%%%%%%%%
% There are 128 possibilities 
% If both appear in expressions, we must include both
decide_table(true, _, _,  true, _, _,  _, true, true) :- !. % 32 - 32

% Only one is available -> include only that
decide_table(false, false, _, true, _, _,  _, false, true) :- !. % 16 - 48
decide_table(false, false, _, false, true, UUnk2, _, false, I2) :- % 8 - 56
	!, not(UUnk2,I2).

decide_table(true, _, _, false, false, _, _, true, false) :- !. % 16 - 72
decide_table(false, true, UUnk1, false, false, _, _, I1, false) :- % 8 - 80
	!, not(UUnk1,I1).

% None of them are available -> include none
decide_table(false, false, _, false, false, _,  _, false, false) :- !. % 8 - 88

% One in expression, the other in argument -> include both
decide_table(true, false, _, false, true, UUnk2, _, true, I2) :- % 8 - 96
	!, not(UUnk2,I2).
decide_table(false, true, UUnk1, true, false, _, _, I1, true) :- % 8 - 104
	!, not(UUnk1,I1).

% One in both, the other in argument -> depends on bound
decide_table(true, true, _, false, true, _, lower, true, false) :- !. % 4 - 108
decide_table(true, true, _, false, true, UUnk2, upper, true, I2) :- % 4 - 112
	!, not(UUnk2,I2).
decide_table(false, true, UUnk1, true, true, _, lower, I1, true) :- % 4 - 116
	!, not(UUnk1,I1).
decide_table(false, true, _, true, true, _, upper, false, true) :- !. % 4 - 120

% Both in argument -> only the one for our bound
decide_table(false, true, UUnk1, false, true, _, lower, I1, false) :- % 4 - 124
	!, not(UUnk1,I1).
decide_table(false, true, _, false, true, UUnk2, upper, false, I2) :- % 4 - 128
	!, not(UUnk2,I2).

% decide_table(_false, true, UUnk1, false, true, _, lower, I1, false) :- % 4 - 124
% 	!, not(UUnk1,I1).
% decide_table(_false, true, _, false, true, UUnk2, upper, false, I2) :- % 4 - 128
% 	!, not(UUnk2,I2).





%decide_not_ground_vars(E,_,E).
decide_not_ground_vars([],_,[]).
decide_not_ground_vars([V|R],Info,[V|NotGR]) :-
	member((V,_,_,_,false,_),Info), !,
	decide_not_ground_vars(R,Info,NotGR).
decide_not_ground_vars([V|R],Info,NotGR) :-
	member((V,_,_,_,true,_),Info), !,
	decide_not_ground_vars(R,Info,NotGR).

:- pred split_in_two_lists(L,OddElements,EvenElements)
	: list * var * var
       => list * list * list.
:- doc(split_in_two_lists/3,
	"Creates two lists with the elements in the odd
         and even positions of @var{L}.").
split_in_two_lists([],[],[]).
split_in_two_lists([E|R],[E|S],T) :-
	split_in_two_lists2(R,S,T).

split_in_two_lists2([],[],[]).
split_in_two_lists2([E|R],S,[E|T]) :-
	split_in_two_lists(R,S,T).

% :- success keep_only_needed_vars(LinEqs,Pos,NumberOfVars,IArgs,FinalEqs)
% 	%: recurrence_relation * list(list(num)) * num * list(var) * var
%        => recurrence_relation(FinalEqs).
:- pred keep_only_needed_vars_(LinEqs,Pos,NumberOfVars,IArgs,FinalEqs)
	: recurrence_relation * list(list(num)) * num * list(var) * var
       => recurrence_relation * list(list(num)) * num * list(var) * recurrence_relation.
:- doc(keep_only_needed_vars_/5,
	"Removes every variable in already linearized
         equations @var{LinEqs} but those in @var{Pos}.
         To check for correct removal, the calls to the
         same predicate must have the same @var{NumberOfVars}.").
keep_only_needed_vars_([],_,_,_,[]).
keep_only_needed_vars_([equation(F,IArgs,Expr,T)|R],Pos,NVars,TArgs,[equation(F,NewArgs,NewExpr,NewT)|KeptR]) :-
	terms_at_positions(IArgs,Pos,NewArgs),
	keep_only_needed_vars_in_expr(Expr,F,Pos,NVars,TArgs,NewExpr),
	keep_only_needed_vars_(R,Pos,NVars,TArgs,KeptR),
	project_tests_into_vars(NewArgs,T,NewT).

:- export(project_tests_into_vars/3).
project_tests_into_vars(_Vs,[],[]).
project_tests_into_vars(Vs,[T|R],[T|RR]):-
	vars_are_contained(T,Vs),!,
	project_tests_into_vars(Vs,R,RR).
project_tests_into_vars(Vs,[_T|R],RR):-
	project_tests_into_vars(Vs,R,RR).

vars_are_contained(T,Vs):-
	varset(T,Vars),
	vars_included(Vars,Vs).

vars_included([],_Set).
vars_included([X|Y],Set):-
	included_eq(X,Set),!,
	vars_included(Y,Set).

included_eq(X,[Y|_T]):-
	X == Y,!.
included_eq(X,[_Y|T]):-
	included_eq(X,T).

	



keep_only_needed_vars_in_expr(Var,_,_,_,_,Var) :-
	var(Var), !.
keep_only_needed_vars_in_expr(Expr,F,Pos,NVars,IArgs,NewExpr) :-
	nonvar(Expr),
	Expr = fun(F,CallArgs), !,
%	Expr =.. [F|CallArgs], !,
	( length(CallArgs,NVars)
	-> terms_at_positions(CallArgs,Pos,NewArgs)
	 ; %error_message("Figuring: ~q ~q", [CallArgs,IArgs]),
	   figure_out_variables(CallArgs,IArgs,NewArgs)
	   %error_message("Afterwards: ~q", [NewArgs])
	   %NewArgs = CallArgs
	),
	keep_only_needed_vars_args(NewArgs,F,Pos,NVars,IArgs,NewKeptArgs),
	NewExpr = fun(F,NewKeptArgs).
keep_only_needed_vars_in_expr(Expr,F,Pos,NVars,IArgs,NewExpr) :-
	nonvar(Expr),
	Expr = fun(G,CallArgs), G \== F,!,
	keep_only_needed_vars_args(CallArgs,F,Pos,NVars,IArgs,NewKeptArgs),
	NewExpr = fun(G,NewKeptArgs).
keep_only_needed_vars_in_expr(Expr,F,Pos,NVars,IArgs,NewExpr) :-
	nonvar(Expr),
	Expr =.. [G|CallArgs], 
	keep_only_needed_vars_args(CallArgs,F,Pos,NVars,IArgs,NewKeptArgs),
	NewExpr =.. [G|NewKeptArgs].

keep_only_needed_vars_args([],_,_,_,_,[]).
keep_only_needed_vars_args([A|R],F,Pos,NVars,IArgs,[KeptA|KeptR]) :-
	keep_only_needed_vars_in_expr(A,F,Pos,NVars,IArgs,KeptA),
	keep_only_needed_vars_args(R,F,Pos,NVars,IArgs,KeptR).

%figure_out_variables(CallArgs,IArgs,NewArgs)
figure_out_variables([],_,[]).
figure_out_variables([V|R],IArgs,[V|FigR]) :-
	ground(V),
	!, figure_out_variables(R,IArgs,FigR).
figure_out_variables([V|R],IArgs,[V|FigR]) :-
	varset(V,[RV]),
	varmember(RV,IArgs),
	%(varmember(RV,IArgs);length(IArgs,1)),
	!, figure_out_variables(R,IArgs,FigR).

% ULQ: The following clause added to figure out variables from
% expressions involving variables. The previous two cluases checks
% only for groundness or exactly one variable. The auxilary
% 'check_inclusion' implementation followed as well. This fixes the
% bug with equations having less arguments in recursive
% part. E.g. f(x, y) = f(x)

figure_out_variables([V|R],IArgs,[V|FigR]) :-
	varset(V,LVars),
	check_intersection(LVars, IArgs),
	% check_inclusion(LVars, IArgs),
	!, figure_out_variables(R,IArgs,FigR).

figure_out_variables([_|R],IArgs,FigR) :-
	figure_out_variables(R,IArgs,FigR).

check_inclusion([], _IArgs).
check_inclusion([RV|Xs], IArgs):-
	varmember(RV,IArgs),
	!,
	check_inclusion(Xs, IArgs).

check_intersection([RV|_Xs], IArgs):-
	varmember(RV,IArgs),
	!.
check_intersection([_RV|Xs], IArgs):-
	check_intersection(Xs, IArgs).


%:- export(solve_non_determinism/3).
:- pred solve_non_determinism(NonDetEquations,Bound,Equations).
:- doc(solve_non_determinism/3,
	"Removes those equations from @var{NonDetEquations} which
         contribute to non-determinism, based on the @var{Bound}.").
solve_non_determinism(NonDetEquations,Bound,Equations) :-
	solve_non_determinism_(NonDetEquations,NonDetEquations,Bound,Equations).
	%solve_non_determinism_for_solutions(Equations,Bound,SolvedEqs).

solve_non_determinism_([],_,_,[]).
solve_non_determinism_([Eq|R],NonDetEqs,Bound,Equations) :-
	solve_non_determinism_(R,NonDetEqs,Bound,EqR),
	( solve_non_determinism_remove(Eq,NonDetEqs,Bound)
	-> Equations = EqR
	 ; Equations = [Eq|EqR]
	).

solve_non_determinism_remove(equation(F1,A1,Expr1,T1),[equation(F2,A2,Expr2,T2)|_],upper) :-
	ground(A1), ground(A2), A1 == A2,
	can_be_compared_nondet(equation(F1,A1,Expr1,T1),equation(F2,A2,Expr2,T2)),
	eq_solver_greater_than(Expr2,Expr1).  %simplify('Greater'(Expr2,Expr1),'True').   %simplify_in_mathematica('Greater'(Expr2,Expr1),'True').
solve_non_determinism_remove(equation(F1,A1,Expr1,T1),[equation(F2,A2,Expr2,T2)|_],upper) :-
	\+ ground(A1), \+ ground(A2),
	can_be_compared_nondet(equation(F1,A1,Expr1,T1),equation(F2,A2,Expr2,T2)),
	eq_solver_greater_than(Expr2,Expr1). %simplify_in_mathematica('Greater'(Expr2,Expr1),'True').
solve_non_determinism_remove(equation(F1,A1,Expr1,T1),[equation(F2,A2,Expr2,T2)|_],lower) :-
	ground(A1), ground(A2), A1 == A2,
	can_be_compared_nondet(equation(F1,A1,Expr1,T1),equation(F2,A2,Expr2,T2)),
	eq_solver_greater_than(Expr1,Expr2). %simplify_in_mathematica('Less'(Expr2,Expr1),'True').
solve_non_determinism_remove(equation(F1,A1,Expr1,T1),[equation(F2,A2,Expr2,T2)|_],lower) :-
	\+ ground(A1), \+ ground(A2),
	can_be_compared_nondet(equation(F1,A1,Expr1,T1),equation(F2,A2,Expr2,T2)),
	eq_solver_greater_than(Expr1,Expr2).%simplify_in_mathematica('Less'(Expr2,Expr1),'True').
solve_non_determinism_remove(Eq,[_|R],Bound) :-
	solve_non_determinism_remove(Eq,R,Bound).


can_be_compared_nondet(Eq1,Eq2):-
	non_recursive_eq(Eq1),
	non_recursive_eq(Eq2),!.
can_be_compared_nondet(Eq1,Eq2):-
	\+ non_recursive_eq(Eq1),
	\+ non_recursive_eq(Eq2),!.



% TODO: NEED TO BE REIMPLEMENTED.
:- pred solve_non_determinism_for_solutions(NonDetEquations,Bound,IsDet,Equations).
:- doc(solve_non_determinism_for_solutions/3,
	"Join those equations from @var{NonDetEquations} which
         contribute to non-determinism, for getting the total
         number of solutions, based on the @var{Bound}.").
solve_non_determinism_for_solutions([],_,_,[]).
solve_non_determinism_for_solutions([E|R],Bound,IsDet,Equations) :-
	solve_non_determinism_for_solutions(R,Bound,IsDet,EqsR),
	join_non_det_eq_for_sol(E,EqsR,Bound,IsDet,Equations).

join_non_det_eq_for_sol(E,[],_,_IsDet,[E]).
join_non_det_eq_for_sol(equation(F,Args1,Expr1,T1),[equation(F,Args2,Expr2,T2)|R],Bound,IsDet,Equations) :-
	Args1 == Args2,
	remove_ground_tautologies(T1,T10),
	remove_ground_tautologies(T2,T20),
	T10 == T20, !, % TODO: Compare set of tests not just syntactically
	% join_non_det_eq_for_sol(equation(F,Args1,Expr1+Expr2),R,Bound,Equations).
	join_exp_with_det_info(Expr1,Expr2,IsDet,Bound,ExprR),
	% TODO: THIS NEEDS A SOLUTION!!.
	join_non_det_eq_for_sol(equation(F,Args1,ExprR,T2),R,Bound,IsDet,Equations).
join_non_det_eq_for_sol(equation(F,Args1,Expr1,T1),[equation(F,Args2,Expr2,T)|R],Bound,IsDet,[equation(F,Args2,Expr2,T)|Equations]) :-
%	Args1 \== Args2, !,
	join_non_det_eq_for_sol(equation(F,Args1,Expr1,T1),R,Bound,IsDet,Equations).



%:- export(join_exp_with_det_info/5).
:- pred join_exp_with_det_info(Exp1,Exp2,IsDet,Bound,ExpR) # "Join both expressions, corresponding to different clauses
to obtain one expressing the joined cost. If @var{IsDet} = true and @var{Bound} = upper, then try to keep the
maximum between the two expressions (Analogously with @var{Bound} = lower, keeping the minimum). If @var{IsDet}=false,
then @var{ExpR}=@var{Exp1}+@var{Exp2}".

join_exp_with_det_info(Exp1,Exp2,true,_,Exp1):-
	eq_solver_equal_than(Exp1,Exp2),!.
join_exp_with_det_info(Exp1,Exp2,true,upper,Exp1):-
	eq_solver_greater_or_equal_than(Exp1,Exp2),!.
join_exp_with_det_info(Exp1,Exp2,true,upper,Exp2):-
	eq_solver_greater_or_equal_than(Exp2,Exp1),!.
join_exp_with_det_info(Exp1,Exp2,true,lower,Exp1):-
	eq_solver_greater_or_equal_than(Exp2,Exp1),!.
join_exp_with_det_info(Exp1,Exp2,true,lower,Exp2):-
	eq_solver_greater_or_equal_than(Exp1,Exp2),!.
join_exp_with_det_info(Exp1,Exp2,_IsDet,upper,Exp1+Exp2):-
	!.
join_exp_with_det_info(_Exp1,_Exp2,_IsDet,lower,0). %:-
% 	extract_constant_part(Exp1,Exp10),
% 	extract_constant_part(Exp2,Exp20),
% 	min_(Exp10,Exp20,Exp3).


% :- export(extract_constant_part/2).
% extract_constant_part(Var,0):-
% 	var(Var),!.
% extract_constant_part(Val,Val):-
% 	num(Val),!.
% extract_constant_part(Fun,0):-
% 	Fun = fun(_,_),!.
% extract_constant_part(Exp,NExp):-
% 	Exp =.. [Op,X1,X2],
% 	extract_constant_part(X1,X1C),
% 	extract_constant_part(X2,X2C),
% 	NExp0 =.. [Op,X1C,X2C],
% 	((Op == '/',X2C == 0) -> NExp = 0; NExp is NExp0). 

:- pred simplify_equations(Equations,SimplifiedEquations).
:- export(simplify_equations/2).
% TODO: Improve simplify with builtin solver in order to avoid using mathematica here.
% The builtin solver does not simplify factorial expression fact(X-1)*X as fact(X).
simplify_equations([],[]).
simplify_equations([equation(I,O,Expr,T)|R],[equation(I,O,Simp,T)|SimpR]) :-
	eq_solver_simplify(Expr,Simp),
	%simplify_in_mathematica(Expr,Simp),
	simplify_equations(R,SimpR).

:- pred add_extra_equations(Equations,Exception,InputArgs,AlreadySolved,AllEqs).
:- doc(add_extra_equations/5,
	"Add those equations from the @var{AlreadySolved} ones
         which are referenced from the initial @var{Equations}
         and returns them in @var{AllEqs}, except for @var{Exception}.").
% First taking themselves needed for coupled equations
add_extra_equations(Eqs,Exception,IArgs,Already,AllEqs) :-
	add_extra_eqs_round(Eqs,Exception,IArgs,Already,RoundEqs),
	sort(Eqs,EqsSorted),
	sort(RoundEqs,RoundEqsSorted),
	( ord_subset(RoundEqsSorted,EqsSorted)
	-> AllEqs = Eqs
	 ; ord_subtract(RoundEqsSorted,EqsSorted,NewEqs),
	   append(Eqs,NewEqs,NextStepEqs),
	   add_extra_equations(NextStepEqs,Exception,IArgs,Already,AllEqs)
	).

add_extra_eqs_round([],_,_,_,[]).
add_extra_eqs_round([equation(F,_,Expr,_T)|R],F,IArgs,Already,RoundEqs) :-
	!,
	get_extra_eqs_expr(Expr,Already,ThisEqs),
	choose_with_correct_args(ThisEqs,F,IArgs,ThisEqs2),
	add_extra_eqs_round(R,F,IArgs,Already,REqs),
	append(ThisEqs2,REqs,RoundEqs).
add_extra_eqs_round([equation(G,_,Expr,_T)|R],F,IArgs,Already,RoundEqs) :-
	G \= F,
	get_extra_eqs_expr(Expr,Already,ThisEqs),
	add_extra_eqs_round(R,F,IArgs,Already,REqs),
	append(ThisEqs,REqs,RoundEqs).

choose_with_correct_args([],_,_,[]).
choose_with_correct_args([equation(F,IArgs,E,T)|R],F,CorrectArgs,Out) :-
	!, choose_with_correct_args(R,F,CorrectArgs,OutR),
	( unifiable(IArgs,CorrectArgs,_)
	-> Out = [equation(F,IArgs,E,T)|OutR]
	 ; Out = OutR
	).
choose_with_correct_args([equation(G,IArgs,E,T)|R],F,CorrectArgs,[equation(G,IArgs,E,T)|CR]) :-
	G \= F, !,
	choose_with_correct_args(R,F,CorrectArgs,CR).

get_extra_eqs_expr(Var,_,[]) :-
	var(Var), !.
get_extra_eqs_expr(Funtor,Already,Eqs) :-
	Funtor = fun(F,Args),!,
	%Funtor =.. [F|Args], !,
	get_extra_eqs_arg_each(Args,Already,ArgEqs),
	( find_function(Already,F,FEqs)
	-> append(FEqs,ArgEqs,Eqs)
	 ; Eqs = ArgEqs
	).
get_extra_eqs_expr(Funtor,Already,Eqs) :-
%	Funtor = fun(F,Args),!,
	Funtor =.. [F|Args], !,
	get_extra_eqs_arg_each(Args,Already,ArgEqs),
	( find_function(Already,F,FEqs)
	-> append(FEqs,ArgEqs,Eqs)
	 ; Eqs = ArgEqs
	).

get_extra_eqs_arg_each([],_,[]).
get_extra_eqs_arg_each([A|R],Already,Eqs) :-
	get_extra_eqs_expr(A,Already,EqA),
	get_extra_eqs_arg_each(R,Already,EqR),
	append(EqA,EqR,Eqs).

find_function([resolution(_,_,_,Vars)|_],F,Eqs) :-
	find_function_in_vars(Vars,F,Eqs), !.
find_function([_|R],F,Eqs) :-
	find_function(R,F,Eqs).

find_function_in_vars([(_,unsolved_equations(_,Eqs))|_],F,Eqs) :-
	Eqs = [equation(F,_,_,_)|_], !.
find_function_in_vars([_|R],F,Eqs) :-
	find_function_in_vars(R,F,Eqs).

remove_duplicate_eqs([],[]).
remove_duplicate_eqs([E|R],Rem) :-
	remove_duplicate_eqs(R,RemR),
	( member(X,RemR), variant(X,E)  %X == E
	-> Rem = RemR
	 ; Rem = [E|RemR]
	).


%:- export(solve_rec_ciaopp_solver/2).
% BEGIN Code to interface CiaoPP recurrence solver
:- pred solve_rec_ciaopp_solver(Equations,Solution).
:- doc(solve_rec_ciaopp_solver/2,
	"Solves the @var{Equations} using the integrated CiaoPP solver.").
solve_rec_ciaopp_solver(Equations,Solution) :-
	Equations = [equation(F,_,_,_Test)|_],
	% We can only solve 1-parameter 1-equation recurrences
	equations_to_general_form(Equations,(VarM,NumberM),[ChosenVar],[EqNF],EqBoundary),
	solve_general_form_rec_rel(EqNF,F,ChosenVar,EqBoundary,SolutionNF),
	general_form_to_vars(SolutionNF,NumberM,VarM,Solution).

:- pred equations_to_general_form(Equations,VarMappings,ChosenVars,EqNF,EqBoundary).
equations_to_general_form(Equations,(Vars,Numbers),ChosenN,EqNF,EqBoundary) :-
	get_input_positions(Equations,IPos),
	get_mappings(Equations,1,IPos,Vars,Numbers),
	discard_tests(Equations,Equations0), % because by now we are not using them.
	analyze_eq_vars(IPos,Equations0,_,Info),
	decide_general_form_vars(Info,ChosenVars),
	terms_at_positions(Numbers,ChosenVars,ChosenN),
	convert_to_equations(Equations0,ChosenVars,Vars,Numbers,EqNF,EqBoundary).

discard_tests([],[]).
discard_tests([equation(A,B,C,_D)|R], [equation(A,B,C,[])|RR]):-
	discard_tests(R,RR).

get_mappings(_,_,[],[],[]).
get_mappings(Equations,I,[Pos|PosR],[Var|VarR],[$(I)|NumberR]) :-
	get_mappings_(Equations,Pos,Var),
	I1 is I+1,
	get_mappings(Equations,I1,PosR,VarR,NumberR).

get_mappings_([],_,0).
get_mappings_([equation(_,IPos,_,_T)|EqR],Pos,Var) :-
	get_mappings_(EqR,Pos,VarR),
	term_at_position(IPos,Pos,EqVar),
	( var(VarR)
	-> Var = VarR
	 ; Var = EqVar
	).

decide_general_form_vars([],[]).
decide_general_form_vars([(V,_,true,_,_)|VR],[V|ChosenR]) :-
	decide_general_form_vars(VR,ChosenR).
decide_general_form_vars([(_,_,false,_,_)|VR],ChosenR) :-
	decide_general_form_vars(VR,ChosenR).

convert_to_equations([],_,_,_,[],[]).
convert_to_equations([equation(F,IArgs,Expr,_T)|EqR],ChosenVars,Vars,Numbers,EqNF,EqB) :-
	convert_to_equations(EqR,ChosenVars,Vars,Numbers,EqNFR,EqBR),
	terms_at_positions(IArgs,ChosenVars,EqInputArgs),
	( ground(EqInputArgs)
	-> % We have a base case
	   substitute_inside_equation(Expr,F,ChosenVars,Vars,Numbers,BoundExpr),
	   ( EqInputArgs = [Base]
	   -> BoundVal = Base
	    ; BoundVal =.. [','|EqInputArgs]
	   ),
	   EqNF = EqNFR, EqB = [val(BoundVal,BoundExpr)|EqBR]
	 ; substitute_inside_equation(Expr,F,ChosenVars,Vars,Numbers,EqExpr),
	   EqNF = [EqExpr|EqNFR], EqB = EqBR
	).

substitute_inside_equation(Var,_,_,Vars,Numbers,Subst) :-
	var(Var), !,
	substitute_vars(Var,Vars,Numbers,Subst).
substitute_inside_equation(Gr,_,_,_,_,Gr) :-
	ground(Gr), !.
substitute_inside_equation(Functor,F,ChosenVars,Vars,Numbers,Subst) :-
	Functor =.. [F|Args], !,
	terms_at_positions(Args,ChosenVars,ChosenArgs),
	substitute_inside_equation_each(ChosenArgs,F,ChosenVars,Vars,Numbers,SubstArgs),
	Subst =.. [F|SubstArgs].
substitute_inside_equation(Functor,F,ChosenVars,Vars,Numbers,Subst) :-
	Functor =.. [G|Args], G \== F, !,
	substitute_inside_equation_each(Args,F,ChosenVars,Vars,Numbers,SubstArgs),
	Subst =.. [G|SubstArgs].

substitute_inside_equation_each([],_,_,_,_,[]).
substitute_inside_equation_each([Arg|RArg],F,ChosenVars,Vars,Numbers,[SubstArg|SubstR]) :-
	substitute_inside_equation(Arg,F,ChosenVars,Vars,Numbers,SubstArg),
	substitute_inside_equation_each(RArg,F,ChosenVars,Vars,Numbers,SubstR).

:- pred general_form_to_vars(Expr,Numbers,Vars,VarsExpr).
general_form_to_vars($(I),Numbers,Vars,Var) :-
	!,
	find_var_by_number($(I),Numbers,Vars,Var).
general_form_to_vars(Functor,Numbers,Vars,Expr) :-
	Functor =.. [F|Args], !,
	general_form_to_vars_each(Args,Numbers,Vars,ExprArgs),
	Expr =.. [F|ExprArgs].

find_var_by_number(A,[A|_],[Var|_],Var) :- !.
find_var_by_number(A,[_|R],[_|VarR],Var) :-
	find_var_by_number(A,R,VarR,Var).

general_form_to_vars_each([],_,_,[]).
general_form_to_vars_each([Arg|ArgR],Numbers,Vars,[V|VR]) :-
	general_form_to_vars(Arg,Numbers,Vars,V),
	general_form_to_vars_each(ArgR,Numbers,Vars,VR).

:- pred solve_general_form_rec_rel(Equation,F,Var,Boundary,Solution).
:- doc(solve_general_form_rec_rel/5,
	"Solves the @var{Equation} for @var{F}(@var{Var}) and
         with @var{Boundary} conditions in @var{Solution}.
         The input and output must be in general form.").
solve_general_form_rec_rel(Equation,_F,_Var,_Boundary,Equation) :-
	num(Equation),!.
solve_general_form_rec_rel(Equation,F,Var,Boundary,Solution) :-
	normal_form(Equation,EqNormalForm),
	normal_form_each_boundary(Boundary,BoundaryNF),
	diff_equ_type(EqNormalForm,Var,F/1,A1n,A2n,Bn,T),
	solve_diff_equ(T,Var,A1n,A2n,Bn,BoundaryNF,SolutionNF),
	general_form(SolutionNF,Solution).

normal_form_each_boundary([],[]).
normal_form_each_boundary([val(Arg,Res)|R],[val(ArgN,ResN)|RN]) :-
	normal_form(Arg,ArgN),
	normal_form(Res,ResN),
	normal_form_each_boundary(R,RN).
% END Code to interface CiaoPP recurrence solver

%:- export(perm/2).
perm([],[]).
perm([X|Xs],[R|Rs]) :-
	select(R,[X|Xs],Y),
	perm(Y,Rs).

% select(X,[X|Xs],Xs).
% select(X,[Y|Ys],[Y|Zs]) :-
% 	select(X,Ys,Zs).

% BEGIN Code to solve equations with max and min and some with two arguments
:- pred solve_special_equation(RecurrenceRelations,Bound,Solution)
	: recurrence_relation * bound * var
       => recurrence_relation * bound * resolution.
:- doc(solve_special_equation/2,
	"Tries to solve a small set of possible
         special @var{RecurrenceRelations}.").

solve_special_equation([A,B],Bound,Solved):-
	solve_by_summation([A,B],Bound,Solved),!.
solve_special_equation([A,B],Bound,Solved):-
	solve_by_summation([B,A],Bound,Solved),!.
solve_special_equation([A,B],Bound,Solved):-
	solve_parallel_decreasing_args([A,B],Bound,Solved),!.
solve_special_equation([A,B],Bound,Solved):-
	solve_parallel_decreasing_args([B,A],Bound,Solved),!.
solve_special_equation([A,B],Bound,Solved):-
	solve_simple_accumulator([A,B],Bound,Solved),!.
solve_special_equation([A,B],Bound,Solved):-
	solve_simple_accumulator([B,A],Bound,Solved),!.
solve_special_equation([A,B],Bound,Solved):-
	solve_basecase_unknown_size([B,A],Bound,Solved).
solve_special_equation([A,B,C],Bound,Solved):-
	%length(Eqs,3),!,
%	error_message("RED: ~q (~q)",[Eqs,_B]),
	solve_extra_var([A,B,C],Bound,Solved).
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation(Eqs,_B,Solved) :-
%	error_message("MINMAX: ~q (~q)",[Eqs,_B]),
	solve_min_max_equation(Eqs,Solved).
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation([A,B],_B,Solved) :-
%	error_message("MINMAX: ~q (~q)",[[A,B],_B]),
	solve_min_max_equation([B,A],Solved).
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation([A,B],_B,Solved) :-
%	error_message("MINMAX: ~q (~q)",[[A,B],_B]),
	solve_argument_fixed_1([A,B],_B,Solved),!.
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation([A,B],_B,Solved) :-
%	error_message("MINMAX: ~q (~q)",[[A,B],_B]),
	solve_argument_fixed_1([B,A],_B,Solved),!.
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation([A,B,C],Bound,Solved) :-
%	error_message("2LISTS: ~q (~q)",[[A,B],Bound]),
	solve_nondet_equation_1([A,B,C],Bound,Solved).
solve_special_equation([A,B],Bound,Solved) :-
	solve_nondet_equation_2([A,B],Bound,Solved).
solve_special_equation([A,B],Bound,Solved) :-
	solve_nondet_equation_2([B,A],Bound,Solved).
solve_special_equation([A,B,C],Bound,Solved) :-
%	error_message("2LISTS: ~q (~q)",[[A,B],Bound]),
	solve_two_lists_equation_2([A,B,C],Bound,Solved).
solve_special_equation([A,B],Bound,Solved) :-
	solve_nondet_equation_2([A,B],Bound,Solved).
solve_special_equation([A,B],Bound,Solved) :-
	solve_nondet_equation_2([B,A],Bound,Solved).
solve_special_equation(Eqs,B,Solved) :-
%	error_message("2LISTS: ~q (~q)",[Eqs,B]),
	solve_two_lists_equation(Eqs,B,Solved).
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation([A,B],Bound,Solved) :-
%	error_message("2LISTS: ~q (~q)",[[A,B],Bound]),
	solve_two_lists_equation([B,A],Bound,Solved).
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation(Eqs,_B,Solved):-
	length(Eqs,3),!,
%	error_message("RED: ~q (~q)",[Eqs,_B]),
	solve_redundant_equation(Eqs,Solved).
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation(Eqs,B,Solved):-
	%length(Eqs,3),!,
%	error_message("RED: ~q (~q)",[Eqs,_B]),
	solve_variable_basecase(Eqs,B,Solved).
%	error_message("SOLV: ~q",[Solved]).
solve_special_equation([A,B],Bound,Solved):-
	%length(Eqs,3),!,
%	error_message("RED: ~q (~q)",[Eqs,_B]),
	solve_variable_basecase([B,A],Bound,Solved).
%	error_message("SOLV: ~q",[Solved]).

% TODO: The simplification should not be necessary here. However,
% some errors occurs when we left the summation as the result.
%eqs([g(0,_C)==(_C,[]),g(_A,_C)==(g(_A-1,_C+_A),[_A>0]).
solve_simple_accumulator([A,B],_Bound,Solved):-
	A = equation(F,[A1,B1],Exp1,[Test1]), %recursive
	B = equation(F,[A2,B2],Exp2,[]), %base
	var(A1),var(B1),var(B2),
	number(A2),
	nonvar(Exp1),unify_with_occurs_check(Exp1,fun(F,[Dec,SecCall])), (SecCall == B1 + A1;SecCall == A1+B1),Dec == A1 - 1,
	Exp2 == B2,
	unify_with_occurs_check(Test1,LT > RT),
	LT == A1, RT == A2,!,
	Solved0 = sum(Var,A2+1,A1,Var) + B2,
	simplify(Solved0,Solved),!.
solve_simple_accumulator([A,B],_Bound,Solved):-
	A = equation(F,[B1,A1],Exp1,[Test1]), %recursive
	B = equation(F,[B2,A2],Exp2,[]), %base
	var(A1),var(B1),var(B2),
	number(A2),
	nonvar(Exp1),unify_with_occurs_check(Exp1,fun(F,[Dec,SecCall])), (SecCall == B1 + A1;SecCall == A1+B1),Dec == A1 - 1,
	Exp2 == B2,
	unify_with_occurs_check(Test1,LT > RT),
	LT == A1, RT == A2,!,
	Solved0 = sum(Var,A2+1,A1,Var) + B2,
	simplify(Solved0,Solved),!.

%eqs([g(0,_C)==(_C,[]),g(_A,_C)==(g(_A-1,_C*_A),[_A>0]).
solve_simple_accumulator([A,B],_Bound,Solved):-
	A = equation(F,[A1,B1],Exp1,[Test1]), %recursive
	B = equation(F,[A2,B2],Exp2,[]), %base
	var(A1),var(B1),var(B2),
	number(A2),
	nonvar(Exp1),unify_with_occurs_check(Exp1,fun(F,[Dec,SecCall])), (SecCall == B1 * A1;SecCall == A1*B1),Dec == A1 - 1,
	Exp2 == B2,
	unify_with_occurs_check(Test1,LT > RT),
	LT == A1, RT == A2,!,
	Solved0 = prod(Var,A2+1,A1,Var) * B2,
	simplify(Solved0,Solved),!.
solve_simple_accumulator([A,B],_Bound,Solved):-
	A = equation(F,[B1,A1],Exp1,[Test1]), %recursive
	B = equation(F,[B2,A2],Exp2,[]), %base
	var(A1),var(B1),var(B2),
	number(A2),
	nonvar(Exp1),unify_with_occurs_check(Exp1,fun(F,[Dec,SecCall])), (SecCall == B1 * A1;SecCall == A1*B1),Dec == A1 - 1,
	Exp2 == B2,
	unify_with_occurs_check(Test1,LT > RT),
	LT == A1, RT == A2,!,
	Solved0 = prod(Var,A2+1,A1,Var) * B2,
	simplify(Solved0,Solved),!.

%eqs([g(C1,_C)==(g1(_C),[]),g(_A,_C)==(g(_A-1,_C + g2(_A)) + w,TR)
% w is any expression that does not depend on _A nor _C
% TR is either [] or [_A > C1]
% Ci > 0 is a constant, i>=1
% ==> g(_A,_C) =  sum(I,C1 + 1,_A,g2(I) + w) + g1(_C)
% NOTE: Temporary, g1(_C) = _C + C2 and g2(_A) = _A or g2(_A) = C3, and w is a constant
solve_simple_accumulator([A,B],_Bound,Solved):-
	((
	     A = equation(F,[A1,B1],Exp1,Test1), %recursive
	     B = equation(F,[C1,B2],Exp2,[]) %base
	 );
	 (
	     A = equation(F,[B1,A1],Exp1,Test1), %recursive
	     B = equation(F,[B2,C1],Exp2,[]) %base
	 )
	),	
	var(A1),var(B1),var(B2),
	number(C1),
	nonvar(Exp1),unify_with_occurs_check(Exp1,fun(F,[Dec,SecCall]) + Exp3),
	nonvar(SecCall),
	(SecCall = SecCallL + SecCallR;SecCall = SecCallR + SecCallL),
	SecCallL == B1, (SecCallR == A1;number(SecCallR)),
	Dec == A1 - 1,
	number(Exp3),
	nonvar(Exp2), unify_with_occurs_check(Exp2, LR + RR), ((LR == B2, number(RR));(number(LR) , RR == B2)),
	(unify_with_occurs_check(Test1,[LT > RT]),
	 LT == A1, RT == C1;
	    Test1 = []),!,
	Solved0 = sum(_Var,C1+1,A1,SecCallR + Exp3) + Exp2,
	simplify(Solved0,Solved),!.

%eqs([g(_N,_D)==(_D,[]),g(_A,_C)==(g(_A-1,_C) OP _C,[]).
solve_by_summation([A,B],_Bound,Solved):-
	((
	    A = equation(F,[A1,B1],Exp1,[]), %recursive
	    B = equation(F,[A2,B2],Exp2,[]) %base
	);
	 (
	    A = equation(F,[B1,A1],Exp1,[]), %recursive
	    B = equation(F,[B2,A2],Exp2,[]) %base
	)),	
	var(A1),var(B1),
	number(A2),
	nonvar(Exp1),
	functor(Exp1,Op,2),
	arg(1,Exp1,Exp1L),
	arg(2,Exp1,Exp1R),
	((nonvar(Exp1L),
	 unify_with_occurs_check(Exp1L,fun(F,[Dec,SecCall])),
	 Exp1R == B1)
	;
	 (nonvar(Exp1R),
	 unify_with_occurs_check(Exp1R,fun(F,[Dec,SecCall])),
	 Exp1L == B1)
	),
	SecCall == B1,Dec == A1 - 1,
	Exp2 == B2,!,
	generate_sum_or_prod(Op,_,A2,A1,B1, Solved0),
	simplify(Solved0,Solved),!.

generate_sum_or_prod('+',Var,Ini,End,Exp,sum(Var,Ini,End,Exp)).
generate_sum_or_prod('*',Var,Ini,End,Exp,prod(Var,Ini,End,Exp)).

%eqs([h(_C,_D,_O)==(g(_O+ -1*_D+1)*h(_C-1,_D,_O)+i(_O+ -1*_D+1),[]), h(1,?,_O)==(0,[])]))).
solve_basecase_unknown_size([A,B],_Bound,Solved):-
	A = equation(F,[G, S1, S2],GE,[]),        %recursive
	B = equation(F,[V, V1, V2],BaseVal,[]),  %base
	V==1,V1=='?',V2==S2,
	ground(BaseVal),
	BaseVal==0,
	var(G),var(S1),var(S2),nonvar(GE),
	(
	    GE = C1*fun(F,[G2, S11, S22])+C2,C3=0
	    ;
	    GE = fun(F,[G2, S11, S22])+C2,C1=1,C3=0
	    ;
	    GE = fun(F,[G2, S11, S22])+C2-C3, C1=1
	),
%	C1==1,
	G2 == G-1,S1==S11,S2==S22,
	Solved = (G-1)*(C2-C3).

% eqs([l(_C,_E)==(l(_C-1,_E)*g(_E)+m(_E),[]),l(1,?)==(0,[])])
solve_basecase_unknown_size([A,B],_Bound,Solved):-
	A = equation(F,[G, S1],GE,[]),        %recursive
	B = equation(F,[V, V1],BaseVal,[]),  %base
	V==1,V1=='?',
	ground(BaseVal),
	BaseVal==0,
	var(G),var(S1),nonvar(GE),
	(
	    GE = C1*fun(F,[G2, S11])+C2,C3=0
	    ;
	    GE = fun(F,[G2, S11])+C2,C1=1,C3=0
	    ;
	    GE = fun(F,[G2, S11])+C2-C3, C1=1
	),
%	C1==1,
	G2 == G-1,S1==S11,
	% Solved = (G-1)*(C2-C3).
	Solved = ((-C1 + C1**G)* (C2 - C3))/((-1 + C1)*C1).

% eqs([j(_D,_H)==(fun(j,[_D-1,_H-1]),[]),j(1,_H)==(V,[]),j(1,1)==(V,[])]
% eqs([j(_D,_H)==(fun(j,[_D-1,_H-1]) + 1,[]),j(1,_H)==(V,[]),j(1,1)==(V,[])]
solve_extra_var([A,B,C],_Bound,Solved):-
	A = equation(F,[G,Extra],GE,[]),
	B = equation(F,[1,Extra1],BaseVal,[]),
	ground(BaseVal),
	C = equation(F,[1,1],BaseVal0,[]),
	BaseVal == BaseVal0,
	var(G),nonvar(GE),
	Extra == Extra1,
	GE = fun(F,[G2,Extra2]),
	G2 == G-1,
	Extra2 == Extra - 1,!,
	Solved = BaseVal.
solve_extra_var([A,B,C],_Bound,Solved):-
	A = equation(F,[G,Extra],GE,[]),
	B = equation(F,[1,Extra1],BaseVal,[]),
	ground(BaseVal),
	C = equation(F,[1,1],BaseVal0,[]),
	BaseVal == BaseVal0,
	var(G),nonvar(GE),
	Extra == Extra1,
	GE = fun(F,[G2,Extra2]) + 1,
	G2 == G-1,
	Extra2 == Extra - 1,!,
	Solved = G - 1 + BaseVal.

% solve_extra_var([A,B],_Bound,Solved):-
% 	A = equation(F,[G,_Extra],GE,[]),
% 	B = equation(F,[GV,Extra1],RecCase,[]),
% 	ground(GE),
% 	ground(G),
% 	RecCase = fun(F,[GM,ExtraR]) + 4*Extra1 + 2,
% 	Extra1 == ExtraR,
% 	GM == GV - 1,
% 	Solved = 2*GV*(2*Extra1 + 1).


%eqs([f(A)==(1,[]),f(A)==(fun(f,[A-1])+1,[])])
solve_variable_basecase(Eqs,Bound,Solved):-
	%perm(Eqs,LEqs),
	Eqs = [ equation(F,[G],GE,[]), equation(F,[V],VE,[])],
	var(G), ground(GE), var(V),G==V,nonvar(VE),VE = fun(F,[XX]) + 1,XX == G - 1,!,
	(Bound == upper -> Solved = max(G,GE); Solved = min(GE,G)).

% eqs([f(C,A)==(fun(f,[C-1,A])-A**(C-1)+5*(A**C*C)-2*A**C+1,[C>0]),f(0,A)==(1,[])])
solve_argument_fixed_1(Eqs,_Bound,SolutionExpr) :-
	Eqs = [equation(F,[A1,A2],Expr, T1), equation(F,[A3,A4],BaseCase,[])],
	ground(A3),A3=0,var(A1),var(A2),var(A4),A2==A4,Expr = fun(F,[A7,A6])-A6**(A5-1)+5*(A6**A5*A5)-2*A6**A5+1,
	var(A5),var(A6),nonvar(A7),A7 == A5 - 1,A1==A5,A2==A6,T1 == [A1 > 0], ground(BaseCase),BaseCase=1,
	SolutionExpr = ((5*(A6-1)*A6*A5-2*A6*(A6+2)+1)*A6**A5+A6*(A6*(A5+3)-2*A5+2)+A5)/(A6-1)**2.
solve_argument_fixed_1(Eqs,_Bound,SolutionExpr) :-
	Eqs = [equation(F,[A1,A2],Expr, T1), equation(F,[A3,A4],BaseCase,[])],
	ground(A3),A3=0,var(A1),var(A2),var(A4),A2==A4,Expr = fun(F,[A7,A6])-A6**(A5-1)+2*(A6**A5*A5) + A6**A5+1,
	var(A5),var(A6),nonvar(A7),A7 == A5 - 1,A1==A5,A2==A6,T1 == [A1 > 0], ground(BaseCase),BaseCase=1,
	SolutionExpr = ((5*(A6-1)*A6*A5-2*A6*(A6+2)+1)*A6**A5+A6*(A6*(A5+3)-2*A5+2)+A5)/(A6-1)**2.

%Solve the following case:
%eqs([f(1)==(1,[]),f(_A)==(_A,[_A>1]),f(_A)==(fun(f,[_A-1]),[_A>1])])
solve_nondet_equation_1(Eqs,Bound,SolutionExpr) :-
	perm(Eqs,LEqs),
	LEqs = [ equation(F,[A1],E1,[]), equation(F,[A2],E2,T2), equation(F,[A3],E3,T3)],
	A1==1,E1==1,var(A2),var(A3), A2 == A3, var(E2),E2==A2, nonvar(E3), E3 = fun(F,[A3 - 1]),T2 == T3,!,
	(Bound = lower -> SolutionExpr = 0; SolutionExpr = A3).

%f(_C)==(_C,[]),f(_C)==(f(_C-1)
solve_nondet_equation_2(Eqs,Bound,SolutionExpr) :-
	Eqs = [ equation(F,[A1],E1,[]), equation(F,[A2],E2,[])],
	var(A1),E1 == A1, var(A2), A2 == A1, nonvar(E2),E2==fun(F,[A1-1]),!,
	(Bound = lower -> SolutionExpr = 0; SolutionExpr = A1).

%[m(_B,_D)==(_D,[]),m(_B,_D)==(m(_B-1,_D),[])]
solve_nondet_equation_2(Eqs,_Bound,SolutionExpr) :-
	Eqs = [ equation(F,[A1,B1],E1,[]), equation(F,[A2,B2],E2,[])],
	var(A1),
	var(B1),
	var(A2),
	E1 == B1,
	A2 == A1,
	B2 == B1,
	nonvar(E2),
	E2==fun(F,[A1-1,B1]),!,
	SolutionExpr = B1.

%[v(_C)==(_C-1,[]),v(_C)==(v(_C-1)+1,[])]
solve_nondet_equation_2(Eqs,_Bound,SolutionExpr) :-
	Eqs = [ equation(F,[A1],E1,[]), equation(F,[A2],E2,[])],
	var(A1),E1 == A1 - 1, var(A2), A2 == A1, nonvar(E2),E2==fun(F,[A1-1]) + 1,!,
	SolutionExpr = A1 - 1.

%Solve the following case:
%g(1)==(1,[]),g(_E)==(_E,[]),g(_E)==(fun(g,[_E-1])+1,[]) ==> _E
solve_redundant_equation(Eqs,SolutionExpr) :-
	perm(Eqs,LEqs),
	LEqs = [ equation(F,[G],GE,_T1), equation(F,[V],VE,_T2), equation(F,[X],XE,_T3)],
	G==1,GE==1,var(V), (VE == V; VE == V - 1 ), var(X), X == V, (XE == fun(F,[X - 1]) + 1;XE == 1 + fun(F,[X - 1])),!,
	SolutionExpr = X.

solve_two_lists_equation(Eqs,Bound,SolutionExpr) :-
	Eqs = [ equation(F,[Ground,Var2],GroundExpr,_T1), equation(F,[Var1,Var2_2],_,_T2) ],
	ground(Ground),ground(GroundExpr),var(Var1),Var2_2 == Var2,
	get_input_positions(Eqs,[ Yes, _ ]),
	keep_only_needed_vars_(Eqs,[Yes],2,[Var1],SimpEqs),
	get_input_positions(SimpEqs,Pos),
	get_mappings(SimpEqs,1,Pos,Vars,_),
	%MODIFIED 10-06-14: Switch to builtin solver.
	solve_rec_with_solver(SimpEqs,Vars,Bound,SolutionExpr).
	%solve_rec_rel_with_mathematica(SimpEqs,Vars,SolutionExpr).

% eqs([g(1,?)==(0,[]),g(_C,_E)==(_E,[]),g(_C,_E)==(fun(g,[_C-1,_E]),[])])
solve_two_lists_equation_2(Eqs,_Bound,SolutionExpr) :-
	Eqs = [ equation(F,[Ground,Ques],GroundExpr,_T1), equation(F,[Var1,Var2_2],VarExp,_T2), equation(F,[Var1,Var2_2],Exp3,_T2)],
	ground(Ground),ground(GroundExpr),Ques == '?',var(Var1),var(Var2_2),Exp3 = fun(F,[Var1-1,Var2_2]),
	GroundExpr = 0,VarExp == Var2_2,!,
	SolutionExpr = Var2_2.
%eqs([f(1,?)==(?,[]),f(_B,_D)==(_D,[]),f(_B,_D)==(min(fun(f,[_B-1,_D]),_D),[])])
solve_two_lists_equation_2(Eqs,_Bound,SolutionExpr) :-
	Eqs = [equation(F,[Ground,Ques],GroundExpr,_T1), equation(F,[Var1,Var2_2],VarExp,_T2), equation(F,[Var1,Var2_2],Exp3,_T2)],
	ground(Ground),ground(GroundExpr),Ques == '?',var(Var1),var(Var2_2),Exp3 = min(fun(F,[Var1-1,Var2_2]),Var2_2),
	GroundExpr = '?',VarExp == Var2_2,!,
	SolutionExpr = Var2_2.

%eqs([f(C,C)==(D,[]),f(_B,_D)==(f(_B-1,_D-1) + G,[]) ==> UP = min(_B,_D)*D, LB= min(_B,_D) {C,D,G numbers}
solve_parallel_decreasing_args(Eqs,Bound,SolutionExpr):-
	Eqs = [ equation(F,[A1,A2],Exp1,[]), equation(F,[B1,B2],Exp2,[]) ],
	number(A1),number(A2),number(Exp1),
	var(B1),var(B2),
	B1 \== B2,
	A1 == A2,
	nonvar(Exp2),
	Exp2 =.. ['+',Op1,Op2],
	nonvar(Op1),nonvar(Op2),
	(Op1 = fun(F,[PB1,PB2]), number(Op2), Un = Op2;(Op2 = fun(F,[PB1,PB2]), number(Op1)),  Un = Op1),
	 PB1 == B1 - 1, PB2 == B2 - 1,!,
	 (Bound == upper ->
	  SolutionExpr = min(B1,B2)*Un + Exp1
	 ;
	     SolutionExpr = min(B1,B2)
	 ).
:- pred solve_min_max_equation(RecurrenceRelations,Solution)
	: recurrence_relation * var
       => recurrence_relation * resolution.
:- doc(solve_min_max_equation/2,
	"Tries to solve a small set of possible
         @var{RecurrenceRelations} involving
         maximum and minimum operators.").

% Assume all functions and results are increasingly monotonic

% | f(c) = ?
% | f(n) = min(f(n-1),g(n))
% => f(n) = g(c+1)
solve_min_max_equation([ equation(F,[GroundParam],Q1,_T1), equation(F,[Var],Min,_T2) ], Result) :-
	ground(GroundParam), var(Var), Q1 == '?',
	( Min = min(L,GofN) ; Min = min(GofN,L) ),
	nonvar(L),
	L = fun(F,[Var-1]),!,
%	L =.. [F,Var-1], !,
	( num(GroundParam)
	-> ResultP is GroundParam + 1
	 ; ResultP = GroundParam + 1
	),
	substitute_vars(GofN,[Var],[ResultP],SubstGofN),
	eq_solver_simplify(SubstGofN,Result).
	%simplify_in_mathematica(SubstGofN,Result).
% | f(c) = ?
% | f(n) = max(f(n-1),g(n))
% => f(n) = g(n)
solve_min_max_equation([ equation(F,[GroundParam],Q1,_T1), equation(F,[Var],Max,_T2) ], GofN):-
	ground(GroundParam), var(Var), Q1 == '?',
	( Max = max(L,GofN) ; Max = max(GofN, L) ),
	nonvar(L),
	L = fun(F,[Var-1]),!.
%	L =.. [F,Var-1], !.

% | f(c,?) = ? or f(c,m) = ?
% | f(n,m) = min(f(n-1,m),g(m)) or f(n,m) = min(f(c,m),g(m))
% => f(n,m) = g(m) (independent of n)
solve_min_max_equation([ equation(F,[GroundParam,Q1],Q2,_T1), equation(F,[VarR,VarNotR],Min,_T2) ], GofVarNotR ) :-
	ground(GroundParam), var(VarR), (Q1 == '?' ; var(Q1)), Q2 == '?',
	( Min = min(L, GofVarNotR) ; Min = min(GofVarNotR, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[WorkVarR,VarNotR])),
%	L =.. [F,WorkVarR,VarNotR],
	( ground(WorkVarR) ; WorkVarR = VarR - 1 ),
	varset(GofVarNotR,[VarNotR]), !.
% | f(c,?) = ? or f(c,m) = ?
% | f(n,m) = max(f(n-1,m),g(m)) or f(n,m) = max(f(c,m),g(m))
% => f(n,m) = g(m) (independent of n)
solve_min_max_equation([ equation(F,[GroundParam,Q1],Q2,_T1), equation(F,[VarR,VarNotR],Max,_T2) ], GofVarNotR ) :-
	ground(GroundParam), var(VarR), (Q1 == '?' ; var(Q1) ), Q2 == '?',
	( Max = max(L, GofVarNotR) ; Max = max(GofVarNotR, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[WorkVarR,VarNotR])),
%	L =.. [F,WorkVarR,VarNotR],
	( ground(WorkVarR) ; WorkVarR = VarR - 1 ),
	varset(GofVarNotR,[VarNotR]), !.

% | f(n,?) = ? or f(n,m) = ?
% | f(n,m) = min(f(n-1,m),g(m))
% => f(n,m) = g(m) (independent of n)
solve_min_max_equation([ equation(F,[GroundParam,Q1],Q2,_T1), equation(F,[VarR,VarNotR],Min,_T2) ], GofVarNotR ) :-
	var(GroundParam), var(VarR), (Q1 == '?' ; var(Q1)), Q2 == '?',
	( Min = min(L, GofVarNotR) ; Min = min(GofVarNotR, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[WorkVarR,VarNotR])),
	( ground(WorkVarR) ; WorkVarR = VarR - 1 ),
	varset(GofVarNotR,[VarNotR]), !.
% | f(n,?) = ? or f(n,m) = ?
% | f(n,m) = max(f(n-1,m),g(m))
% => f(n,m) = g(m) (independent of n)
solve_min_max_equation([ equation(F,[GroundParam,Q1],Q2,_T1), equation(F,[VarR,VarNotR],Max,_T2) ], GofVarNotR ) :-
	var(GroundParam), var(VarR), (Q1 == '?' ; var(Q1) ), Q2 == '?',
	( Max = max(L, GofVarNotR) ; Max = max(GofVarNotR, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[WorkVarR,VarNotR])),
	( ground(WorkVarR) ; WorkVarR = VarR - 1 ),
	varset(GofVarNotR,[VarNotR]), !.



% | f(c,?) = ?
% | f(n,m) = f(n-1,m)
% | f(n,m) = min(f(n-1,m),g(m))
% => f(n,m) = g(m) (independent of n)
solve_min_max_equation([ Eq1, Eq2, Eq3 ], GofVarNotR) :-
	Eq1 = equation(F,[GroundParam,Q1],Q2,_T1),
	Q1 == '?', Q2 == '?',
	( Eq2 = equation(F,[VarR,VarNotR],Min,_T2),
	  nonvar(Min), Min = min(_,_),
	  Eq3 = equation(F,[VarR,VarNotR],L,_T3)
	-> true
	 ; Eq3 = equation(F,[VarR,VarNotR],Min,_T3),
	   nonvar(Min), Min = min(_,_),
	   Eq2 = equation(F,[VarR,VarNotR],L,_T2)
	),
	ground(GroundParam), var(VarR),
	( unify_with_occurs_check(Min,min(L, GofVarNotR))
	; unify_with_occurs_check(Min,min(GofVarNotR, L))
	),
	nonvar(L),
	L = fun(F,[VarR-1,VarNotR]),
%	L =.. [F,VarR-1,VarNotR],
	varset(GofVarNotR,[VarNotR]), !.
% | f(c,?) = ?
% | f(n,m) = f(n-1,m)
% | f(n,m) = max(f(n-1,m),g(m))
% => f(n,m) = g(m) (independent of n)
solve_min_max_equation([ Eq1, Eq2, Eq3 ], GofVarNotR) :-
	Eq1 = equation(F,[GroundParam,Q1],Q2,_T1),
	Q1 == '?', Q2 == '?',
	( Eq2 = equation(F,[VarR,VarNotR],Max,_T2),
	  nonvar(Max), Max = max(_,_),
	  Eq3 = equation(F,[VarR,VarNotR],L,_T3)
	-> true
	 ; Eq3 = equation(F,[VarR,VarNotR],Max,_T3),
	   nonvar(Max), Max = max(_,_),
	   Eq2 = equation(F,[VarR,VarNotR],L,_T2)
	),
	ground(GroundParam), var(VarR),
	( unify_with_occurs_check(Max,max(L, GofVarNotR))
	; unify_with_occurs_check(Max,max(GofVarNotR, L))
	),
	nonvar(L),
	L = fun(F,[VarR-1,VarNotR]),
%	L =.. [F,VarR-1,VarNotR],
	varset(GofVarNotR,[VarNotR]), !.

% | f(c,?,m) = g(m)
% | f(x,n,m) = min(f(x-1,n,m),g(n))
% => g(min(n,m))
:- pred solve_min_max_equation(I,O):(term(I),var(O)) => (atom(O)).
solve_min_max_equation([ equation(F,[GroundParam,Q,Var2],GofVar2,_T1), equation(F,[Var,Var1,Var2],Min,_T2) ], Result ) :-
	ground(GroundParam), var(Var), Q == '?',
	( Min = min(L, GofVar1) ; Min = min(GofVar1, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[Var-1,Var1,Var2])),
%	L =.. [F,Var-1,Var1,Var2],
	varset(GofVar2,[Var2]),
	varset(GofVar1,[Var1]),
	unifiable(GofVar1,GofVar2,_), !,
	min_(Var1,Var2,MinVars),
	substitute_vars(GofVar1,[Var1],[MinVars],SubstG),
	eq_solver_simplify(SubstG,Result).
	%simplify_in_mathematica(SubstG,Result).
% | f(c,?,m) = g(m)
% | f(x,n,m) = max(f(x-1,n,m),g(n))
% => g(max(n,m))
solve_min_max_equation([ equation(F,[GroundParam,Q,Var2],GofVar2,_T1), equation(F,[Var,Var1,Var2],Max,_T2) ], Result ) :-
	ground(GroundParam), var(Var), Q == '?',
	( Max = max(L, GofVar1) ; Max = max(GofVar1, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F, [Var-1,Var1,Var2])),
%	L =.. [F,Var-1,Var1,Var2],
	varset(GofVar2,[Var2]),
	varset(GofVar1,[Var1]),
	unifiable(GofVar1,GofVar2,_), !,
	max_(Var1,Var2,MaxVars),
	substitute_vars(GofVar1,[Var1],[MaxVars],SubstG),
	eq_solver_simplify(SubstG,Result).
	%simplify_in_mathematica(SubstG,Result).

% | f(n,c,?) = ?
% | f(n,x,m) = min(f(n,x-1,m),g(n,m))
% => g(n,m)
solve_min_max_equation([ equation(F,[N,G,Q1],Q2,_T1), equation(F,[N,X,M],Min,_T2) ], Result) :-
	ground(G), var(N), var(M), Q1 == '?', Q2 == '?',
	( Min = min(L,GOfNM) ; Min = min(GOfNM, L) ),
	nonvar(L),
	unify_with_occurs_check(L, fun(F,[N,X-1,M])),
%	L =.. [F,N,X-1,M],
	( varset(GOfNM,[N,M]) ; varset(GOfNM,[N]) ; varset(GOfNM,[M]) ), !,
	  eq_solver_simplify(GOfNM,Result).
	  %simplify_in_mathematica(GOfNM,Result).

%%%%
%([g(1,?,_J)==(?,[]),g(_D,_F,_J)==(max(_F*_J,fun(g,[_D-1,_F,_J])),[])])
% | f(c,?,n) = ?
% | f(x,n,m) = min(f(x-1,n,m),g(n,m))
% => g(n,m)
solve_min_max_equation([ equation(F,[G,Q1,N],Q2,_T1), equation(F,[X,M,N],Min,_T2) ], Result) :-
	ground(G), var(N), var(M), Q1 == '?', Q2 == '?',
	( Min = min(L,GOfNM) ; Min = min(GOfNM, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[X-1,M,N])),
%	L =.. [F,N,X-1,M],
	varset(GOfNM,VSet),included_vars(VSet,[M,N]),!,
	%( varset(GOfNM,[N,M]) ; varset(GOfNM,[N]) ; varset(GOfNM,[M]) ), !,
	  Result = GOfNM.
	  %eq_solver_simplify(GOfNM,Result).
	  %simplify_in_mathematica(GOfNM,Result).

%([g(1,?,_J)==(?,[]),g(_D,_F,_J)==(max(_F*_J,fun(g,[_D-1,_F,_J])),[])])
% | f(c,?,n) = ?
% | f(x,n,m) = max(f(x-1,n,m),g(n,m))
% => g(n,m)
solve_min_max_equation([ equation(F,[G,Q1,N],Q2,_T1), equation(F,[X,M,N],Max,_T2) ], Result) :-
	ground(G), var(N), var(M), Q1 == '?', Q2 == '?',
	( Max = max(L,GOfNM) ; Max = max(GOfNM, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F,[X-1,M,N])),
%	L =.. [F,N,X-1,M],
	varset(GOfNM,VSet),included_vars(VSet,[M,N]),!,
	% ( varset(GOfNM,[N,M]) ; varset(GOfNM,[N]) ; varset(GOfNM,[M]) ), !,
	  Result = GOfNM.
	  %eq_solver_simplify(GOfNM,Result).
	  %simplify_in_mathematica(GOfNM,Result).

% [h(_D,Cons,?)==(?,[]),h(_D,_F,_H)==(max(_H+_D,h(_H+_D,_F-1,_H)),[])] => h(X,Y,Z) = X + sum(I,Cons,_F,_H)
solve_min_max_equation([A,B], Result) :-
	A = equation(F,Arg1,Exp1,[]),
	B = equation(F,Arg2,Exp2,[]),
	Exp1 == '?', 
	nonvar(Exp2),
	is_min_max_expr(Exp2),
	length(Arg1,3),
	perm([1,2,3],[P1,P2,P3]),
	nth(P1,Arg1,A1),nth(P2,Arg1,A2),nth(P3,Arg1,A3),
	nth(P1,Arg2,B1),nth(P2,Arg2,B2),nth(P3,Arg2,B3),
	var(A1),number(A2),A3 == '?',
	perm([1,2],[Pos1,Pos2]),
	arg(Pos1,Exp2,RecCall),
	arg(Pos2,Exp2,Base),
	functor(RecCall,fun,_),
	RecCall = fun(F,RecCallArgs),
	nth(P1,RecCallArgs,CallP1),
	nth(P2,RecCallArgs,CallP2),
	nth(P3,RecCallArgs,CallP3),
	nonvar(CallP1), nonvar(CallP2),
	(CallP1 == B1+B3;CallP1 == B3+B1),
	(CallP2 == B2-1;CallP2 == B2-1),
	CallP3 == B3,
	Base == CallP1,
	!,
	NExp2 = sum(_,A2,B2,B3) + B1,
	simplify(NExp2,Result).

% | f(n,c,?) = ?
% | f(n,x,m) = max(f(n,x-1,m),g(n,m))
% => g(n,m)
solve_min_max_equation([ equation(F,[N,G,Q1],Q2,_T1), equation(F,[N,X,M],Max,_T2) ], Result) :-
	ground(G), var(N), var(M), Q1 == '?', Q2 == '?',
	( Max = max(L,GOfNM) ; Max = max(GOfNM, L) ),
	nonvar(L),
	unify_with_occurs_check(L,fun(F, [N,X-1,M])),
%	L =.. [F,N,X-1,M],
	( varset(GOfNM,[N,M]) ; varset(GOfNM,[N]) ; varset(GOfNM,[M]) ), !,
	eq_solver_simplify(GOfNM,Result).


is_min_max_expr(min(_, _)):-!.
is_min_max_expr(max(_, _)).

%:- export(included_vars/2).
included_vars([],_).
included_vars([X|Y],T):-
	included_var(X,T),
	included_vars(Y,T).

included_var(X,[D|_T]):-
	D == X,!.
included_var(X,[_|T]):-
	included_var(X,T).



simplify_min_max([],[]).
simplify_min_max([equation(F,Vars,Expr,T)|R],[equation(F,Vars,SimpExpr,T)|SimpR]) :-
	simplify_min_max_expr(Expr,SimpExpr),
	simplify_min_max(R,SimpR).

simplify_min_max_expr(Var,Var) :- var(Var), !.
simplify_min_max_expr(min(A,B),Min) :-
	!,
	simplify_min_max_expr(A,ASimp),
	simplify_min_max_expr(B,BSimp),
	min_(ASimp,BSimp,Min).
simplify_min_max_expr(max(A,B),Max) :-
	!,
	simplify_min_max_expr(A,ASimp),
	simplify_min_max_expr(B,BSimp),
	max_(ASimp,BSimp,Max).
simplify_min_max_expr(F,FSimp) :-
	F = fun(R,Args),
%	F =.. [R|Args],
	simplify_min_max_expr_list(Args,SimpArgs),
	FSimp = fun(R,SimpArgs).
%	FSimp =.. [R|SimpArgs].
simplify_min_max_expr(F,FSimp) :-
	F =.. [R|Args],
	simplify_min_max_expr_list(Args,SimpArgs),
	FSimp =.. [R|SimpArgs].


simplify_min_max_expr_list([],[]).
simplify_min_max_expr_list([A|Args],[ASimp|SimpArgs]) :-
	simplify_min_max_expr(A,ASimp),
	simplify_min_max_expr_list(Args,SimpArgs).
% END Code to solve equations with max and min

% BEGIN Code to work with symbolic expressions
:- pred max_(X,Y,Z) : term * term * var => term * term * term.
:- doc(max_/3,
	"@var{Z} is an expression representing the
         maximum between @var{X} and @var{Y}.").
% Special cases
max_(A,B,A) :- A == B, !.
max_(Q,A,A) :- Q == '?', !.
max_(A,Q,A) :- Q == '?', !.
max_(I,_,inf) :- I == 'inf', !.
max_(_,I,inf) :- I == 'inf',!.
max_(Z,A,A) :- Z == 0, !.
max_(A,Z,A) :- Z == 0, !.
% Simplification
max_(A,B,C) :-
	nonvar(A),
	A = max(D,E),
	D == B,!,
	max_(B,E,C).
max_(A,B,C) :-
	nonvar(A),
	A = max(D,E),
	E == B,!,
	max_(B,D,C).
max_(B,A,C) :-
	nonvar(A),
	A = max(D,E),
	D == B,!,
	max_(B,E,C).
max_(B,A,C) :-
	nonvar(A),
	A = max(D,E),
	E == B,!,
	max_(B,D,C).
% Variables
max_(N,M,max(N,M)) :- var(N), var(M), N @=< M, !.
max_(N,M,max(M,N)) :- var(N), var(M), M @=< N, !.
% Syntactic
max_(N,M,O) :-
	nonvar(M),
	( M = P+Q, P == N
	; M = Q+P, P == N
	), !,
	O = N+Q.
max_(M,N,O) :-
	nonvar(M),
	( M = P+Q, P == N
	; M = Q+P, P == N
	), !,
	O = N+Q.
max_(N,M,O) :-
	nonvar(M),
	M = P-_, P == N,
	!,
	O = N.
max_(M,N,O) :-
	nonvar(M),
	M = P-_, P == N,
	!,
	O = N.
% Not ground elements
max_(N,M,N):-
	eq_solver_greater_or_equal_than(N,M),!.
max_(N,M,M):-
	eq_solver_greater_or_equal_than(M,N),!.


max_(N,M,max(N,M)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), N @=< M, !.
max_(N,M,max(M,N)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), M @=< N, !.
max_(N,M,O) :- ground(N), ground(M), num(N), num(M), !,
	( N > M -> O = N ; O = M ).
max_(N,M,max(M,N)) :- N @>= M, !.
max_(N,M,max(N,M)) :- N @< M.

:- pred min_(X,Y,Z) : term * term * var => term * term * term.
:- doc(min_/3,
	"@var{Z} is an expression representing the
         minimum between @var{X} and @var{Y}.").
% Special cases
min_(A,B,A) :- A == B, !.
min_(Q,A,A) :- Q == '?', !.
min_(A,Q,A) :- Q == '?', !.
min_(I,A,A) :- I == 'inf', !.
min_(A,I,A) :- I == 'inf', !.
min_(Z,_,0) :- Z == 0, !.
min_(_,Z,0) :- Z == 0, !.
% Simplification
min_(A,B,C) :-
	nonvar(A),
	A = min(D,E),
	D == B,!,
	min_(B,E,C).
min_(A,B,C) :-
	nonvar(A),
	A = min(D,E),
	E == B,!,
	min_(B,D,C).
min_(B,A,C) :-
	nonvar(A),
	A = min(D,E),
	D == B,!,
	min_(B,E,C).
min_(B,A,C) :-
	nonvar(A),
	A = min(D,E),
	E == B,!,
	min_(B,D,C).
% Variables
min_(N,M,min(N,M)) :- var(N), var(M), M @=< N, !.
min_(N,M,min(M,N)) :- var(N), var(M), N @=< M, !.
% Syntactic
min_(N,M,O) :-
	nonvar(M),
	( M = P+_, P == N
	; M = _+P, P == N
	), !,
	O = N.
min_(M,N,O) :-
	nonvar(M),
	( M = P+_, P == N
	; M = _+P, P == N
	), !,
	O = N.
min_(N,M,O) :-
	nonvar(M),
	M = P-Q, P == N,
	!,
	O = N-Q.
min_(M,N,O) :-
	nonvar(M),
	M = P-Q, P == N,
	!,
	O = N-Q.
% Not ground elements
% Using solver
min_(N,M,N):-
	eq_solver_greater_or_equal_than(M,N),!.
min_(N,M,M):-
	eq_solver_greater_or_equal_than(N,M),!.
min_(N,M,min(N,M)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), M @=< N, !.
min_(N,M,min(M,N)) :- ( (\+ ground(N)) ; (\+ ground(M)) ), N @=< M, !.
min_(N,M,O) :- ground(N), ground(M), num(N), num(M), !,
	( N < M -> O = N ; O = M ).
min_(N,M,N) :- N @=< M, !.
min_(N,M,M) :- N @> M.

:- pred add_(X,Y,Z) : term * term * var => term * term * term.
:- doc(add_/3,
	"@var{Z} is an expression representing the
         sum of @var{X} and @var{Y}.").
add_(Q,_,'?') :- Q == '?', !.
add_(_,Q,'?') :- Q == '?', !.
add_(A,B,C) :-
	ground(A), arithexpression(A),
	ground(B), arithexpression(B),
	!, C is A + B.
add_(A,B,A+B).

:- pred subtract_(X,Y,Z) : term * term * var => term * term * term.
:- doc(subtract_/3,
	"@var{Z} is an expression representing the
         difference between @var{X} and @var{Y}.").
subtract_(Q,_,'?') :- Q == '?', !.
subtract_(_,Q,'?') :- Q == '?', !.
subtract_(A,B,C) :-
	ground(A), arithexpression(A),
	ground(B), arithexpression(B),
	!, C is A - B.
subtract_(A,B,A-B).
% END Code to work with symbolic expressions


% % :- pred equal_than(Exp1,Exp2) # "Checks if @var{Exp1} is equal than @var{Exp2} for all sound assignment to their variables.".
% equal_than(Exp1,Exp2):-
% 	current_pp_flag(req_solver, S),
% 	equal_than_(S,Exp1,Exp2).

solve_rec_with_solver(AllEqs,FinalVars,Bound,SolutionExpr):-
	current_pp_flag(req_solver, S),
	normalize_eqs_(AllEqs,Eqs0),
	unify_vars(Eqs0),
	slice_args(Eqs0, RemVars,Eqs),
	remove_elements(RemVars,FinalVars,Vars),
	Eqs = [equation(_,Vars,_,_)|_],
	denormalize_tests(Eqs, EqsR),
	eq_solver_rsolve(S,EqsR,Vars,Bound,SolutionExpr).


% :- export(equal_than_/3).
% equal_than_(mathematica,X,Y):-
% 	\+ eq_solver_greater_than(X,Y),
% 	\+ eq_solver_greater_than(Y,X).
% 	%simplify_in_mathematica('Equal'(X,Y),'True').


% % % TODO: Give a better implementation for builtin equal_than.
% equal_than_(builtin,X,Y):-
% 	X == Y.


:- export(eq_solver_simplify/2).
eq_solver_simplify(Exp,SimpExp):-
	catch((
		 simplify(Exp,SimpExp)
		 %error_message("SIMP::: ~q >>>> ~q",[Exp,SimpExp])
		 ),
		 Ex,
		 (
		     show_user_message(Ex), 
		     (Ex = not_implemented(_Op,_Solver,_Exp) -> SimpExp=Exp;throw(Ex))
		 )  
	     ).

eq_solver_greater_than(X,Y):-
	catch(
		 greater_than(X,Y),
		 Ex,
		 (
		     show_user_message(Ex),
		     throw(Ex)
		 )
	     ).


eq_solver_equal_than(X,Y):-
	catch(
		 equal_than(X,Y),
		 Ex,
		 (
		     show_user_message(Ex),
		     throw(Ex)
		 )
	     ).


eq_solver_greater_or_equal_than(X,Y):-
	catch(
		 greater_or_equal_than(X,Y),
		 Ex,
		 (
		     show_user_message(Ex),
		     throw(Ex)
		 )
	     ).

eq_solver_rsolve(St,X,Y,Z,W):-
	%rsolve(St,X,Y,Z,W).
	catch(
		 rsolve(St,X,Y,Z,W),
		 Ex,
		 (
		     show_user_message(Ex),
		     throw(Ex)
		 )
	     ).



%PREVIOUS VERSION OF solve_rec
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %PART1 - SET UP THE INITIAL EQUATIONS (of recurrence_relation type)
% 	%error_message("Very beginning ~q",[RecRel]),
% 	% Make equations linear, that is, flatten data structures
	
% 	% RecRel0 = equation(N,I,O,Tests),
% 	% RecRel = equation(N,I,O),


% 	get_input_positions(RecRel,IPos),
% 	linearize_equations(RecRel,IPos,Lin),
% %PART1 - END
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%	

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %PART2 - Choose relevant variables (taking into account how they are used in the equation)
% 	terms_at_positions(InputArgs,IPos,LinIArgs),
% 	% Analyze and decide which input arguments to obtain
% 	get_input_positions(Lin,LinIPos),
% 	analyze_eq_vars(LinIPos,Lin,Bound,Info),
% 	length(Info,NumberOfVars),
% 	decide_eq_vars(Info,Bound,ChosenVars),
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 	%terms_at_positions(LinIArgs,ChosenVars,ChosenIArgs),
% 	( 
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %PART3 - Take off irrelevant variables and also check if is a kind of intermediate equation not solvable
%           ( ChosenVars = [], Lin = [equation(MF,_,MExpr,_)], MExpr = fun(MF,_)  %functor(MExpr,MF,_)
% 	  ; Lin = [equation(MF,G,_,_)], has_ground_element(G)
% 	  )
% 	-> % We have an equation of the form f(x) = f(x)
% 	   % or one of the form f(cte, x) = cte
% 	   % We can safely just depend in each case of min or max variables
% 	   split_in_two_lists(LinIPos,VarsMin,VarsMax),
% 	   ( Bound = lower -> OnlyVars = VarsMin ; OnlyVars = VarsMax ),
% 	   terms_at_positions(LinIArgs,OnlyVars,ChosenOnlyIArgs),
% 	   keep_only_needed_vars(Lin,OnlyVars,NumberOfVars,ChosenOnlyIArgs,LinStripped),
% 	   debug_message("EQ UNSOLVED (f(x)=f(x) or f(cte,x)=cte) >>> ~q",[LinStripped]),
% 	   Solution = unsolved_equations(ChosenOnlyIArgs,LinStripped)
% 	 ; decide_not_ground_vars(ChosenVars,Info,ChosenVars2),
% 	   terms_at_positions(LinIArgs,ChosenVars2,ChosenIArgs2),
% 	   keep_only_needed_vars(Lin,ChosenVars2,NumberOfVars,ChosenIArgs2,OnlyVarsEqs),
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %PART4 - Simplify and remove duplicate equations
% 	   simplify_min_max(OnlyVarsEqs,MinMaxSimpEqs),
% 	   %error_message("Before solving non-det (are solutions: ~q ~q",[AreSolutionEqs,MinMaxSimpEqs]),
% 	   ( AreSolutionEqs == true
% 	   -> 
% 	   %FinalEqsDup = MinMaxSimpEqs
% 	   % TODO: Now, in order to solve non determinism, the constraints of the
% 	   % equations have to be taken into account, instead that just
% 	   % the information that the original predicate is deterministic.
% 	   % The following call is commented out until the predicate being corrected.
% 	   solve_non_determinism_for_solutions(MinMaxSimpEqs,Bound,IsDet,FinalEqsDup)
% 	    ; solve_non_determinism(MinMaxSimpEqs,Bound,FinalEqsDup)
% 	   ),
% 	   % Final check for duplicates
% 	   simplify_equations(FinalEqsDup,FinalEqsDup0),
% 	   remove_duplicate_eqs(FinalEqsDup0,FinalEqs),
% 	   % simplify_equations(FinalEqsNotSimp,FinalEqs),
% 	   %error_message("Going to solve ~q",[FinalEqs]),
% 	   % Add referenced equations
% 	   [equation(F,_,_,_)|_] = FinalEqs, % Get function name
% 	   add_extra_equations(FinalEqs,F,ChosenIArgs2,AlreadySolved,AllEqs0),
% 	   simplify_equations(AllEqs0,AllEqs1),
% 	   remove_duplicate_eqs(AllEqs1,AllEqs),
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% 	%PART5 - Check if it is an equation already solved (i.e. trivially solved)
% 	% See if it is already solved
% 	( already_solved(F,FinalEqs,AllEqs,Bound,Expr)
% 	-> % Simplify in Mathematica
% 	   % Change: already simplified above
% 	   % simplify_in_mathematica(Expr,SolutionExpr),
% 	   debug_message("EQ ALREADY SOLVED>>> (~q) >>> ~q",[AllEqs,Expr]),
% 	      Solution = closed_form(Expr)
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 	    ; % Convert to Mathematica form
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %PART6 - Prepare the equation for the solver and send it to it (if corresponds)
% 	      %error_message("Mathematica: ~q",[AllEqs]),
% 	      get_input_positions(FinalEqs,FinalPos),
% 	      get_mappings(FinalEqs,1,FinalPos,FinalVars,_),
% 	      ( (%current_pp_flag(output_resources,OUTPUT),(OUTPUT=equations;OUTPUT=both)
% 		  TrySolve == just_equations
% 	       ) ->
% 		Solution = unsolved_equations(ChosenIArgs2,FinalEqs)
% 	      ;
% 	      (
% 	      	  %MODIFIED 06-10-14: Switch to builtin solver.
% 	      	  debug_message("EQ READY FOR SOLVER>>> ~q",[AllEqs]),
% 		  %mathequations2equations(AllEqs,AllEqsG),
% 	      	  solve_rec_with_solver(AllEqs,FinalVars,Bound,SolutionExpr),
% 	      	  debug_message("EQ SOLUTION FROM SOLVER>>> ~q",[SolutionExpr])

% 	      	  %solve_rec_rel_with_mathematica(AllEqs,FinalVars,SolutionExpr)
% 	      -> Solution = closed_form(SolutionExpr)
% 	       ; ( solve_special_equation(FinalEqs,Bound,SolutionExpr)
% 	      	 -> simplify_min_max_expr(SolutionExpr,SolutionExprS),
% 	      	 debug_message("EQ SOLVED>>> ~q",[SolutionExprS]),
% 	      	    Solution = closed_form(SolutionExprS)
% 	      	  ; % For not solvable equations
% 	      	     debug_message("EQ NOT SOLVED>>> ~q",[FinalEqs]),
% 	      	     Solution = unsolved_equations(ChosenIArgs2,FinalEqs)
% 	      	 )
% 	      ))
% 	   )
% 	   %error_message("Solution ~q",[Solution]),
% 	   %error_message(".")
% 	).
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
