:- module(eq_solver_graphs,
	[ graph/1
	, node/1
	, full_edge/1
	, half_edge/1
	, nodes/2
        , neighbours/3
	, insert_node/3
	, insert_edge/3
        , dijkstra_all/4
	, minimal_path/4
	, has_cycles/3
	, merge_graphs/3
        , intersect_graphs/3
        , reverse_graph/2
	, maximum_connected_component/3
        , maximum_connected_component_list/3
	, strong_components/2
        , export_to_graphviz/3
	, varmember/2
	],
	[ assertions,regtypes,basicmodes,hiord
	]).

%TODO: Look if this can not be moved to library. 

:- use_module(library(aggregates)).
:- use_module(library(lists)).
:- use_module(library(sets)).
:- use_module(library(sort)).
:- use_module(library(streams)).
:- use_module(library(write)).

:- doc(title, "Graphs library").
:- doc(author, "Alejandro Serrano").
:- doc(module, "This module includes definitions and predicates
	to work with directed labelled graphs.").
	
:- regtype graph(G)
	# "@var{G} is a directed labelled graph.".
:- doc(graph/1,
	"A graph can be a list of nodes and edges
         or a list of nodes and neighbours.

         @includedef{graph/1}

         ").
graph(graph(Nodes,Edges)) :-
	list(Nodes,node),
	list(Edges,full_edge).
graph(graph(NodesAndNeighbours)) :-
	list(NodesAndNeighbours,node_neighbour).

:- regtype node(N) # "@var{N} is a graph node.".
node(_N).

:- regtype node_neighbour(N)
	# "@var{N} is a node along with all his neighbours.".
node_neighbour((Node,Neighbours)) :-
	node(Node),
	list(Neighbours,half_edge).

:- regtype full_edge(E)
	# "@var{E} is an edge between two nodes with a label.".
full_edge((Origin,_Label,Destination)) :-
	node(Origin),
	node(Destination).

:- regtype half_edge(E)
	# "@var{E} defines a label and destination for an edge.".
half_edge((_Label,Destination)) :-
	node(Destination).

:- pred nodes(Graph,Nodes)
	: graph * var
       => graph * list(node).
:- doc(nodes/2,	"@var{Nodes} is the set of nodes in the @var{Graph}.").
nodes(graph(N, _), N).
nodes(graph(N), Nodes) :- findall(Node, member((Node, _), N), Nodes).

:- pred neighbours(Graph,Node,Neighbours)
	: graph * node * var
       => graph * node * list(half_edge).
:- doc(neighbours/3,
	"@var{Neighbours} include a set of label and destinations
         that can be reached from the @var{Node} in the @var{Graph}.").
%neighbours(graph(E, N), A, Nb) :- findall((L, B), edge_label(graph(E, N), A, B, L), Nb).
neighbours(graph(_, []), _, []).
neighbours(graph(E, [(A, L, B)|R]), C, [(L, B)|NbR]) :-
	A == C,
	neighbours(graph(E, R), C, NbR).
neighbours(graph(E, [(A, _, _)|R]), C, NbR) :-
	A \== C,
	neighbours(graph(E, R), C, NbR).
neighbours(graph(N), A, Neighbours) :- member((A, Neighbours), N).

%edge_label(graph(_, Edges), A, B, Label) :- find_edge_label(A, B, Edges, Label).
%find_edge_label(A, B, [(C, Label, D)|_], Label) :-
%	A == C, B == D.
%find_edge_label(A, B, [_|R], Label) :- find_edge_label(A, B, R, Label).

:- pred insert_node(Graph,Node,NewGraph)
	: graph * node * var
       => graph * node * graph.
:- doc(insert_node/3,
	"@var{NewGraph} is the previouse @var{Graph} with
         @var{Node} inserted if it didn't exist before").
insert_node(graph(Nodes,Edges),Node,graph(NewNodes,Edges)) :-
	insert(Nodes,Node,NewNodes).
insert_node(graph(NodesAndNeighbours),Node,graph(NewNodesAndNeighbours)) :-
	nodes(graph(NodesAndNeighbours), AlreadyNodes),
	( member(Node, AlreadyNodes)
	-> NewNodesAndNeighbours = NodesAndNeighbours
	 ; insert(NodesAndNeighbours,(Node,[]),NewNodesAndNeighbours)
	).

:- pred insert_edge(Graph,Edge,NewGraph)
	: graph * full_edge * var
       => graph * full_edge * graph.
:- doc(insert_edge/3,
	"@var{NewGraph} is the previouse @var{Graph} with
         @var{Edge} inserted if it didn't exist before.").
insert_edge(graph(N,E),(Origin,Label,Dest),graph(NewNodes,NewEdges)) :-
	insert_node(graph(N,E),Origin,G1),
	insert_node(G1,Dest,graph(NewNodes,Edges2)),
	insert(Edges2,(Origin,Label,Dest),NewEdges).
insert_edge(graph(NNeigh),(Origin,Label,Dest),graph(NewNNeigh)) :-
	insert_node(graph(NNeigh),Origin,graph(NNeigh1)),
	insert_node(graph(NNeigh1),Dest,graph(NNeigh2)),
	insert_edge_aux(NNeigh2,(Origin,Label,Dest),NewNNeigh).

insert_edge_aux([],_,[]).
insert_edge_aux([(Origin,Neighbours)|R],(Origin,Label,Dest),[(Origin,NewNeigh)|R]) :-
	!,
	insert(Neighbours,(Label,Dest),NewNeigh).
insert_edge_aux([E|R],Edge,[E|NewR]) :-
	insert_edge_aux(R,Edge,NewR).

:- pred dijkstra_all(Graph,Node,Unvisited,Distances)
	: graph * node * var * var
       => graph * node * list(node) * list.
:- doc(dijkstra_all/4,
	"Gets the minimum distances from @var{Node} from the
         rest of nodes in the @var{Graph}, and the information
         is written out in @var{Distances}, which is comprised
         of triples (Node, Distance, Minimal Path).
         @var{Unvisited} is the set of non-reachable nodes.").
dijkstra_all(G, A, Unvisited, Distances) :-
        nodes(G, Nodes),
        dijkstra_all_aux(G, Nodes, Unvisited, [(A, 0, [])], Distances).

dijkstra_all_aux(G, InUnvisited, OutUnvisited, InDistances, OutDistances) :-
        find_min_unvisited(InDistances, InUnvisited, A, D, MinPath), !,
        visit(G, A, InUnvisited, VUnvisited, InDistances, VDistances, D, MinPath),
        dijkstra_all_aux(G, VUnvisited, OutUnvisited, VDistances, OutDistances).
dijkstra_all_aux(_, InUnvisited, InUnvisited, InDistances, InDistances).

find_min_unvisited([(A, D, P)|R], Unvisited, B, Min, MinPath) :-
        member(A, Unvisited), !,
        find_min_unvisited_aux(R, Unvisited, A, D, P, B, Min, MinPath).
find_min_unvisited([(A, _, _)|R], Unvisited, B, Min, MinPath) :-
        \+ member(A, Unvisited),
        find_min_unvisited(R, Unvisited, B, Min, MinPath).

find_min_unvisited_aux([], _, A, D, P, A, D, P).
find_min_unvisited_aux([(B, E, Q)|R], Unvisited, _, D, _, C, Min, MinPath) :-
        member(B, Unvisited),
        E =< D, !,
        find_min_unvisited_aux(R, Unvisited, B, E, Q, C, Min, MinPath).
find_min_unvisited_aux([(B, E, _)|R], Unvisited, A, D, P, C, Min, MinPath) :-
        member(B, Unvisited),
        E > D, !,
        find_min_unvisited_aux(R, Unvisited, A, D, P, C, Min, MinPath).
find_min_unvisited_aux([(B, _, _)|R], Unvisited, A, D, P, C, Min, MinPath) :-
        \+ member(B, Unvisited), !,
        find_min_unvisited_aux(R, Unvisited, A, D, P, C, Min, MinPath).
        
visit(G, A, InUnvisited, OutUnvisited, InDistances, OutDistances, D, MinPath) :-
        % Get all unvisited neighbours without cycles
        neighbours(G, A, Neighbours_),
        findall((Label, Node), (member((Label, Node), Neighbours_), Node \= A), Neighbours),
        % Get distance to neighbours
        D1 is D + 1,
        % Update all unvisited neighbours
        update_distances_foreach(Neighbours, D1, MinPath, InUnvisited, InDistances, OutDistances),
        delete(InUnvisited, A, OutUnvisited).

update_distances_foreach([], _, _,  _, Distances, Distances).
update_distances_foreach([(L, N)|R], D1, MinPath, Unvisited, InDistances, OutDistances) :-
        member(N, Unvisited),
        update_distances((L, N), D1, MinPath, InDistances, UDistances),
        update_distances_foreach(R, D1, MinPath, Unvisited, UDistances, OutDistances).
update_distances_foreach([(_, N)|R], D1, MinPath, Unvisited, InDistances, OutDistances) :-
        \+ member(N, Unvisited),
        update_distances_foreach(R, D1, MinPath, Unvisited, InDistances, OutDistances).

update_distances((L, N), D1, MinPath, [], [(N, D1, [N, L|MinPath])]).
update_distances((_, N), D1, _,       [(N, P, B)|R], [(N, P, B)|R])  :- P =< D1, !.
update_distances((L, N), D1, MinPath, [(N, P, _)|R], [(N, D1, [N, L|MinPath])|R]) :- P > D1, !.
update_distances((L, N), D1, MinPath, [(M, P, B)|R], [(M, P, B)|S])  :- M \= N, update_distances((L, N), D1, MinPath, R, S).

:- pred minimal_path(Graph,A,B,Path)
	: graph * node * node * var
       => graph * node * node * list(node).
:- doc(minimal_path/4,
	"Obtains the minimal @var{Path} from @var{A} to @var{B}
         in the @var{Graph}.").
minimal_path(G, A, B, [A|MinPath]) :-
        dijkstra_all(G, A, _, Distances),
        find_distance_path(B, Distances, Path),
        reverse(Path, MinPath).

find_distance_path(B, [(B, _, MinPath)|_], MinPath).
find_distance_path(B, [_|R], MinPath) :- find_distance_path(B, R, MinPath).


:- calls has_cycles(Graph,A,B) : graph * node * node.
:- doc(has_cycles/3,
	"Check whether there is a cycle in @var{Graph}
         between nodes @var{A} and @var{B}.").
% A cycle happens when there is a path A -> B and one B -> A
has_cycles(G, A, B) :-
        nodes(G, N),
        member(A, N),
        minimal_path(G, A, B, _),
        A \= B,
        minimal_path(G, B, A, _).

:- pred merge_graphs(G1,G2,Merged)
	: graph * graph * var
       => graph * graph * graph.
:- doc(merge_graph/3,
	"Merges @var{G1} and @var{G2} into @var{Merged}.
         Nodes and edges from both are preserved,
	 combining nodes with same names and
         edges with same endpoints and label.").
merge_graphs(graph(N1, E1), graph(N2, E2), graph(N, E)) :-
        ord_union(N1, N2, N),
        ord_union(E1, E2, E).
merge_graphs(graph(G1), graph(G2), graph(G)) :-
        nodes(G1, N1),
        nodes(G2, N2),
        ord_union(N1, N2, N),
        merge_neighbours(G1, G2, N, [], G).

merge_neighbours(_, _, [], Acc, RAcc) :-
        reverse(Acc, RAcc).
merge_neighbours(G1, G2, [N|R], Acc, RAcc) :-
        neighbours_or_empty(G1, N, Nb1),
        neighbours_or_empty(G2, N, Nb2),
        ord_union(Nb1, Nb2, Nb),
        merge_neighbours(G1, G2, R, [(N, Nb)|Acc], RAcc).

neighbours_or_empty(G, N, Nb) :-
        neighbours(G, N, Nb), !.
neighbours_or_empty(_, _, []).

:- pred intersect_graphs(G1,G2,Intersected)
	: graph * graph * var
       => graph * graph * graph.
:- doc(intersect_graph/3,
	"Intersects @var{G1} and @var{G2} into @var{Intersected}.
         Only nodes and edges (including labels)
         included in both are preserved.").
intersect_graphs(graph(N1, E1), graph(N2, E2), graph(N, E)) :-
        ord_intersection(N1, N2, N),
        ord_intersection(E1, E2, E).

:- pred reverse_graph(Graph,Reversed)
	: graph * var => graph * graph.
:- doc(reverse_graph/2,
	"Gets a copy of @var{Graph} with edges reversed.").
reverse_graph(graph(N, E), graph(N, RE)) :-
        reverse_edges(E, RE).

reverse_edges([], []).
reverse_edges([(A,L,B)|Rest], [(B,L,A)|RevRest]) :-
        reverse_edges(Rest, RevRest).

:- pred maximum_connected_component(Graph,Node,Component)
	: graph * node * var
       => graph * node * graph.
:- doc(maximum_connected_component/3,
	"Gets the maximum connected @var{Component}
         in the @var{Graph} containing @var{Node}.").
maximum_connected_component(graph(N, E), A, graph(MN, ME)) :-
        dijkstra_all(graph(N, E), A, Unvisited, _),
        ord_subtract(N, Unvisited, MN),
        findall((O, L, T), (member((O, L, T), E), member(O, MN), member(T, MN)), MEUnsorted),
        sort(MEUnsorted, ME).

:- pred maximum_connected_component_list(Graph,Nodes,Component)
	: graph * list(node) * var
       => graph * list(node) * graph.
:- doc(maximum_connected_component_list/3,
	"Gets the maximum connected @var{Component}
         in the @var{Graph} containing all @var{Nodes}.").
maximum_connected_component_list(_, [], graph([], [])).
maximum_connected_component_list(G, [A|R], MCC) :-
        maximum_connected_component(G, A, MCCA),
        maximum_connected_component_list(G, R, MCCR),
        merge_graphs(MCCA, MCCR, MCC).

:- pred strong_components(Graph,Components)
	: graph * var
       => graph * list(list(node)).
:- doc(strong_components/2,
	"Performs strong component analysis of @var{Graph}
         using the path-based algorithm. The result is a
         list @var{Components} of list of nodes in each
         strongly connected component.").
strong_components(graph(N, E), L) :-
        pbsca(graph(N, E), N, _, 1, _, [], _, [], _, [], _, [], L).

% pbsca(Graph, InUnassigned, OutUnassigned, InC, OutC, InPreN, OutPreN,  InS, OutS, InP, OutP, InL, OutL)
pbsca(_, [], [], _, _, _, _, _, _, _,  _, L, L).
pbsca(G, [A|RestA], OutA, InC, OutC, InPreN, OutPreN, InS, OutS, InP, OutP, InL, OutL) :-
        pbsca(G, A, [A|RestA], StepA, InC, StepC, InPreN, StepPreN, InS, StepS, InP, StepP, InL, StepL),
        pbsca(G, StepA, OutA, StepC, OutC, StepPreN, OutPreN, StepS, OutS, StepP, OutP, StepL, OutL).

:- push_prolog_flag(multi_arity_warnings, off).

% pbsca(Graph, Vertex, InUnassigned, OutUnassigned, InC, OutC, InPreN, OutPreN, InS, OutS, InP, OutP, InL, OutL)
pbsca(G, V, InA, OutA, InC, OutC, InPreN, OutPreN, InS, OutS, InP, OutP, InL, OutL) :-
        delete(InA, V, StepA),
        % Set the preorder number to C and increment C
        StepPreN = [V=InC|InPreN],
        StepC is InC + 1,
        % Push v to S and P
        StepS = [V|InS],
        StepP = [V|InP],
        % For each neighbour
        neighbours_or_empty(G, V, Nb),
        foreach_nb_pbsca(G, Nb, StepA, OutA, StepC, OutC, StepPreN, OutPreN, StepS, Step2S, StepP, Step2P, InL, Step2L),
        ( Step2P = [V|RestStep2P]
        -> % V is the top element of P
           pop_until_v(Step2S, V, Component, OutS),
           OutP = RestStep2P,
           OutL = [Component|Step2L]
         ; OutS = Step2S,
           OutP = Step2P,
           OutL = Step2L
        ).

:- pop_prolog_flag(multi_arity_warnings).

foreach_nb_pbsca(_, [], InA, InA, InC, InC, InPreN, InPreN, InS, InS, InP, InP, InL, InL).
foreach_nb_pbsca(G, [(_, V)|R], InA, OutA, InC, OutC, InPreN, OutPreN, InS, OutS, InP, OutP, InL, OutL) :-
        do_nb_pbsca(G, V, InA, StepA, InC, StepC, InPreN, StepPreN, InS, StepS, InP, StepP, InL, StepL),
        foreach_nb_pbsca(G, R, StepA, OutA, StepC, OutC, StepPreN, OutPreN, StepS, OutS, StepP, OutP, StepL, OutL).

do_nb_pbsca(_, V, InA, InA, InC, InC, InPreN, InPreN, InS, InS, InP, OutP, InL, InL) :-
        % If assigned preorder number
        find_preorder_number(InPreN, V, N), !,
        ( findall(Comp, ( member(Comp, InL), member(V, Comp) ), Comps), Comps \= []
        -> true % Already assigned to a connected component
         ; pop_until_preorder_less(InPreN, N, InP, OutP)
        ).
do_nb_pbsca(G, V, InA, OutA, InC, OutC, InPreN, OutPreN, InS, OutS, InP, OutP, InL, OutL) :-
        pbsca(G, V, InA, OutA, InC, OutC, InPreN, OutPreN, InS, OutS, InP, OutP, InL, OutL).

find_preorder_number([V=N|_], V, N) :- !.
find_preorder_number([_|R], V, N) :- find_preorder_number(R, V, N).

pop_until_preorder_less(_, _, [], []).
pop_until_preorder_less(PreN, N, [W|R], OutP) :-
        find_preorder_number(PreN, W, M),
        ( M > N
        -> pop_until_preorder_less(PreN, N, R, OutP)
         ; OutP = [W|R]
        ).

pop_until_v([W|Rest], V, [V], Rest) :- V == W, !.
pop_until_v([W|R], V, [W|S], Rest) :- V \== W, pop_until_v(R, V, S, Rest).

:- calls export_to_graphviz(Graph,GraphName,FileName)
	: graph * atom * atom.
:- doc(export_to_graphviz/3,
	"Export the @var{Graph} into file @var{FileName}
         using @var{GraphName} as title.").
export_to_graphviz(graph(_,Edges), Name, OutputFile) :-
        % Open file
        open(OutputFile, 'write', Stream),
        % Write init of document
        print(Stream, 'digraph '),
        print(Stream, Name),
        print(Stream, ' {\n'),
        % Write edges
        export_edges(Stream, Edges),
        % Write end of document
        print(Stream, ' }'),
        % Close file
        close(Stream).

export_edges(_, []).
export_edges(Stream, [(A, L, B)|R]) :-
        print(Stream, '"'),
        print(Stream, A),
        print(Stream, '" -> "'),
        print(Stream, B),
        print(Stream, '" [ label = "'),
        print(Stream, L),
        print(Stream, '" ];\n'),
        export_edges(Stream, R).


% TODO: This should be in a different module.
:- pred varmember(X,Lst)
	: var * list(var)
       => var * list(var).
:- doc(varmember/2,
	"Checks whether @var{X} is in @var{Lst}.").
varmember(X,[Y|_]) :- X == Y.
varmember(X,[_|R]) :- varmember(X,R).