:- module(eq_solver_mathematica,
	[ call_mathematica/4
	, solve_rec_rel_with_mathematica/3
	, simplify_in_mathematica/2
	, expand_in_mathematica/2
	, expand_functions_in_mathematica/2
	, solve_in_mathematica/3
	, element/4
	, stop_mathematica/0
%        , replace_term/3
	],
	[assertions,
	 nativeprops,
	 regtypes,
	 dcg,
	 datafacts]).

:- use_module(library(process)).
:- use_module(library(stream_utils)).
:- use_module(library(format)).
:- use_module(library(lists)).
:- use_module(library(sort)).
:- use_module(library(system)).
:- use_module(library(terms_vars)).

:- use_module(library(javall/javart)).

:- use_module(eq_solver_graphs, [varmember/2]).
:- use_module(eq_solver_term_position).
:- use_module(eq_solver_mathematica_config).
%:- use_module(mathematica_link,[evaluate/2]).%,stop_kernel/0,kernel_is_running/0]).
%:- use_module(library(streams)).
%:-use_module(library(write)).

:- use_module(library(messages)).
% :- multifile issue_debug_messages/1.
% :- data issue_debug_messages/1.
% issue_debug_messages(mathematica).

:- doc(call_mathematica/4,
	"Calls Mathematica to perform the actions in @var{String}.
         Then the result is parsed back, using the correspondence
         between variable numbers and actual variables in @var{Var}.").

%% In order to communicate with mathematica using 
%% Java interface, uncomment the following line.

%:- compilation_fact(eq_solver_mathematica__use_java_interface).

:- if(defined(eq_solver_mathematica__use_java_interface)).

:- data mathematica_started/2.
%mathematica_started(no,nil).

get_mathematica_handle(H) :-
	( current_fact(mathematica_started(yes,H))
	; mathematica_jlink(ClassPath),
	  mathematica_kernel_link(KLink),
	  catch(java_start(ClassPath),java_exception(_),true),
	  java_create_object('com.wolfram.jlink.MathLinkFactory',MLFactory),
	  java_invoke_method(MLFactory,createKernelLink(KLink,H)),
	  java_invoke_method(H,discardAnswer(_)),
	  asserta_fact(mathematica_started(yes,H))
	).

stop_mathematica:-
	(
	    retract_fact(mathematica_started(yes,_)) ->
	    java_stop
	; 
	    true
	).

%Call to mathematica using Java Interface.
call_mathematica(String,Vars,Rule,Output) :-
%	note_message("JAVA: ~s",[String]),nl,
	get_mathematica_handle(H),
	java_invoke_method(H,evaluateToOutputForm(String, 0, U)), !,
	debug_message("MATH RAW RESULT >>> ~q",[U]),
	element(Vars,O,U,[]),
	% If we receive $Failed, we should not succeed
	( ground(O), O = '$Failed'
	-> fail
	 ; !, ( Rule = yes
	   -> O =.. ['List',Solutions|_],
	      Solutions =.. ['List',R|_],
	      R = 'Rule'(_,Output)
	    ; Output = O
	   )
%	   note_message("JAVA Output: ~q",[Output]),nl
	).

:- else.

%:- use_module(library(mathematica)).
:- import(selected_mathematica_interface_auto,[
	evaluate/2,
	close_link/0
	]).


%Call to mathematica using mathematica_link module. 
call_mathematica(String,Vars,Rule,Output) :-
%	note_message("CMD: ~s",[String]),nl,
 	% Wrap with OutputForm command:
 	append("OutputForm[",String,TmpStr),
 	append(TmpStr,"]",CmdString),
	%%%% RH: the catch is a kludge to silently fail when
	%%%%     mathematica recieves an ill-forme expression.
        \+ member(0'?,  CmdString),
        catch(
		 evaluate(CmdString,OutputChars),
		 error(mathematica_error(_), _),
		 fail
	     ),
 	  element(Vars,O,OutputChars,[]),
	  ( ground(O), O = '$Failed'
 	  -> fail
 	  ; !, ( Rule = yes
 	       -> O =.. ['List',Solutions|_],
 	       Solutions =.. ['List',R|_],
 	       R = 'Rule'(_,Output)
 	       ; Output = O
 	       )
 	  ).

stop_mathematica:-
	close_link.

:- endif.

%call_mathematica(String,Vars,Output) :-
%	% Save script to file
%	mktemp_in_tmp('mathXXXXXX',FileName),
%	string_to_file(String,FileName),
%	% Get command line
%	mathematica_path(Path),
%	% Run it
%	process_call(Path,['-script',FileName],[stdout(string(OutputChars))]),
%	element(Vars,Output,OutputChars,[]),
%	% Delete the temp file
%	delete_file(FileName).

:- pred solve_rec_rel_with_mathematica(Eqs,Vars,Solution).
:- doc(solve_rec_rel_with_mathematica/3,
	"Solves the recurrence relation posed by @var{Eqs}
         on variables @var{Vars} into its @var{Solution}.").
solve_rec_rel_with_mathematica(Eqs,Vars,Solution) :-
	convert_to_mathematica(Eqs,MEqs,MFns,MVars),
	generate_mathematica_text(MEqs,MFns,MVars,MText),
	%note_message("Mathematica: ~s",[MText]),nl,
	call_mathematica(MText,Vars,yes,Solution).
	%note_message("Output: ~q",[Solution]).

:- pred convert_to_mathematica(Eqs,MathematicaEqs,MathematicaFns,Vars).
:- doc(convert_to_mathematica/4,
	"Converts a set of @var{Equations} into the list
         of @var{MathematicaEqs}, @var{MathematicaFns} and @var{Vars}
         that are needed to call Mathematica later.").
convert_to_mathematica(Eqs,MathematicaEqs,MathematicaFns,Vars) :-
	%note_message("convert_to_mathematica: ~q",[Eqs]),nl,
	% Get the maximum number of variables
	get_max_var_number(Eqs,N),
	% As all equations to Mathematica must have the same number
	% of parameters, we extend those with less
	generate_n_vars(N,Vars),
	get_fn_names(Eqs,Fns),
	sort(Fns,FnsSorted),
	convert_to_mathematica_(Eqs,Vars,Fns,MathematicaEqs),
	generate_mathematica_fn(FnsSorted,Vars,MathematicaFns),
	debug_message("CONVERTED TO MATH >>> ~q",[MathematicaEqs]).

get_max_var_number([],0).
get_max_var_number([equation(_,Vars,_,_)|R],N) :-
	get_max_var_number(R,NR),
	length(Vars,NV),
	max(NR,NV,N).

max(N,M,N) :- N >= M, !.
max(N,M,M) :- N < M.

get_fn_names([],[]).
get_fn_names([equation(F,_,_,_)|R],[F|FR]) :-
	get_fn_names(R,FR).

convert_to_mathematica_([],_,_,[]).
convert_to_mathematica_([equation(F,Vars,Expr,_)|R],N,Fns,[Eq|EqR]) :-
	complete_until(Vars,N,AllVars),
	Lhs =.. [F|AllVars],
	CompleteEq = (Lhs == Expr),
	prune_non_vars(AllVars,N,SubstL,SubstR),
	substitute_vars(CompleteEq,SubstL,SubstR,EqNotComplete),
	complete_fn_names(EqNotComplete,Fns,N,Eq),
	convert_to_mathematica_(R,N,Fns,EqR).

complete_until([],Vars,Vars).
complete_until([E|R],[V|RVars],[V|CR]) :-
	E == '?', complete_until(R,RVars,CR).
complete_until([E|R],[_|RVars],[E|CR]) :-
	E \== '?', complete_until(R,RVars,CR).

complete_until_zeros([],[],[]).
complete_until_zeros([],[_|R],[0|ZR]) :-
	complete_until_zeros([],R,ZR).
complete_until_zeros([E|R],[_|RVars],[E|CR]) :-
	complete_until_zeros(R,RVars,CR).



replace_bad_chars_in_name([],[]).
replace_bad_chars_in_name([0'_|R],[0'z|RepR]) :-
	!, replace_bad_chars_in_name(R,RepR).
replace_bad_chars_in_name([0':|R],[0'z,0'z|RepR]) :-
	!, replace_bad_chars_in_name(R,RepR).
replace_bad_chars_in_name([E|R],[E|RepR]) :-
	replace_bad_chars_in_name(R,RepR).




generate_mathematica_fn([],_,[]).
generate_mathematica_fn([F0|R],Vars,[Expr|ER]) :-
%	atom_codes(F0,FC),
%	replace_bad_chars_in_name(FC,FNC),
%	atom_codes(F,FNC),
	Expr =.. [F0|Vars],
	generate_mathematica_fn(R,Vars,ER).

prune_non_vars([],[],[],[]).
prune_non_vars([V|R],[_|S],T,U) :-
	nonvar(V), !,
	prune_non_vars(R,S,T,U).
prune_non_vars([V|R],[E|S],[V|T],[E|U]) :-
	var(V), !,
	prune_non_vars(R,S,T,U).

complete_fn_names((Head==Body),Fns,Vars,(CHead==CBody)) :-
	Head =.. [H|HArgs],
	complete_until(HArgs,Vars,CHArgs),
	CHead =.. [H|CHArgs],
	complete_fn_names_expr(Body,Fns,Vars,CBody).

complete_fn_names_expr(Expr,_,_,_) :-
	var(Expr), !, fail.
complete_fn_names_expr(Expr,Fns,Vars,Eq) :-
	Expr =.. [F|Args],
	complete_fn_names_each(Args,Fns,Vars,CArgs),
	( member(F,Fns)
	-> complete_until_zeros(CArgs,Vars,CCArgs)
	 ; CCArgs = CArgs
	),
	Eq =.. [F|CCArgs].

complete_fn_names_each([],_,_,[]).
complete_fn_names_each([E|R],Fns,Vars,[CE|CR]) :-
	complete_fn_names_expr(E,Fns,Vars,CE),
	complete_fn_names_each(R,Fns,Vars,CR).

:- pred generate_mathematica_text(MEqs,MFns,MVars,MText).
:- doc(generate_mathematica_text/4,
	"Generates the Mathematica code to solve the equations
         @var{MEqs} on functions @var{MFns} and variables
         @var{MVars}. These elements can be generated from the
         equations with @tt{convert_to_mathematica}.").
generate_mathematica_text(MEqs,MFns,MVars,MText) :-
	generate_eqs_text(MEqs,EqsText),
	generate_rest_text(MFns,FnsText),
	generate_rest_text(MVars,VarsText),
	FormatText = "FullForm[FullSimplify[RSolve[~s,~s,~s], Element[~s,Integers] ]]",
	format_to_string(FormatText,[EqsText,FnsText,VarsText,VarsText],MText).

generate_rest_text(Eqs,Text) :-
	generate_expr_text_each(Eqs,EqsText,false),
	join_by(EqsText,",",QText),
	append("{",QText,QParen),
	append(QParen,"}",Text).

generate_eqs_text(Eqs,Text) :-
	generate_eq_text_each(Eqs,EqsText),
	join_by(EqsText,",",QText),
	format_to_string("{~s}",[QText],Text).

generate_eq_text_each([],[]).
generate_eq_text_each([(Lhs==Rhs)|Rs],[EqText|RsText]) :-
	generate_expr_text(Lhs,LText,false),
	generate_expr_text(Rhs,RText,true),
	format_to_string("~s==~s",[LText,RText],EqText),
	generate_eq_text_each(Rs,RsText).

:- pred simplify_in_mathematica(Expression,Solution)
	: term * var
       => term * term.
:- doc(simplify_in_mathematica/2,
	"Tries to simplify a @var{Expression} by calling Mathematica.").
simplify_in_mathematica(Expr,Solution) :-
	% Simplify in Mathematica
	term_variables(Expr,FinalVars),
	length(FinalVars,FinalVarsLength),
	generate_n_vars(FinalVarsLength,VarsAsNumbers),
	substitute_vars(Expr,FinalVars,VarsAsNumbers,MExpr),
	generate_simplify_mathematica_text(MExpr,VarsAsNumbers,MText),
	( call_mathematica(MText,FinalVars,no,SolutionExpr)
	-> Solution = SolutionExpr
	 ; Solution = Expr
	).

generate_simplify_mathematica_text(Expr,Vars,MText) :-
	generate_expr_text(Expr,ExprText,true),
	generate_integer_assumptions(Vars,AssText),
	FormatText = "FullForm[FullSimplify[~s,{~s}]]",
	format_to_string(FormatText,[ExprText,AssText],MText).

generate_integer_assumptions([],"") :- !.
generate_integer_assumptions([V],Ass) :-
	!, format_to_string("Element[~q,Integers]",[V],Ass).
generate_integer_assumptions([V|R],Ass) :-
	generate_integer_assumptions(R,RAss),
	format_to_string("Element[~q,Integers],~s",[V,RAss],Ass).

:- pred expand_in_mathematica(Expression,Solution)
	: term * var
       => term * term.
:- doc(expand_in_mathematica/2,
	"Expands @var{Expression} into sums and powers.").
expand_in_mathematica(Expr,Solution) :-
	% Simplify in Mathematica
	term_variables(Expr,FinalVars),
	length(FinalVars,FinalVarsLength),
	generate_n_vars(FinalVarsLength,VarsAsNumbers),
	substitute_vars(Expr,FinalVars,VarsAsNumbers,MExpr),
	generate_expand_mathematica_text(MExpr,MText),
	( call_mathematica(MText,FinalVars,no,SolutionExpr)
	-> Solution = SolutionExpr
	 ; Solution = Expr
	).

generate_expand_mathematica_text(Expr,MText) :-
	generate_expr_text(Expr,ExprText,true),
	FormatText = "FullForm[Expand[~s]]",
	format_to_string(FormatText,[ExprText],MText).




% preorder_exp(-1/5*(('Power'(5,1/2)-5)*('Power'(1/2*('Power'(5,1/2)-1),int(_A))*'Cos'('Pi'*int(_A))))+('Power'(5,-1/2)+1)*'Power'(1/2*('Power'(5,1/2)+1),int(_A))-1,L).

% %:- export(replace_term/3).
% :- pred replace_term(Exp,ListReplaces,Result) 
% 	# "For each pari (@var{T},@var{R}) in @var{ListReplaces}, it replaces each subterm (or term head) in @var{Exp} that unifies with @var{T} with @var{R}.".
%  :- test replace_term(3 + exp(3,cos(pi*(3+D))),[(cos(pi*A),exp(-1,A))],L) => (L = 3 + exp(3,exp(-1,3+D))) # "cos(pi*A) ==> exp(-1,A)".
% :- test replace_term('Power'(2,A),[('Power',exp)],L) => (L = exp(2,A)).
% % TODO: Probably this predicate is already implemented somewhere else. So, it should
% % be removed is that is the case.
% replace_term(Exp,[],Exp).
% replace_term(Exp,[(Term,Rep)|Rest],Result):-
% 	replace_term_(Exp,(Term,Rep),Result0),!,
% 	replace_term(Result0,Rest,Result).

% replace_term_(Exp,(Term,Rep),Rep):-
% 	var(Exp),
% 	Exp == Term.
% replace_term_(Exp,(_Term,_Rep),Exp):-
% 	var(Exp).
% replace_term_(Exp,(Exp,Rep),Rep).
% replace_term_(Exp,(Term,Rep),Res):-
% 	Exp =.. [ExpHead|Args],
% 	replace_term_args_(Args,(Term,Rep),NArgs),
% 	(ExpHead == Term -> (Res =.. [Rep|NArgs]) ; (Res =.. [ExpHead|NArgs])).

% replace_term_args_([],_,[]).
% replace_term_args_([Arg|Args],Tup,[NArg|NArgs]):-
% 	replace_term_(Arg,Tup,NArg),
% 	replace_term_args_(Args,Tup,NArgs).

% preorder_exp(Exp,[Exp]):-
% 	var(Exp).
% preorder_exp([],[]).
% preorder_exp(Exp,[Exp]):-
% 	num(Exp).
% preorder_exp(Exp,[Exp]):-
% 	atm(Exp).
% preorder_exp(Exp,[F/Ar|PreArgs]):-
% 	\+list(Exp),
% 	functor(Exp,F,Ar),
% 	Exp =.. [_|Args],
% 	preorder_exp(Args,PreArgs0),
% 	list_concat(PreArgs0,PreArgs).
% preorder_exp([Exp|Rest],[Pre|PreRest]):-
% 	preorder_exp(Exp,Pre),
% 	preorder_exp(Rest,PreRest).
% :- export(build_exp/2).
% build_exp(L,Exp):-
% 	build_sub_exp(1,L,[Exp],_).



% build_sub_exp(_,[],[],[]).
% build_sub_exp(N,[P|List],[P|Subs],Rest):-
% 	var(P),!,
% %	build_sub_exp(Ar,List,ListExp,Rest0),
% 	N1 is N -1,
% 	build_sub_exp(N1,List,Subs,Rest).
% build_sub_exp(N,[P/Ar|List],[Sub|Subs],Rest):-
% 	build_sub_exp(Ar,List,ListExp,Rest0),
% 	Sub =.. [P|ListExp],
% 	N1 is N -1,
% 	build_sub_exp(N1,Rest0,Subs,Rest).
% build_sub_exp(N,[P|List],[P|Subs],Rest):-
% %	build_sub_exp(Ar,List,ListExp,Rest0),
% 	N1 is N -1,
% 	build_sub_exp(N1,List,Subs,Rest).


	



expand_functions_in_mathematica(Expr,Solution) :-

	term_variables(Expr,FinalVars),
	length(FinalVars,FinalVarsLength),
	generate_n_vars(FinalVarsLength,VarsAsNumbers),
	substitute_vars(Expr,FinalVars,VarsAsNumbers,MExpr),
	generate_expand_functions_mathematica_text(MExpr,MText),
	( call_mathematica(MText,FinalVars,no,SolutionExpr)
	-> Solution = SolutionExpr
	 ; Solution = Expr
	).

generate_expand_functions_mathematica_text(Expr,MText) :-
	generate_expr_text(Expr,ExprText,true),
	FormatText = "FullForm[FullSimplify[FunctionExpand[~s]]]",
	format_to_string(FormatText,[ExprText],MText).






:- pred solve_in_mathematica(Equation,Var,Solution).
:- doc(solve_in_mathematica/2,
	"Solves the @var{Equation} for the @var{Var} into @var{Solution}.").

%Special trivial cases
solve_in_mathematica(A=B,Y,A):-
	Y==B,!.
solve_in_mathematica(A=B,Y,B):-
	Y==A,!.


solve_in_mathematica(Equation,V,Solution) :-
	% Generate equation without variables
	%note_message("EQUATION>>> ~q",[Equation]),
	term_variables(Equation,FinalVars),
	length(FinalVars,FinalVarsLength),
	generate_n_vars(FinalVarsLength,VarsAsNumbers),
	substitute_vars(Equation,FinalVars,VarsAsNumbers,MExpr),
	find_variable_number(FinalVars,VarsAsNumbers,V,MVar),
	generate_solve_mathematica_text(MExpr,MVar,MText),
	( call_mathematica(MText,FinalVars,yes,SolutionExpr)
	-> Solution = SolutionExpr
	 ; Solution = Equation
	).
%	note_message("EQUATION: ~q (~q) ::: SOLUTiON: ~q",[Equation,V,Solution]).

:- export(find_variable_number/4).
find_variable_number([V|_],[N|_],W,N) :- V == W, !.
find_variable_number([_|R],[_|S],W,N) :- find_variable_number(R,S,W,N).

generate_solve_mathematica_text(Lhs=Rhs,Var,MText) :-
	generate_expr_text(Lhs,LhsText,false),
	generate_expr_text(Rhs,RhsText,false),
	generate_expr_text(Var,VarText,false),
	FormatText = "FullForm[ Solve[~s==~s,~s] ]",
	format_to_string(FormatText,[LhsText,RhsText,VarText],MText).

generate_n_vars(0,[]) :- !.
generate_n_vars(N,Vars) :-
	N > 0, !,
	N1 is N - 1,
	generate_n_vars(N1,Vars1),
	atom_number(Atom,N),
	atom_concat('var',Atom,ThisVar),
	append(Vars1,[ThisVar],Vars).

generate_expr_text(Expr,Text,_) :-
	num(Expr), !,
	number_codes(Expr,Text).
generate_expr_text(Expr,Text,Paren) :-
	Expr =.. [F|Args],
	atom_codes(F,FText),
	generate_expr_text_each(Args,ArgsText,Paren),
	( ArgsText = []
	-> QText = FText
	 ; ( not_bracketed(F)
	   -> ( ArgsText = [OneArg]
	      -> format_to_string("~s~s",[FText,OneArg],QText) % Unary
	       ; ArgsText = [OneArg,TwoArg],
		 format_to_string("~s~s~s",[OneArg,FText,TwoArg],QText) % Binary
	      )
	   ; join_by(ArgsText,",",Inner),
	     format_to_string("~s[~s]",[FText,Inner],QText)
	   )
	),
	( Paren = true
	-> format_to_string("(~s)",[QText],Text)
	 ; Text = QText
	).

generate_expr_text_each([],[],_).
generate_expr_text_each([E|R],[EText|RText],Paren) :-
	generate_expr_text(E,EText,Paren),
	generate_expr_text_each(R,RText,Paren).

not_bracketed(X) :- member(X,['+','-','*','/','==']).

join_by([E],_,E).
join_by([E|R],Sep,Text) :-
	join_by(R,Sep,RText),
	append(E,Sep,ESep),
	append(ESep,RText,Text).

% BEGIN Parser for Mathematica expressions
element(_,N)    --> number(N), !.
element(_,N)    --> negative_number(N), !.
element(Vars,V) --> variable(Vars,V), !.
element(Vars,I) -->
	identifier(F),
	spaces,
	element_list(Vars,Args),
	{
	    process_element(F,Args,I)
	}.

element_list(_,[]) -->
	lbrace,
	spaces,
	rbrace,
	!.
element_list(Vars,L)  -->
	lbrace,
	!,
	element_list_(Vars,L).
element_list(_,[]) --> [].

element_list_(Vars,[E|R]) -->
	element(Vars,E),
	spaces,
	comma,
	spaces,
	element_list_(Vars,R).
element_list_(Vars,[E])   -->
	element(Vars,E),
	spaces,
	rbrace.

variable(Vars,Var) -->
	"var",!,
	number(N),
	{ nth(N,Vars,Var) }.
	
identifier('+') --> "Plus", !.
identifier('*') --> "Times", !.
identifier('/') --> "Rational", !.

identifier(A) -->
	chars_no_space(X),
	{ atom_codes(A,X) }.

process_element(_,Args,'$Failed') :-
	varmember('$Failed',Args), !.
process_element('C',_,'$Failed') :- !.
%% process_element('*',[N,Arg],I) :-
%% 	N == -1, !,
%% 	( ground(Arg), num(Arg)
%% 	-> I is -Arg
%% 	 ; I = -Arg
%% 	).
%% process_element('*',[N,Arg],I) :-
%% 	ground(N), num(N), N < 0, !,
%% 	MinN is -N,
%% 	process_element('*',[MinN,Arg],IIn),
%% 	( ground(IIn), num(IIn)
%% 	-> I is -IIn
%% 	 ; I = -IIn
%% 	).
process_element('*',Args,I) :-
	!, generate_two_by_two('*',Args,I).
process_element('+',Args,I) :-
	!, generate_sum(Args,I).
process_element('Power',[A,B],'Exp'(B)) :-
	A == 'E', !.
%process_element('Power',[A,B],A ^ B) :- !.
process_element('Gamma',[A+B],'Factorial'(Var)) :-
	A == 1, !, Var = B.
process_element('Gamma',[A+B],'Factorial'(Var)) :-
	B == 1, !, Var = A.
process_element('Pochhammer',[1,Var],'Factorial'(Var)) :- !.


process_element(F,Args,I) :-
	I =.. [F|Args].

generate_two_by_two(F,[A,B],Solution) :-
	!, Solution =.. [F,A,B].
generate_two_by_two(F,[C|R],Solution) :-
	generate_two_by_two(F,R,Inner),
	Solution =.. [F,C,Inner].

generate_sum([A],A).
generate_sum([E|R],B) :-
	generate_sum(R,RSum),
	( ground(E), num(E), E < 0
	-> Em is -E, Expr = RSum - Em
	 ; Expr = RSum + E
	),
	( ground(Expr), arithexpression(Expr)
	-> B is Expr
	 ; B = Expr
	).


spaces --> space, spaces.
spaces --> [].

negative_number(N) -->
	"-",
	number(M),
	{ N is -M }.

number(N) -->
	number_parts(X),
	{ number_codes(N,X) }.

number_parts(N) -->
	digits(A),
	dot,
	digits(B),
	{ append(A,[0'.|B],N) }.
number_parts(A) -->
	digits(A).
number_parts([0'0,0'.|B]) -->
	dot,
	digits(B).
	
digits([D|R]) --> digit(D), digits(R).
digits([])    --> [].

chars_no_space([H|T]) --> char_no_space(H), chars_no_space(T).
chars_no_space([])    --> [].

char_no_space(X) --> [X], 
	{\+ comma([X],[]), \+ space([X],[]),
         \+ lbrace([X],[]), \+ rbrace([X],[])}.

dot      --> [0'.].
comma    --> [0',].
lbrace   --> [0'[].
rbrace   --> [0']].
digit(X) --> [X], {X > 47, X<58}.
space    --> [X], {X=<32}. % covers EOF, ^I, ^G, ^J, ^K, ^L, ^M, ^Z, ESC, ^_
%eol      --> [0'
%].
% END Parser for Mathematica expressions
