:- module(eq_solver_term_position,
	[ var_position/3
	, vars_position/2
	, leaves_position/2
	, term_at_position/3
	, terms_at_positions/3
	, substitute_vars/4
	, change_term_at_position/4
	, change_terms_at_positions/4
	],
	[ assertions
	, nativeprops
	]).


% TODO: look if this should not go to library/contrib.

:- use_module(library(lists)).

:- doc(title, "Positions on terms").
:- doc(author, "Alejandro Serrano").
:- doc(module, "This module includes predicates that allow
	to navigate terms using lists of positions.").

:- pred var_position(Term,Var,Position).
:- doc(var_position/3,
	"Gets the position of @var{Var} in @var{Term}.").
var_position(Term,Var,Position) :-
	vars_position(Term,AllPositions),
	terms_at_positions(Term,AllPositions,AllVars),
	find_var_position(AllVars,AllPositions,Var,Position).

find_var_position([A|_],[P|_],B,P) :- A == B, !.
find_var_position([_|R],[_|S],B,P) :- find_var_position(R,S,B,P).

:- pred vars_position(Term,VariablePositions)
	: term * var => term * list(list(num)).
:- doc(vars_position/2,
	"Gets the positions of the variables in a term
         as a list of functor argument positions.").
:- texec pred vars_position(Term,VariablePositions)
	: Term = a(X,b(Y,Z))
       => VariablePositions = [[1],[2,1],[2,2]].
 
% Example:
% ?- vars_position([Z,h(g(Y,X)),C,V],  P).
% P = [[1],[2,1,1,1],[2,1,1,2],[2,2,1],[2,2,2,1]] ? 

vars_position(Var,[[]]) :- var(Var), !.
vars_position(Functor,Positions) :-
	functor(Functor,_,N),
	N > 0, !,
	vars_position_each(Functor,N,1,Positions).
vars_position(_,[]).

vars_position_each(_Functor,N,I,[]) :-
	I > N, !.
vars_position_each(Functor,N,I,VarPositions) :-
	I =< N, !,
	arg(I,Functor,Arg),
	vars_position(Arg,PositionsInArg),
	put_in_front(PositionsInArg,I,VarPositionsI),
	I1 is I + 1,
	vars_position_each(Functor,N,I1,VarPositionsR),
	append(VarPositionsI,VarPositionsR,VarPositions).

:- pred leaves_position(Term,VariablePositions)
	: term * var => term * list(list(num)).
:- doc(leaves_position/2,
	"Gets the positions of the leaves in a term
         as a list of functor argument positions.
         A leaf is a variable or an atom.").
:- texec pred leaves_position(Term,VariablePositions)
	: Term = a(X,b(Y,c))
       => VariablePositions = [[1],[2,1],[2,2]].
leaves_position(Var,[[]]) :- var(Var), !.
leaves_position(Leaf,[[]]) :-
	num(Leaf), !.
%	functor(Leaf,_,0), !.
leaves_position(Functor,Positions) :-
	functor(Functor,_,N),
	N > 0, !,
	leaves_position_each(Functor,N,1,Positions).
leaves_position(_,[]).

leaves_position_each(_Functor,N,I,[]) :-
	I > N, !.
leaves_position_each(Functor,N,I,VarPositions) :-
	I =< N, !,
	arg(I,Functor,Arg),
	leaves_position(Arg,PositionsInArg),
	put_in_front(PositionsInArg,I,VarPositionsI),
	I1 is I + 1,
	leaves_position_each(Functor,N,I1,VarPositionsR),
	append(VarPositionsI,VarPositionsR,VarPositions).

put_in_front([],_,[]).
put_in_front([L|R],N,[[N|L]|NR]) :-
	put_in_front(R,N,NR).

:- pred term_at_position(Term,Position,InnerTerm)
	: term * list(num) * var
       => term * list(num) * term.
:- doc(term_at_position/3,
	"Gets the @var{InnerTerm} found at the
	 specified @var{Position} in @var{Term}").
term_at_position(X,[],X) :- !.
term_at_position(T,[E|R],Inner) :- !,
	arg(E,T,OneStepChild),
	term_at_position(OneStepChild,R,Inner).

:- pred terms_at_positions(Terms,Positions,InnerTerms)
	: term * list(list(num)) * var
       => term * list(list(num)) * list(term).
:- doc(terms_at_positions/3,
	"Gets the @var{InnerTerms} found at the
	 specified @var{Positions} in @var{Term}.").
terms_at_positions(_,[],[]) :- !.
terms_at_positions(Term,[Pos|PosR],[Inner|InnerR]) :- !,
	term_at_position(Term,Pos,Inner),
	terms_at_positions(Term,PosR,InnerR).

:- pred substitute_vars(Term,VariablesInTerm,SubstitutionTerms,SubstitutedTerm)
	: term * list(var) * list(term) * var
       => term * list(var) * list(term) * term.
:- doc(substitute_vars/4,
	"Gets a @var{SubstitutedTerm} by replacing each ocurrence
         of a variable in @var{VariablesInTerm} by the term in
         @var{SubstitutionTerms} in the same position.").
substitute_vars(Var,VarsInTerm,SubstTerms,SubstitutedTerm) :-
	var(Var), !,
	substitute_one_var(Var,VarsInTerm,SubstTerms,SubstitutedTerm).
substitute_vars(Ground,_,_,Ground) :-
	ground(Ground), !.
substitute_vars(Functor,VarsInTerm,SubstTerms,Term) :-
	\+ ground(Functor), \+ var(Functor),
	functor(Functor,F,N),!,
	functor(Term,F,N),
	substitute_vars_each(Functor,1,N,VarsInTerm,SubstTerms,Term).

substitute_one_var(Var,[],[],Var) :- !.
substitute_one_var(Var,[V|_],[T|_],T) :-
	Var == V, !.
substitute_one_var(Var,[V|RV],[_|RT],Subst) :-
	Var \== V, !,
	substitute_one_var(Var,RV,RT,Subst).

substitute_vars_each(_,I,N,_,_,_) :-
	I > N, !.
substitute_vars_each(Functor,I,N,VarsInTerm,SubstTerms,Term) :-
	I =< N, !,
	arg(I,Functor,Arg),
	substitute_vars(Arg,VarsInTerm,SubstTerms,SubstArg),
	arg(I,Term,SubstArg),
	I1 is I + 1,
	substitute_vars_each(Functor,I1,N,VarsInTerm,SubstTerms,Term).

:- pred change_term_at_position(Term,Position,NewInner,NewTerm)
	: term * list(num) * term * var
       => term * list(num) * term * term.
:- doc(change_term_at_position/4,
	"Changes the subterm from @var{Term} at @var{Position}
         with the new @var{InnerTerm} creating @var{NewTerm}.").
change_term_at_position(_,[],NewInner,NewInner) :- !.
change_term_at_position(Term,[N|R],NewInner,NewTerm) :- !,
	functor(Term,F,M),
	functor(NewTerm,F,M),
	change_term_at_position_arg(Term,1,M,N,R,NewInner,NewTerm).

change_term_at_position_arg(Term,I,M,N,R,NewInner,NewTerm) :-
	I =< M,
	arg(I,Term,Arg),
	( I == N
	-> % Changing what we want
	   change_term_at_position(Arg,R,NewInner,NewArg),
	   arg(I,NewTerm,NewArg)
	 ; arg(I,NewTerm,Arg)
	),
	( I < M
	-> I1 is I + 1,
	   change_term_at_position_arg(Term,I1,M,N,R,NewInner,NewTerm)
	 ; true
	).

:- pred change_terms_at_positions(Term,Positions,NewInners,NewTerm)
	: term * list(list(num)) * list(term) * var
       => term * list(list(num)) * list(term) * term.
:- doc(change_terms_at_positions/4,
	"Changes the subterms from @var{Terms} at each of the
         specified @var{Positions} with @var{NewInners},
	 resulting in @var{NewTerm}.").
change_terms_at_positions(Term,[],[],Term) :- !.
change_terms_at_positions(Term,[Pos|PosR],[Inner|InnerR],NewTerm) :- !,
	change_term_at_position(Term,Pos,Inner,NewTerm),
	change_terms_at_positions(Term,PosR,InnerR,NewTerm).