:- module(eq_solver_types,_,[assertions,nativeprops,regtypes]).

:- doc(title, "Types").
:- doc(author, "Alejandro Serrano").
:- doc(module, "Definition of types used in extended recurrence equations.").
:- use_module(library(messages)).
:- use_module(library(hiordlib),[maplist/2]).
%%%%%%%%%%%%%%%%%%%%%%%%%%%% EXPRESSION TYPE DEFINITIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                                    %
% TODO: This was implemented extending arithexpression with function calls.          %
% Needs revision.                                                                    %
%                                                                                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	

% TODO: Add summations and products (as in general_form).
% TODO: Is algebraic_expression/1 really a property?.
:- prop algebraic_expression(X)
	# "@var{X} is an algebraic expression (as we can find in plai size/resource analysis)".

algebraic_expression(X) :- var(X),!.
algebraic_expression(X) :- num(X),!.
algebraic_expression(X) :- algebraic_function_call(X),!.
algebraic_expression(Exp):- algebraic_builtin(Exp),!.
algebraic_expression(-X) :- algebraic_expression(X).
algebraic_expression(+X) :- algebraic_expression(X).
algebraic_expression(--X) :- algebraic_expression(X).
algebraic_expression(++X) :- algebraic_expression(X).
algebraic_expression(X+Y) :- algebraic_expression(X), algebraic_expression(Y).
algebraic_expression(X-Y) :- algebraic_expression(X), algebraic_expression(Y).
algebraic_expression(X*Y) :- algebraic_expression(X), algebraic_expression(Y).
algebraic_expression(X//Y) :- algebraic_expression(X), algebraic_expression(Y).
algebraic_expression(X/Y) :- algebraic_expression(X), algebraic_expression(Y).
algebraic_expression(abs(X)) :- algebraic_expression(X).
algebraic_expression(sign(X)) :- algebraic_expression(X).
algebraic_expression(float_integer_part(X)) :- algebraic_expression(X).
algebraic_expression(float_fractional_part(X)) :- algebraic_expression(X).
algebraic_expression(float(X)) :- algebraic_expression(X).
algebraic_expression(X**Y) :- algebraic_expression(X), algebraic_expression(Y).
algebraic_expression(exp(X)) :- algebraic_expression(X).
algebraic_expression(log(X)) :- algebraic_expression(X).
algebraic_expression(log(A,B)) :- algebraic_expression(A),algebraic_expression(B).
algebraic_expression(sqrt(X)) :- algebraic_expression(X).
algebraic_expression(sin(X)) :- algebraic_expression(X).
algebraic_expression(cos(X)) :- algebraic_expression(X).
algebraic_expression(atan(X)) :- algebraic_expression(X).
algebraic_expression([X]) :- algebraic_expression(X).
algebraic_expression(inf). % Added infinity.
algebraic_expression('?'). % Added unknown value/variable.
algebraic_expression(sum(Var,I,F,Exp)):- var(Var),algebraic_expression(I),algebraic_expression(F),algebraic_expression(Exp),!.
algebraic_expression(prod(Var,I,F,Exp)):- var(Var),algebraic_expression(I),algebraic_expression(F),algebraic_expression(Exp),!.


:- regtype algebraic_function_call(F)
	# "@var{F} is an expression representing a user-defined function call.".

algebraic_function_call(fun(Name,Args)):-
	atom(Name),
	maplist(algebraic_expression,Args),!.

:- regtype algebraic_builtin(F)
	# "@var{F} is an expression representing a builtin function call.".

algebraic_builtin(fact(X)) :- algebraic_expression(X). % Added factorial as a builtin function.
algebraic_builtin(lucasl(X)) :- algebraic_expression(X). % Added fibonacci as a builtin function.
algebraic_builtin(fibo(X)) :- algebraic_expression(X). % Added lucasl as a builtin function.
algebraic_builtin(max(X,Y)):-
	algebraic_expression(X),
	algebraic_expression(Y).
algebraic_builtin(min(X,Y)):-
	algebraic_expression(X),
	algebraic_expression(Y).


:- regtype solver_strategy(S)
	# "@var{S} is a solver strategy.".
% solver_strategy(chain(L)):-
% 	list(L,backend_solver).
% solver_strategy(chain):-
% 	list(L,backend_solver).
solver_strategy(chain).
solver_strategy(St):-
	backend_solver(St).

:- regtype solver_operation(Op)
	# "@var{Op} is an operation implemented by backend solvers".
solver_operation(simplify).
solver_operation(expand).
solver_operation(greater_than).
solver_operation(less_than).
solver_operation(solve_rec_rel).


:- regtype solver_exception(Ex)  #   "@var{Ex} is an exception thrown
by the solver.  @tt{not_implemented/2} arises when for a particular
operation, the chosen backend solver is not able to perform it, either
due to it fails, or there is no implementation  for that operation in
the solver.  @tt{backend_exception/2} is a wrapper for an exception
originated in some backend solver.".


solver_exception(not_implemented(Op,Solver,Exp)):-
	solver_operation(Op),
	backend_solver(Solver),
	list(Exp).

solver_exception(backend_exception(Op,Solver,Args,BEx)):-
	solver_operation(Op),
	backend_solver(Solver),
	list(Args),
	backend_exception(BEx).

:- regtype backend_exception(Ex)
	# "@var{Ex} is an exception thrown by a backend solver".
backend_exception(Ex):-
	mathematica_exception(Ex).

:- regtype mathematica_exception(Ex)
	# "@var{Ex} is an exception thrown by the backend solver mathematica".
mathematica_exception(error(_,_)).


:- regtype backend_solver(BS)
	# "@var{BS} is a backend CAS or Recurrence equations solver or algebraic utility provider".
backend_solver(mathematica).
backend_solver(builtin).
backend_solver(ppl).


% TODO: Define properly (based on element/4 dcg pred (in eq_solver_mathematica), and mathematica syntax documentation.
:- regtype mathematica_expression(E)
	# "@var{E} is an algebraic expression recognized by the Mathematica CAS system.". 
mathematica_expression(X):-
	term(X).

:- regtype predicate_name(Name)
	# "@var{Name} is a predicate name.".
:- doc(predicate_name/1,
	"A predicate name consists of an atomic name
         and an arity, @var{F}/@var{N}:

         @includedef{predicate_name/1}

         ").
predicate_name(F/N) :-
	atom(F),
	int(N).

:- regtype predicate(Pred)
	# "@var{Pred} is the representation of a predicate.".
:- doc(predicate/1,
	"A predicate has a name and a set of clauses:

         @includedef{predicate/1}

        ").
predicate(predicate(Name,InputArgs,OutputArgs,Clauses)) :-
	predicate_name(Name),
	list(InputArgs),
	list(OutputArgs),
	list(Clauses,clause).

:- regtype clause(Clause)
	# "@var{Clause} is the representation of a clause.".
:- doc(clause/1,
	"A clause Clause is composed of its input and output
         arguments and the body of it:

         @includedef{clause/1}

         ").
clause(clause(InputArgs,OutputArgs,Body)) :-
	list(InputArgs),
	list(OutputArgs),
	body(Body).

:- regtype body(Body)
	# "@var{Body} is the body of a clause, that is, a list of literals.".
body(Body) :- list(Body,literal).

:- regtype literal(Lit)
	# "@var{Lit} is one of the allowed literals".
:- doc(literal/1,
	"A literal Lit can be an assignment or a call:

         @includedef{literal/1}

         ").
literal(assign(_X,_Y)).
literal(call(Predicate,InputArgs,OutputArgs)) :-
	predicate_name(Predicate),
	list(InputArgs),
	list(OutputArgs).


:- regtype boolean(B)
	# "Represents a boolean value".
boolean(true).
boolean(false).

:- regtype yesno(B)
	# "Either yes or no atom".
yesno(yes).
yesno(no).




% TODO: There should be a regtype for valid arguments.
:- regtype equation(Eq)
	# "@var{Eq} is an equation.".
equation(equation(Function,Args,Expression,Constraints)):-
	atom(Function),
	list(Args),
	algebraic_expression(Expression),!,
	test_set(Constraints).

:- regtype test_set(C)
	# "@var{C} is a list of constraints (arithmetic comparisons or unifications)".

test_set([]).
test_set([C|R]):-
	test(C),
	test_set(R).

:- regtype test(C)
	# "@var{C} is an arithmetic comparison or a unification (between algebraic expressions)".
test(C):-
	C =.. [Op|Args],
	maplist(algebraic_expression,Args),
	(Op == '>' ; Op == '<' ; Op == '>=' ;
	    Op == '=<' ; Op == '=:=' ;Op == '=').

:- regtype recurrence_relation(Rel)
	# "@var{Rel} is a recurrence relation.".
recurrence_relation(Rel) :-
	maplist(equation,Rel).

% :- regtype bound(B)
% 	# "@var{B} is either lower or upper bound.".
% bound(lower).
% bound(upper).

:- regtype resolution(S)
	# "@var{S} is the resolution of a predicate".
:- doc(resolution/1,
	"A resolution of the predicate it refers consists
         of the structure of input and output arguments
         along with an expression for each variable
         appearing in the latter.

         @includedef{resolution/1}
         @includedef{var_solution/1}
         @includedef{solution/1}

         ").
resolution(resolution(Name,InputArgs,OutputArgs,VarSolutions)) :-
	predicate_name(Name),
	%list(OutputArgs),
	list(InputArgs),
	list(OutputArgs),
	maplist(var_solution,VarSolutions).

var_solution((Var,Solution)) :-
	var(Var),
	solution(Solution).

solution(closed_form(_Expression)).
solution(unsolved_equations(ChosenArgs,Equations)) :-
	list(ChosenArgs),
	list(equation,Equations).

:- pred opposite_bound(X,Y) : bound * var => bound * bound.
:- pred opposite_bound(X,Y) : var * bound => bound * bound.
:- pred opposite_bound(X,Y) : bound * bound => bound * bound.
:- doc(opposite_bound/2,
	"@var{X} is the opposite type of bound to @var{Y}.").
opposite_bound(lower,upper).
opposite_bound(upper,lower).
