:- module(pred_solver,
	[ % example/2,
	  % example2/2,
	  preds_solve/3,
	  preds_solve_/4,
	  pred_solve/8,
	  pred_solve/7,
	  preprocess_clause/8
	],
	[assertions,
	 nativeprops,
	 regtypes,
	 datafacts]).

:- use_module(library(lists)).
%:- use_module(library(sets)).
%:- use_module(library(sort)).
:- use_module(library(terms_vars)).

:- use_module(eq_solver).
:- use_module(eq_solver_graphs).
:- use_module(eq_solver_term_position).
:- use_module(eq_solver_types).

% For storing names of equations.
:- use_module(domain(resources/res_abs_aux),[equation_name/1]).


:- use_module(library(messages)).
% :- multifile issue_debug_messages/1.
% :- data issue_debug_messages/1.
% issue_debug_messages(pred_solver).
%:- doc(filetype,application).

 
% Example
% :- doc(example/2,
% 	"Example of solving an append predicate:

%          @includedef{example/2}

%         ").
% example([X1,X2,Y1,Y2,Z1,Z2],Solution) :-
% 	pred_solve(predicate(append/3,[list(X1,X2),list(Y1,Y2)],[list(Z1,Z2)],
% 	    [ clause([list(0,0),list(B1,B2)],[list(C1,C2)],
% 	        [assign(list(C1,C2),list(B1,B2))]
%               )
% 	    , clause([list(T1,T2),list(U1,U2)],[list(V1,V2)],
% 	        [ assign(list(TB1,TB2),list(T1-1,T2-1))
% 		, call(append/3,[list(TB1,TB2),list(U1,U2)],[list(VB1,VB2)])
% 		, assign(list(V1,V2),list(VB1+1,VB2+1))
% 		]
%               )
% 	    ]), [], 0, true,solve,Solution).

% example2([X1,X2,Y1,Y2],Solution) :-
% 	pred_solve(predicate(nrev/2,[list(X1,X2)],[list(Y1,Y2)],
% 	    [ clause([list(0,0)],[list(0,0)],
% 	        [ ]
% 	      ),
% 	      clause([list(T1,T2)],[list(U1,U2)],
% 	        [ assign(list(TB1,TB2),list(T1-1,T2-1))
% 		, call(nrev/2,[list(TB1,TB2)],[list(UB1,UB2)])
% 		, call(append/3,[list(UB1,UB2),list(1,1)],[list(U1,U2)])
% 		]
%               )
% 	    ]),
% 	    [ resolution(append/3,[list(A1,A2),list(B1,B2)],[list(C1,C2)],
% 	        [ (C1,closed_form(A1+B1))
% 		, (C2,closed_form(A2+B2))
% 		]
% 	      )
% 	    ],
% 	    0,true,solve,
% 	    Solution).

:- doc(title, "Numerical predicate solver").
:- doc(author, "Alejandro Serrano").
:- doc(module, "This module includes predicates that convert
	a set of pseudo-Prolog numerical predicates as used
        in size analysis in numerical functions, when possible.").
	

% TODO: preds_solve should receives the determinacy information for each predicate. At the moment is assuming non-det for all the predicates.
:- pred preds_solve/3 : list(predicate) * atom *  var
	             => list(predicate) * atom * list(resolution).
:- doc(preds_solve(Preds,TrySolve,Resolutions),
	"Solves the set of predicates in @var{Preds}.
         The predicates must be in topological order.").
preds_solve(Preds,TrySolve,Resolutions) :- preds_solve_(Preds,[],TrySolve,Resolutions).

:- pred preds_solve_/4 : list(predicate) * list(resolution) * atom * var
	              => list(predicate) * list(resolution) * atom * list(solution).
:- doc(preds_solve_(Preds,AlreadySolved,TrySolve,Resolutions),
	"Solves the set of predicates in @var{Preds}
         using the information of @var{AlreadySolved} ones.
         The predicates must be in topological order.").
preds_solve_([],AlreadySolved,_TrySolve,AlreadySolved).
preds_solve_([Pred|R],AlreadySolved,TrySolve,Resolutions) :-
	pred_solve(Pred,r,AlreadySolved,0,false,false,TrySolve,Resolution), !,
	preds_solve_(R,[Resolution|AlreadySolved],TrySolve,Resolutions).

:- pred pred_solve/8 : predicate * atom * list(resolution) * num * term * term * atom * var
	            => predicate * atom *list(resolution) * num * term * term * atom * resolution.
:- doc(pred_solve(Preds,AlreadySolved,NbResources,TrySolve,Resolution),
	"Solves one predicates @var{Pred} using the information
         of @var{AlreadySolved} ones. The last @var{NbResources}
         are taken as resource equations.").
pred_solve(predicate(Name,IArgs,OArgs,Clauses),RFlag,AlreadySolved,NbResources,IsDet,CanFail,TrySolve,Resolution) :-
	debug_message("SOLVING: ~q",[predicate(Name,IArgs,OArgs,Clauses)]),
	vars_position(IArgs,IPos),
	vars_position(OArgs,OPos),
	% Substitute known solutions and desugar assignments
	preprocess_clauses(IPos,IArgs,Clauses,AlreadySolved,PreClausesAndGraphs),
	% Get equations for each output position
        generate_equations(Name,IPos,OPos,PreClausesAndGraphs,Equations),
	debug_message("EQ GENERATED>>> ~q",[Equations]),
	solve_recs(Name,RFlag,Equations,IArgs,OArgs,AlreadySolved,NbResources,IsDet,CanFail,TrySolve,Resolution).
	%note_message("SOL: ~q",[Resolution]).

:- pred pred_solve/7 : predicate  * list(resolution) * num * term * term * atom * var
	            => predicate  * list(resolution) * num * term * term * atom * resolution.
:- doc(pred_solve(Preds,AlreadySolved,NbResources,TrySolve,Resolution),
	"Solves one predicates @var{Pred} using the information
         of @var{AlreadySolved} ones. The last @var{NbResources}
         are taken as resource equations.").
pred_solve(Pred,AlreadySolved,NbResources,IsDet,CanFail,TrySolve,Resolution) :-
	pred_solve(Pred,r,AlreadySolved,NbResources,IsDet,CanFail,TrySolve,Resolution).

preprocess_clauses(_,_,[],_,[]) :- !.
preprocess_clauses(IPos,PredIArgs,[E|R],AlreadySolved,[(PreE,Graph,PredVars,ClauseVars)|PreG]) :- !,
	preprocess_clause(IPos,PredIArgs,E,AlreadySolved,PreE,Graph,PredVars,ClauseVars),
	preprocess_clauses(IPos,PredIArgs,R,AlreadySolved,PreG).
	





fix_clause_tests(_IPos, PredIArgs, clause(K,IArgs,_OArgs,_Body,T),_Graph,T):-
	error_message("NOT IMPLEMENTED YET!>>> ~q(~q, ~q): ~q",[K,PredIArgs,IArgs,T]).



:- pred preprocess_clause/8
	: list(list(num)) * list * clause * list(resolution) * var * var * var * var
       => list(list(num)) * list * clause * list(resolution) * clause * graph * list * list.
:- doc(preprocess_clause(IPos,PredIArgs,Clause,AlreadySolved,
	PreprocessedClause,DepGraph,PredVars,ClauseVars),
	"Preprocesses a @var{Clause} from a predicate,
	 returning both a @var{PreprocessedClause} and
         the variable dependency graph @var{DepGraph}.
         It will use the @var{AlreadySolved} equations.
         Additionally, it defines the mapping between
         variables in predicate head, @var{PredVars} and
         those in the clause heas, @var{ClauseVars}.").
% TODO: Fix the preprocessing step so that cuts are not needed
preprocess_clause(IPos,PredIArgs,clause(K,IArgs,OArgs,Body,T),AlreadySolved,
	clause(K,IArgs,OArgs,PreBody,T),Graph,PredVars,ClauseVars) :-
        substitute_solutions_desugar_assignments_and_var_dependency_graph(Body,AlreadySolved,PreBody,Graph),
	% fix_clause_tests(IPos,PredIArgs,clause(K,IArgs,OArgs,Body,T), Graph, NT),
	%!, substitute_solutions(Body,AlreadySolved,BodySoluted),
	%!, desugar_assignments(BodySoluted,PreBody),
	%!, var_dependency_graph(PreBody,Graph),
	!, pred_to_clause_mapping(IPos,PredIArgs,IArgs,PredVars,ClauseVars),
	!.

substitute_solutions_desugar_assignments_and_var_dependency_graph([],_,[],graph([],[])).
substitute_solutions_desugar_assignments_and_var_dependency_graph([call(PredName,CallIArgs,CallOArgs)|R],AlreadySolved,Body,Graph) :-
	member(resolution(PredName,SolIArgs,SolOArgs,TheSolution),AlreadySolved),!,
	create_solution_expr(resolution(PredName,SolIArgs,SolOArgs,TheSolution),SolExpr),
	vars_position(SolIArgs,SolPos),
	terms_at_positions(SolIArgs,SolPos,SolVars),
	terms_at_positions(CallIArgs,SolPos,CallVars),
	substitute_vars(SolExpr,SolVars,CallVars,SubstitutedSolution),
	create_assign(CallOArgs,SubstitutedSolution,NAssign),
	!,substitute_solutions_desugar_assignments_and_var_dependency_graph2([NAssign|R],AlreadySolved,Body,Graph).
substitute_solutions_desugar_assignments_and_var_dependency_graph([E|R],AlreadySolved,Body,Graph) :-
	substitute_solutions_desugar_assignments_and_var_dependency_graph2([E|R],AlreadySolved,Body,Graph).

substitute_solutions_desugar_assignments_and_var_dependency_graph2([assign(X,Y)|R],AlreadySolved,Body,Graph) :-
	substitute_solutions_desugar_assignments_and_var_dependency_graph(R,AlreadySolved,RBody,RGraph),
	desugar_one_assignment(X,Y,Desugared),
	append(Desugared,RBody,Body),
	update_var_dependency_graph(Desugared,RGraph,Graph).
substitute_solutions_desugar_assignments_and_var_dependency_graph2([call(P,IArgs,OArgs)|R],AlreadySolved,[call(P,IArgs,OArgs)|RBody],Graph) :-
	substitute_solutions_desugar_assignments_and_var_dependency_graph(R,AlreadySolved,RBody,RGraph),
	update_var_dependency_graph([call(P,IArgs,OArgs)],RGraph,Graph).

create_assign(OArgs,Subs,assign(OArgs,NSubs)):-
	generate_matching_subs(OArgs,Subs,NSubs).

:- use_module(library(terms_check),[instance/2]).
:- export(generate_matching_subs/3).
generate_matching_subs([],_,[]).
generate_matching_subs([O|OT],[S|ST],[S|NST]):-
	instance(S,O),!,
	generate_matching_subs(OT,ST,NST).
generate_matching_subs([O|OT],[S|ST],[S|NST]):-
	O =.. [Res,(CC,_,_,_)],
	S =.. [Res,(CC,_,_,_)],!,	
	generate_matching_subs(OT,ST,NST).
generate_matching_subs([O|OT],Subs,[NO|NST]):-
	O =.. [Res,(CC,Env,_,_)],
	NO =.. [Res,(CC,Env,0,0)],!,
	generate_matching_subs(OT,Subs,NST).


update_var_dependency_graph([],Graph,Graph).
update_var_dependency_graph([assign(X,Y)|R],Graph,NewGraph) :-
	ground(Y), !,
	update_var_dependency_graph(R,Graph,RGraph),
	add_edges_between(RGraph,assign(X,Y),[X],[Y],NewGraph).
update_var_dependency_graph([assign(X,Y)|R],Graph,NewGraph) :-
	update_var_dependency_graph(R,Graph,RGraph),
	varset(Y,RhsVars),
	add_edges_between(RGraph,assign(X,Y),[X],RhsVars,NewGraph).
update_var_dependency_graph([call(P,IArgs,OArgs)|R],Graph,NewGraph) :-
	update_var_dependency_graph(R,Graph,RGraph),
	varset(IArgs,IVars),
	varset(OArgs,OVars),
	add_edges_between(RGraph,call(P,IArgs,OArgs),OVars,IVars,NewGraph).

%% :- pred substitute_solutions/3 : body * list(resolution) * var
%% 	                      => body * list(resolution) * body.
%% :- doc(substitute_solutions(Body,AlreadySolved,BodySoluted),
%% 	"Substitute the already known solutions to predicates
%%          in @var{AlreadySolved} in the @var{Body}.").
%% substitute_solutions([],_,[]).
%% substitute_solutions([call(PredName,CallIArgs,CallOArgs)|R],AlreadySolved,[assign(CallOArgs,SubstitutedSolution)|SubstR]) :-
%% 	member(resolution(PredName,SolIArgs,SolOArgs,TheSolution),AlreadySolved),!,
%% 	create_solution_expr(resolution(PredName,SolIArgs,SolOArgs,TheSolution),SolExpr),
%% 	vars_position(SolIArgs,SolPos),
%% 	terms_at_positions(SolIArgs,SolPos,SolVars),
%% 	terms_at_positions(CallIArgs,SolPos,CallVars),
%% 	substitute_vars(SolExpr,SolVars,CallVars,SubstitutedSolution),
%% 	substitute_solutions(R,AlreadySolved,SubstR).
%% substitute_solutions([NotCall|R],AlreadySolved,[NotCall|SubstR]) :-
%% 	substitute_solutions(R,AlreadySolved,SubstR).

create_solution_expr(resolution(_,_,OArgs,VarSols),SolExpr) :-
	create_var_expr_pairs(VarSols,Vars,Exprs),
	substitute_vars(OArgs,Vars,Exprs,SolExpr).

create_var_expr_pairs([],[],[]).
create_var_expr_pairs([(V,closed_form(Expr))|R],[V|VarsR],[Expr|ExprsR]) :-
	create_var_expr_pairs(R,VarsR,ExprsR).
create_var_expr_pairs([(V,unsolved_equations(ChosenArgs,[equation(F,_,_,_)|_]))|R],[V|VarsR],[Expr|ExprsR]) :-
	%get_unsolved_eqs_params(Eqs,F,ChosenArgs),
	Expr = fun(F,ChosenArgs),
	%Expr =.. [F|ChosenArgs],
	create_var_expr_pairs(R,VarsR,ExprsR).

%% get_unsolved_eqs_params([equation(F,Args,_,_)],F,Args) :- !.
%% get_unsolved_eqs_params([equation(_,Args,_,_)|R],F,NewArgs) :-
%% 	get_unsolved_eqs_params(R,F,RArgs),
%% 	merge_unsolved_eqs_params(RArgs,Args,NewArgs).
%% 
%% merge_unsolved_eqs_params([],[],[]).
%% merge_unsolved_eqs_params([A|X],[B|Y],[C|Z]) :-
%% 	( ground(A)
%% 	-> C = B
%% 	 ; C = A
%% 	),
%% 	merge_unsolved_eqs_params(X,Y,Z).

%% :- pred desugar_assignments/2 : body * var => body * body.
%% :- doc(desugar_assignments(Body,DesugaredBody),
%% 	"Converts all assignments of complex structures in @var{Body}
%%          into a bunch of different simple assignments of the form
%%          @var{Variable} = @var{Term}.").
%% desugar_assignments([],[]).
%% desugar_assignments([assign(X,Y)|R],DesugaredBody) :-
%% 	!,
%% 	desugar_one_assignment(X,Y,DesugaredLiterals),
%% 	desugar_assignments(R,DesugaredRest),
%% 	append(DesugaredLiterals,DesugaredRest,DesugaredBody).
%% desugar_assignments([E|R],[E|DesugaredRest]) :-
%% 	desugar_assignments(R,DesugaredRest).

desugar_one_assignment(Lhs,Rhs,Literals) :-
	vars_position(Lhs,Positions),
	desugar_each_variable(Lhs,Rhs,Positions,Literals).

desugar_each_variable(_Lhs,_Rhs,[],[]) :- !.
desugar_each_variable(Lhs,Rhs,[Position|R],[assign(LhAs,RhAs)|RestAssigns]) :- 
	term_at_position(Lhs,Position,LhAs),
	term_at_position(Rhs,Position,RhAs),!,
	desugar_each_variable(Lhs,Rhs,R,RestAssigns).

%%MKL: Added for supporting res_plai_stprf analysis%%
desugar_each_variable(Lhs,Rhs,[Position|R],[assign(LhAs,0)|RestAssigns]) :- !,
	term_at_position(Lhs,Position,LhAs),
	desugar_each_variable(Lhs,Rhs,R,RestAssigns).



:- pred var_dependency_graph/2 : body * var => body * graph.
:- doc(var_dependency_graph(Body,Graph),
	"Gets the directed dependency @var{Graph} between
         variables in the desugared @var{Body}.").
var_dependency_graph([],graph([],[])).
var_dependency_graph([assign(X,Y)|R],Graph) :-
	ground(Y), !,
	var_dependency_graph(R,RGraph),
	add_edges_between(RGraph,assign(X,Y),[X],[Y],Graph).
var_dependency_graph([assign(X,Y)|R],Graph) :-
	var_dependency_graph(R,RGraph),
	varset(Y,RhsVars),
	add_edges_between(RGraph,assign(X,Y),[X],RhsVars,Graph).
var_dependency_graph([call(P,IArgs,OArgs)|R],Graph) :-
	var_dependency_graph(R,RGraph),
	varset(IArgs,IVars),
	varset(OArgs,OVars),
	add_edges_between(RGraph,call(P,IArgs,OArgs),OVars,IVars,Graph).

add_edges_between(G,_,[],_,G) :- !.
add_edges_between(G,Label,[E|R],Dests,NewG) :- !,
	insert_node(G,E,G1),
	add_edges_with_origin(G1,Label,E,Dests,G2),
	add_edges_between(G2,Label,R,Dests,NewG).

add_edges_with_origin(G,_,_,[],G) :- !.
add_edges_with_origin(G,Label,O,[D|R],NewG) :- !,
	insert_node(G,D,G1),
	insert_edge(G1,(O,Label,D),G2),
	add_edges_with_origin(G2,Label,O,R,NewG).

:- pred pred_to_clause_mapping(IPos,PredArgs,ClauseArgs,PredVars,ClauseVars)
	: list(list(num)) * list * list * var * var
       => list(list(num)) * list * list * list * list.
:- doc(pred_to_clause_mapping/5,
	"Creates the mapping between variables in @var{PredArgs}
         to variables in @var{ClauseArgs}, so later the latter
         can be replaced with the former.").
pred_to_clause_mapping([],_,_,[],[]).
pred_to_clause_mapping([Pos|R],PredArgs,ClauseArgs,[PredVar|PredR],[ClauseVar|ClauseR]) :-
	term_at_position(ClauseArgs,Pos,ClauseVar),
	var(ClauseVar), !,
	term_at_position(PredArgs,Pos,PredVar),
	pred_to_clause_mapping(R,PredArgs,ClauseArgs,PredR,ClauseR).
pred_to_clause_mapping([Pos|R],PredArgs,ClauseArgs,PredR,ClauseR) :-
	term_at_position(ClauseArgs,Pos,ClauseVar),
	nonvar(ClauseVar), !,
	pred_to_clause_mapping(R,PredArgs,ClauseArgs,PredR,ClauseR).

:- pred generate_equations(PredName,IPos,OPos,ClausesAndGraphs,Equations)
	: predicate_name * list(list(num)) * list(list(num)) * list * var
       => predicate_name * list(list(num)) * list(list(num)) * list * list(recurrence_relation).
:- doc(generate_equations/5,
	"Creates the set of @var{Equations} from the
         preprocessed @var{ClausesAndGraphs}, using the
         input and output positions.").

% For each output position
generate_equations(_,_,[],_,[]) :- !.
generate_equations(PredName,IPos,[OPos|OR],Clauses,[Eq|EqR]) :- !,
	generate_equations_pos(PredName,IPos,OPos,Clauses,Eq),
	generate_equations(PredName,IPos,OR,Clauses,EqR).


:- use_module(domain(resources/eq_solver/solver/solver_utils_trans)).
% For each clause
generate_equations_pos(_,_,_,[],[]) :- !.
generate_equations_pos(PredName,IPos,OPos,[(clause(_K,IArgs,OArgs,_,T),Graph,PredV,ClauseV)|R],[Eq|EqR]) :- !,
	% Get the variable name for this case
	term_at_position(OArgs,OPos,TheArg),
	% Generate equation in terms of clause input
	generate_equation_name(PredName,OPos,EqName),
	generate_equation_clause(EqName,PredName,IArgs,T,TheArg,Graph,EqClause),
%	mathequations2equations([EqClause0],[EqClause]),
	debug_message("EQUATION>>>~q {~q}",[EqClause,T]),
	% Change to be in terms of predicate input
	substitute_vars(EqClause,ClauseV,PredV,Eq),
	generate_equations_pos(PredName,IPos,OPos,R,EqR).



%DONE%TODO: Modify this function. The names of the functions should be more
%readable. These should not be created thinking in Mathematica notation/limitations.
% Generate the name of the function to solve
% wich is name_arity_position


generate_equation_name(PName,Pos,Name) :-
	( PName = F/N
	-> atom_concat(F,':',FSlash),
	   atom_number(NAtom,N),
	   atom_concat(FSlash,NAtom,PredAtom)
	 ; PredAtom = PName
	),
	generate_position_list(Pos,PosAtom),
	atom_concat(PredAtom,PosAtom,AlmostName),
	atom_codes(AlmostName,AlmostCodes),
	%Codes = AlmostCodes,
	replace_bad_chars_in_name(AlmostCodes,Codes),
	atom_codes(Name,Codes),!,
	asserta_fact(equation_name(Name)).

generate_position_list([],'').
generate_position_list([N|R],Atom) :-
	generate_position_list(R,RAtom),
	atom_number(NAtom,N),
	atom_concat(NAtom,RAtom,NoSlashAtom),
	atom_concat('z',NoSlashAtom,Atom).

replace_bad_chars_in_name([],[]).
replace_bad_chars_in_name([0'_|R],[0'z|RepR]) :-
	!, replace_bad_chars_in_name(R,RepR).
replace_bad_chars_in_name([0':|R],[0'z,0'z|RepR]) :-
	!, replace_bad_chars_in_name(R,RepR).
replace_bad_chars_in_name([E|R],[E|RepR]) :-
	replace_bad_chars_in_name(R,RepR).

% Here we do the work for a position and clause
generate_equation_clause(EqName,PredName,IArgs,T,TheArg,RGraph,Eq) :-
	PredName = Function/N,
	% Get the result in this position
	( ground(TheArg)
	-> Eq = equation(EqName,IArgs,TheArg,T)
	 ; neighbours(RGraph, TheArg, Nb),
	   ( Nb = [(Label,_)|_]
	   -> ( Label = assign(_,Y)
	      -> CurrentEq = Y % Bare assignment
	       ; ( Label = call(FName, CallArgs, OutArgs)
		 -> ( ( ( FName = F/M, F == Function, M == N )
		      ; ( FName == Function, F = FName )
		      )
		    -> % We have to find the good equation name
		       var_position(OutArgs,TheArg,Position),
		       generate_equation_name(PredName,Position,ThisEqName),
		       CurrentEq = fun(ThisEqName,CallArgs)
		       %CurrentEq =.. [ThisEqName|CallArgs]
		     ; ( FName = F/M ; F = FName ),
			 CurrentEq = fun(F,CallArgs)
		    %CurrentEq =.. [F|CallArgs] % Function call
		    )
		  ; fail
		 )
	      ) ,
	      % Now we have to recursively replace things
	      term_variables(CurrentEq,EqVars),
	      replace_vars_in_equation(EqName,EqVars,CurrentEq,PredName,IArgs,T,RGraph,FinalEq),
	      Eq = equation(EqName,IArgs,FinalEq,T)
	    ; % We have no neighbours, it's done
	      Eq = equation(EqName,IArgs,TheArg,T)
	   )
	).
generate_equation_clause(EqName,PredName,IArgs,T,TheArg,RGraph,Eq) :-
	% Get the result in this position
	( ground(TheArg)
	-> Eq = equation(EqName,IArgs,TheArg,T)
	 ; neighbours(RGraph, TheArg, Nb),
	   ( Nb = [(Label,_)|_]
	   -> ( Label = assign(_,Y)
	      -> CurrentEq = Y % Bare assignment
	       ; ( Label = call(F, CallArgs, OutArgs)
		 -> ( F == PredName
		    -> % We have to find the good equation name
		       var_position(OutArgs,TheArg,Position),
		       generate_equation_name(PredName,Position,ThisEqName),
		       CurrentEq = fun(ThisEqName,CallArgs)
		       %CurrentEq =.. [ThisEqName|CallArgs]
		    ; %CurrentEq =.. [F|CallArgs] % Function call
			CurrentEq = fun(F,CallArgs) % Function call
		    )
		  ; fail
		 )
	      ) ,
	      % Now we have to recursively replace things
	      term_variables(CurrentEq,EqVars),
	      replace_vars_in_equation(EqName,EqVars,CurrentEq,PredName,IArgs,T,RGraph,FinalEq),
	      Eq = equation(EqName,IArgs,FinalEq,T)
	    ; % We have no neighbours, it's done
	      Eq = equation(EqName,IArgs,TheArg,T)
	   )
	).

replace_vars_in_equation(_,[],Eq,_,_,_,_,Eq) :- !.
replace_vars_in_equation(EqName,[V|R],Eq,PredName,IArgs,T,RGraph,OutEq) :-
	( ground(V) ; varmember(V,IArgs) ), !,
	replace_vars_in_equation(EqName,R,Eq,PredName,IArgs,T,RGraph,OutEq).
replace_vars_in_equation(EqName,[V|R],Eq,PredName,IArgs,T,RGraph,OutEq) :-
	generate_equation_clause(EqName,PredName,IArgs,T,V,RGraph,equation(_,_,InnerEq,_)),
	substitute_vars(Eq,[V],[InnerEq],StepEq),
	replace_vars_in_equation(EqName,R,StepEq,PredName,IArgs,T,RGraph,OutEq).