:- module(solver_utils,
	[
	    simplify/3,
	    simplify/2,
	    expand/3,
	    expand/2,
	    solve/4,
	    solve/3,
	    greater_than/3,
	    greater_than/2,
	    solve_rec_rel/5,
	    solve_rec_rel/4,
	    greater_or_equal_than/2,
	    greater_or_equal_than/3,
	    equal_than/2,
	    equal_than/3,
	    clean_solver_cache/0,
	    order_of_function/2
	],
	[assertions,
	 nativeprops,
	 regtypes,
	 datafacts]).

:- doc(title, "Solver Utilities").
:- doc(author, "Maximiliano Klemen").
:- doc(module, "API for accessing to services provided by the algebraic/recurrence solvers").

:- use_module(library(terms_check),[instance/2]).

% TODO: Handle algebraic expressions that contains '?' as a subexpression.
% (or maybe take off '?' from algebraic_expression definition).

:- use_module(domain(resources/eq_solver/solver/solver_messages),[show_user_message/1]).
:- use_module(domain(resources/eq_solver/eq_solver_types),
	[
	    backend_solver/1,
	    recurrence_relation/1,
	    algebraic_expression/1
	]
).

:- use_module(domain(resources/eq_solver/solver/solver_utils_ppl),
	[
	    solve_special_rec_rel_ppl/3
	]
).

:- use_module(domain(resources/eq_solver/solver/solver_utils_builtin),
	[
	    simplify_in_ciaopp/2,
	    solve_rec_ciaopp_solver/2,
	    greater_than_ciaopp/2,
	    greater_or_equal_than_ciaopp/2,
	    equal_than_ciaopp/2,
	    order_of_function/2
	]
).

:- use_module(domain(resources/eq_solver/solver/solver_utils_mathematica),
	[
	    simplify_in_mathematica/2,
	    expand_in_mathematica/2,
	    solve_in_mathematica/3,
	    solve_rec_rel_with_mathematica/3,
	    greater_than_mathematica/2,
	    greater_or_equal_than_mathematica/2,
	    equal_than_mathematica/2
	]
	     ).
:- use_module(domain(resources/eq_solver/solver/solver_utils_translations),[translate/4]).
:- use_module(ciaopp(preprocess_flags)).
:- use_module(library(messages),[debug_message/2,error_message/2]).

:- pred simplify(Solver,Exp,SimpExp):  (backend_solver(Solver),
	algebraic_expression(Exp),var(SimpExp)) =>
	(algebraic_expression(SimpExp)) # "@var{SimpExp} is a
      simplified expression equivalent  to @var{Exp}, obtained using
      the @var{Solver}  backend solver.".

% simplify(A,B,C):-
% 	statistics(runtime,[_,_]),
% 	statistics(runtime,[G,_]),
% 	simplify_(A,B,C),
% 	statistics(runtime,[Gp,_]),
% 	GG is round(1000*(Gp - G))/1000,
% 	counting_time(GG).

simplify(_,N,N):-
	var(N),!.
simplify(_,N,N):-
	num(N),!.
simplify(_,N,N):-
	N == inf,!.
simplify(_,X,X):-
	X == '?',!. 
simplify(_,max(A,B),A):-
	B == '?',!.
simplify(_,max(A,B),B):-
	A == '?',!. 
simplify(_,min(A,B),A):-
	B == '?',!.
simplify(_,min(A,B),B):-
	A == '?',!. 
simplify(_,max(A,_B),A):-
	A == inf,!. 
simplify(_,max(_A,B),B):-
	B == inf,!. 
simplify(_,max(A,_B),A):-
	A == inf,!. 
simplify(_,max(A,B),B):-
	A == 0,!. 
simplify(_,max(A,B),A):-
	B == 0,!. 
simplify(_,min(A,B),A):-
	B == inf,!. 
simplify(_,min(A,B),B):-
	A == inf,!. 
simplify(_,min(A,_B),A):-
	A == 0,!. 
simplify(_,min(_A,B),B):-
	B == 0. 
simplify(mathematica,Exp,Simp):-
	translate(algebraic,mathematica,Exp,Exp0),
	solver_utils_mathematica:simplify_in_mathematica(Exp0,Simp0),
	translate(mathematica,algebraic,Simp0,Simp),!.
simplify(builtin,Exp,Simp):-
	translate(algebraic,builtin,Exp,Exp0),
	solver_utils_builtin:simplify_in_ciaopp(Exp0,Simp0),
	translate(builtin,algebraic,Simp0,Simp),!.
simplify(Solver,Exp,Exp):-
	throw(not_implemented(simplify,Solver,[Exp])).

:- pred simplify(Exp,SimpExp):
	(algebraic_expression(Exp),var(SimpExp)) =>
	 (algebraic_expression(SimpExp)) # "@var{SimpExp} is a
      simplified expression equivalent to @var{Exp}, obtained using
      the default backend_solver.".

simplify(Exp,SimpExp):-
	current_pp_flag(math_system,Solver),
	simplify(Solver,Exp,SimpExp).

:- pred expand(Solver,Exp,ExpExp): (backend_solver(Solver),
	algebraic_expression(Exp),var(ExpExp)) =>
	(algebraic_expression(ExpExp)) # "@var{ExpExp} is the expanded
      equivalent version of the expression @var{Exp}, obtained using
      @var{Solver} as the backend solver".

% expand(A,B,C):-
% 	statistics(runtime,[_,_]),
% 	statistics(runtime,[G,_]),
% 	expand_(A,B,C),
% 	statistics(runtime,[Gp,_]),
% 	GG is round(1000*(Gp - G))/1000,
% 	counting_time(GG).

expand(_,X,_Y):-
	(X=='?'),!,fail.
expand(mathematica,Exp,ExpExp):-
	translate(algebraic,mathematica,Exp,Exp0),
	solver_utils_mathematica:expand_in_mathematica(Exp0,ExpExp0),
	translate(mathematica,algebraic,ExpExp0,ExpExp),!.
expand(Solver,Exp,Exp):-
	throw(not_implemented(expand,Solver,[Exp])).

:- pred expand(Exp,ExpExp): (algebraic_expression(Exp),var(ExpExp)) =>
	(algebraic_expression(ExpExp)) # "@var{ExpExp} is the expanded
	equivalent version of the expression @var{Exp}, obtained using
	the default backend solver".

expand(Exp,ExpExp):-
	current_pp_flag(math_system,Solver),
	expand(Solver,Exp,ExpExp).

:- pred solve(Solver,Exp,Var,Sol): (backend_solver(Solver),
	algebraic_expression(Exp),var(Var),var(ExpExp)) =>
	(algebraic_expression(Sol)) # "@var{Exp} is an unification of
      the form L=R, where L and R are algebraic_expressions,
      possibly containing @var{Var} as a subexpression. @var{Sol} is
      the expression that results from isolating the variable
      @var{Var} in the equation @var{Exp}.  This operation is
      performed by @var{Solver} backend solver.".

% solve(A,B,C,D):-
% 	statistics(runtime,[_,_]),
% 	statistics(runtime,[G,_]),
% 	solve_(A,B,C,D),
% 	statistics(runtime,[Gp,_]),
% 	GG is round(1000*(Gp - G))/1000,
% 	counting_time(GG).

solve(_,Exp1=Exp2,Var,Exp1):-
	Exp2==Var,!.
solve(_,Exp1=Exp2,Var,Exp2):-
	Exp1==Var,!.
solve(_,Exp1=Exp2,Var,Exp2 + 1):-
	Exp1==Var - 1,!.
solve(_,Exp1=Exp2,Var,Exp1 + 1):-
	Exp2==Var - 1,!.
solve(mathematica,Exp1=Exp2,Var,Sol):-
	translate(algebraic,mathematica,Exp1,Exp10),
	translate(algebraic,mathematica,Exp2,Exp20),
	solver_utils_mathematica:solve_in_mathematica(Exp10=Exp20,Var,Sol0),
	translate(mathematica,algebraic,Sol0,Sol).
solve(Solver,Exp,Var,_Sol):-
	throw(not_implemented(solve,Solver,[Exp,Var])).

:- pred solve(Exp,Var,Sol):
	(algebraic_expression(Exp),var(Var),var(ExpExp)) =>
	 (algebraic_expression(Sol)) # "Same as @pred{solve/4}, but
      using the default backend solver".  

solve(Exp,Var,Sol):-
	current_pp_flag(math_system,Solver), 
	solve(Solver,Exp,Var,Sol).

:- pred greater_than(Solver,X,Y):
	(backend_solver(Solver),algebraic_expression(X),algebraic_expression(Y))
 # " Succeeds if @var{X} > @var{Y}, using @var{Solver} as the
      backend solver.".

% greater_than(A,B,C):-
% 	statistics(runtime,[_,_]),
% 	statistics(runtime,[G,_]),
% 	greater_than_(A,B,C),
% 	statistics(runtime,[Gp,_]),
% 	GG is round(1000*(Gp - G))/1000,
% 	counting_time(GG).

greater_than(_,X,Y):-
	(X=='?';Y=='?'),!,
	 fail.
greater_than(mathematica,X,Y):-!,
	translate(algebraic,mathematica,X,X0),
	translate(algebraic,mathematica,Y,Y0),
	solver_utils_mathematica:greater_than_mathematica(X0,Y0).   
greater_than(builtin, X, Y):-!,
	translate(algebraic,builtin,X,X0),
	translate(algebraic,builtin,Y,Y0),
	solver_utils_builtin:greater_than_ciaopp(X0,Y0).   
greater_than(Solver,X,Y):-!,
	throw(not_implemented(greater_than,Solver,[X,Y])).

:- pred greater_than(X,Y):
	(algebraic_expression(X),algebraic_expression(Y))
 # " Succeeds if @var{X} > @var{Y}, using the default backend solver.".

greater_than(X,Y):-
	current_pp_flag(math_system,Solver),
	greater_than(Solver,X,Y).

:- pred greater_or_equal_than(Solver,X,Y):
	(backend_solver(Solver),algebraic_expression(X),algebraic_expression(Y))
 # " Succeeds if @var{X} >= @var{Y}, using @var{Solver} as the
      backend solver.".

% greater_or_equal_than(A,B,C):-
% 	statistics(runtime,[_,_]),
% 	statistics(runtime,[G,_]),
% 	greater_or_equal_than_(A,B,C),
% 	statistics(runtime,[Gp,_]),
% 	GG is round(1000*(Gp - G))/1000,
% 	counting_time(GG).

greater_or_equal_than(_,X,Y):-
	(X=='?';Y=='?'),!,
	 fail.

greater_or_equal_than(mathematica,X,Y):-!,
	translate(algebraic,mathematica,X,X0),
	translate(algebraic,mathematica,Y,Y0),
	catch(solver_utils_mathematica:greater_or_equal_than_mathematica(X0,Y0),_,fail).   
greater_or_equal_than(builtin, X, Y):-!,
	translate(algebraic,builtin,X,X0),
	translate(algebraic,builtin,Y,Y0),
	solver_utils_builtin:greater_or_equal_than_ciaopp(X0,Y0).   
greater_or_equal_than(Solver,X,Y):-!,
	throw(not_implemented(greater_or_equal_than,Solver,[X,Y])).

:- pred greater_or_equal_than(X,Y):
	(algebraic_expression(X),algebraic_expression(Y))
 # " Succeeds if @var{X} >= @var{Y}, using the default backend solver.".

greater_or_equal_than(X,Y):-
	current_pp_flag(math_system,Solver),
	greater_or_equal_than(Solver,X,Y).

:- pred equal_than(Solver,X,Y):
	(backend_solver(Solver),algebraic_expression(X),algebraic_expression(Y))
 # " Succeeds if @var{X} = @var{Y}, using @var{Solver} as the
      backend solver.".

% equal_than(A,B,C):-
% 	statistics(runtime,[_,_]),
% 	statistics(runtime,[G,_]),
% 	equal_than_(A,B,C),
% 	statistics(runtime,[Gp,_]),
% 	GG is round(1000*(Gp - G))/1000,
% 	counting_time(GG).

equal_than(_,X,Y):-
	(X=='?';Y=='?'),!,
	 fail.
equal_than(mathematica,X,Y):-!,
	translate(algebraic,mathematica,X,X0),
	translate(algebraic,mathematica,Y,Y0),
	solver_utils_mathematica:equal_than_mathematica(X0,Y0).   
equal_than(builtin, X, Y):-!,
	translate(algebraic,builtin,X,X0),
	translate(algebraic,builtin,Y,Y0),
	solver_utils_builtin:equal_than_ciaopp(X0,Y0).   
equal_than(Solver,X,Y):-!,
	throw(not_implemented(equal_than,Solver,[X,Y])).


:- pred equal_than(X,Y):
	(algebraic_expression(X),algebraic_expression(Y))
 # " Succeeds if @var{X} = @var{Y}, using the default backend solver.".

equal_than(X,Y):-
	current_pp_flag(math_system,Solver),
	equal_than(Solver,X,Y).


:- pred order_of_function(X,Y):
	(algebraic_expression(X),var(Y)) => algebraic_expression(Y)
 # " Succeeds if @var{Y} is the order of the expression @var{X}.".

order_of_function(X, Y):-!,
	translate(algebraic,builtin,X,X0),
	solver_utils_builtin:order_of_function(X0,Y0),   
	translate(builtin,algebraic,Y0,Y).

:- pred solve_rec_rel(Solver,RecRel,Vars,Bound,Sol):
	(backend_solver(Solver),recurrence_relation(RecRel),
	 list(Vars,var),bound(Bound),var(Sol)) =>
	 (algebraic_expression(Sol)) # "Tries to solve the recurrence
      @var{RecRel}, or find a sound bound @var{Bound}, using
      @var{Solver} as the backend solver.".

% solve_rec_rel(A,B,C,D,E):-
% 	statistics(runtime,[_,_]),
% 	statistics(runtime,[G,_]),
% 	solve_rec_rel_(A,B,C,D,E),
% 	statistics(runtime,[Gp,_]),
% 	GG is round(1000*(Gp - G))/1000,
% 	counting_time(GG).

solve_rec_rel(_,RecRel,_,Bound,Sol):-
	solver_cache(RecRel,Bound,Sol),!.
solve_rec_rel(mathematica,RecRel,Vars,_B,Sol):-
	translate(algebraic,mathematica,RecRel,MRecRel),
	solver_utils_mathematica:solve_rec_rel_with_mathematica(MRecRel,Vars,Sol0),
	translate(mathematica,algebraic,Sol0,Sol),
	save_solution_in_cache(RecRel,_,Sol),!.
solve_rec_rel(builtin,RecRel,_Vars,_B,Sol):-
	translate(algebraic,builtin,RecRel,MRecRel),
	solver_utils_builtin:solve_rec_ciaopp_solver(MRecRel,Sol0),
	translate(builtin,algebraic,Sol0,Sol),
	save_solution_in_cache(RecRel,_,Sol),!.
solve_rec_rel(ppl,RecRel,_Vars,Bound,Sol):-
	solver_utils_ppl:solve_special_rec_rel_ppl(RecRel,Bound,Sol0),
	simplify(builtin,Sol0,Sol),
	save_solution_in_cache(RecRel,Bound,Sol),!.
solve_rec_rel(Solver,RecRel,Vars,Bound,_Sol):-
	debug_message("The following RR could not be solved by ~q: (~q, ~q, ~q)",[RecRel,Vars,Bound, Solver]),!,
	fail.

:- pred solve_rec_rel(RecRel,Vars,Bound,Sol):
	(recurrence_relation(RecRel),
	 list(Vars,var),bound(Bound),var(Sol)) =>
	 (algebraic_expression(Sol)) # "Tries to solve the recurrence
      @var{RecRel}, or find a sound bound @var{Bound}, using the
      default solver.".

solve_rec_rel(RecRel,Vars,Bound,Sol):-
	current_pp_flag(req_solver,Solver),
	solve_rec_rel(Solver,RecRel,Vars,Bound,Sol).


:- data solved/3.
:- export(solved/3).
% solver_cache(R,B,S):-
% 	current_fact(solved(R0,B,S0)),
% 	instance(R,R0),
% 	R = R0,
% 	S = S0,!.
solver_cache(_,_,_):-fail.
save_solution_in_cache(R,B,S):-
	normalize_equations(R,_,NR),
	asserta_fact(solved(NR,B,S)).

clean_solver_cache:-
	retractall_fact(solved(_,_,_)).

:- pred normalize_equations(Eqs,Var,NEqs):
	(list(Eqs,recurrence_relation),var(Var),var(NEqs)) 
 # "Normalize the recurrence relation @var{Eqs}, by replacing the
      name of the equations uniformly by the var @var{Var}.".
normalize_equations([],_,[]).
normalize_equations([X|T],Var,[NX|NT]):-
	normalize_equation(X,Var,NX),
	normalize_equations(T,Var,NT).

%:- export(normalize_equations/3).
normalize_equation(equation(F,Args,Exp,T),NV,equation(NV,Args,NExp,T)):-
	normalize_exp(Exp,F,NV,NExp),!.

normalize_exp(X,_,_,X):-
	var(X),!.
normalize_exp(X,_,_,X):-
	num(X),!.
normalize_exp(fun(F,Args),F,NV,fun(NV,NArgs)):-
	normalize_exp_each(Args,F,NV,NArgs),!.
normalize_exp(Exp,F,NV,NExp):-
	Exp =.. [Op|Args],
	normalize_exp_each(Args,F,NV,NArgs),
	NExp =.. [Op|NArgs],!.

normalize_exp_each([],_,_,[]).
normalize_exp_each([E|T],F,NV,[NE|TT]):-
	normalize_exp(E,F,NV,NE),
	normalize_exp_each(T,F,NV,TT).

% :- export(get_var_function_name/2).
% :- data var_function_name/2.
% get_var_function_name(F,N):-
% 	current_fact(var_function_name(F,'$v'(NV))),NV = N,!.
% get_var_function_name(F,N):-
% 	asserta_fact(var_function_name(F,N)).
% :- use_module(engine(runtime_control),[statistics/2]).
% :- use_module(ciaopp(p_unit)).

% :- export([solver_time_spent/2,solver_clean_times/0]).
% :- data solver_time_spent/2.

% counting_time(Time):-
% 	curr_module(M),
% 	retract_fact(solver_time_spent(M,Previous)),!,
% 	New is Previous + Time,
% 	asserta_fact(solver_time_spent(M,New)).
% counting_time(Time):-
% 	curr_module(M),
% 	asserta_fact(solver_time_spent(M,Time)).

% solver_clean_times:-
% 	retractall_fact(solver_time_spent(_,_)).

