:- module(solver_utils_mathematica,
	[
	    stop_mathematica/0,
	    translate_from_algebraic/2,
	    translate_to_algebraic/2,
	    simplify_in_mathematica/2,
	    expand_in_mathematica/2,
	    solve_in_mathematica/3,
	    solve_rec_rel_with_mathematica/3,
	    greater_than_mathematica/2,
	    greater_or_equal_than_mathematica/2,
	    equal_than_mathematica/2
	],
	 [assertions,
	  nativeprops,
	  regtypes,
	  dcg,
	  datafacts]).

:- use_module(library(aggregates)).
:- use_module(library(format)).
:- use_module(library(lists)).
:- use_module(library(sort)).
:- use_module(library(sets)).
:- use_module(library(terms_check)).
:- use_module(library(terms_vars)).
:- use_module(library(iso_misc)).
:- use_module(library(messages)).

:- use_module(ciaopp(preprocess_flags), [current_pp_flag/2] ).
:- use_module(domain(resources/term_utilities)).

:- use_module(domain(resources/eq_solver/eq_solver_graphs), [varmember/2]).
:- use_module(domain(resources/eq_solver/eq_solver_term_position), [substitute_vars/4]).
:- use_module(domain(resources/eq_solver/solver/solver_utils_trans), [wrap_math_fun/2, unwrap_fun/2]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%CODE ORIGINALLY IN eq_solver_mathematica MODULE%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% TODO: Revision to remove unused code.

:- doc(call_mathematica/4,
	"Calls Mathematica to perform the actions in @var{String}.
         Then the result is parsed back, using the correspondence
         between variable numbers and actual variables in @var{Var}.").

%% In order to communicate with mathematica using 
%% Java interface, uncomment the following line.

%:- compilation_fact(eq_solver_mathematica__use_java_interface).

:- if(defined(eq_solver_mathematica__use_java_interface)).

:- data mathematica_started/2.
%mathematica_started(no,nil).

get_mathematica_handle(H) :-
	( current_fact(mathematica_started(yes,H))
	; mathematica_jlink(ClassPath),
	  mathematica_kernel_link(KLink),
	  catch(java_start(ClassPath),java_exception(_),true),
	  java_create_object('com.wolfram.jlink.MathLinkFactory',MLFactory),
	  java_invoke_method(MLFactory,createKernelLink(KLink,H)),
	  java_invoke_method(H,discardAnswer(_)),
	  asserta_fact(mathematica_started(yes,H))
	).

stop_mathematica:-
	(
	    retract_fact(mathematica_started(yes,_)) ->
	    java_stop
	; 
	    true
	).

%Call to mathematica using Java Interface.
call_mathematica(String,Vars,Rule,Output) :-
%	note_message("JAVA: ~s",[String]),nl,
	get_mathematica_handle(H),
	java_invoke_method(H,evaluateToOutputForm(String, 0, U)), !,
	debug_message("MATH RAW RESULT >>> ~q",[U]),
	element(Vars,O,U,[]),
	% If we receive $Failed, we should not succeed
	( ground(O), O = '$Failed'
	-> fail
	 ; !, ( Rule = yes
	   -> O =.. ['List',Solutions|_],
	      Solutions =.. ['List',R|_],
	      R = 'Rule'(_,Output)
	    ; Output = O
	   )
%	   note_message("JAVA Output: ~q",[Output]),nl
	).

:- else.


%:- use_module(library(mathematica)).
:- import(selected_mathematica_interface_auto,[
evaluate/2 ,
evaluate/2 ,
close_link/0	
	]).


%Call to mathematica using mathematica_link module. 
call_mathematica(String,Vars,Rule,Output) :-
%	note_message("CMD: ~s",[String]),nl,
 	% Wrap with OutputForm command:
 	append("OutputForm[",String,TmpStr),
 	append(TmpStr,"]",CmdString),
 	evaluate(CmdString,OutputChars),
 	  element(Vars,O,OutputChars,[]),
	  ( ground(O), O = '$Failed'
 	  -> fail
 	  ; !, ( Rule = yes
 	       -> O =.. ['List',Solutions|_],
 	       Solutions =.. ['List',R|_],
 	       R = 'Rule'(_,Output)
 	       ; Output = O
 	       )
 	  ).
%foo.


%Call to mathematica using mathematica_link module. 
call_mathematica_fv(String,Vars,FreeVar,Rule,Output) :-
%	note_message("CMD: ~s",[String]),nl,
 	% Wrap with OutputForm command:
 	append("OutputForm[",String,TmpStr),
 	append(TmpStr,"]",CmdString),
 	evaluate(CmdString,OutputChars),
 	  elementfv(Vars,FreeVar,O,OutputChars,[]),
	  ( ground(O), O = '$Failed'
 	  -> fail
 	  ; !, ( Rule = yes
 	       -> O =.. ['List',Solutions|_],
 	       Solutions =.. ['List',R|_],
 	       R = 'Rule'(_,Output)
 	       ; Output = O
 	       )
 	  ).
%foo.



stop_mathematica:-
	close_link.

:- endif.

:- pred solve_rec_rel_with_mathematica(Eqs,Vars,Solution).
:- doc(solve_rec_rel_with_mathematica/3,
	"Solves the recurrence relation posed by @var{Eqs}
         on variables @var{Vars} into its @var{Solution}.").
solve_rec_rel_with_mathematica(Eqs,Vars,Solution) :-
	catch(
		 (
		     convert_to_mathematica(Eqs,MEqs,MFns,MVars,FreeVars),
		     generate_mathematica_text(MEqs,MFns,MVars,MText),
%		     note_message("Mathematica: ~s",[MText]),nl,
		     \+ member(63,MText), % Checks if there is any '?'
		     call_mathematica_fv(MText,Vars,FreeVars,yes,Solution),
		     \+ contains_sum_expression(Solution) % By now we are not handling Summations.
		 ),
		 Ex,
		 (
		     (mathematica_exception(Ex) -> 
		     throw(backend_exception(solve_rec_rel,mathematica,[Eqs,Vars],Ex));throw(Ex))
		 )
	     ).   



:- pred contains_sum_expression(F): (mathematica_expression(F)) #
	"Checks if the expression contains terms with the functor 'Sum' as
      subexpression".
      
contains_sum_expression(F):-
	var(F),!,fail.
contains_sum_expression(F):-
	F =.. ['Sum'|_],!.
contains_sum_expression(F):-
	F =.. [_|Args],!,
	contains_sum_expression_args(Args).

contains_sum_expression_args([]):-
	!,fail.
contains_sum_expression_args([A|_R]):-
	contains_sum_expression(A),!.
contains_sum_expression_args([_A|R]):-
	contains_sum_expression_args(R),!.



% :- pred convert_to_mathematica(Eqs,MathematicaEqs,MathematicaFns,Vars).
% :- doc(convert_to_mathematica/4,
% 	"Converts a set of @var{Equations} into the list
%          of @var{MathematicaEqs}, @var{MathematicaFns} and @var{Vars}
%          that are needed to call Mathematica later.").
convert_to_mathematica(Eqs,MathematicaEqs,MathematicaFns,Vars,FreeVars) :-
	%note_message("convert_to_mathematica: ~q",[Eqs]),nl,
	% Get the maximum number of variables
	get_max_var_number(Eqs,N),
	% As all equations to Mathematica must have the same number
	% of parameters, we extend those with less
	generate_n_vars(N,Vars),
	get_fn_names(Eqs,Fns),
	sort(Fns,FnsSorted),
	convert_to_mathematica_(Eqs,Vars,Fns,MathematicaEqs0),
	generate_mathematica_fn(FnsSorted,Vars,MathematicaFns),
	replace_free_vars(N,MathematicaEqs0,MathematicaEqs,[],FreeVars),!,
	debug_message("CONVERTED TO MATH >>> ~q",[MathematicaEqs]).

replace_free_vars(N,Eqs,NEqs,PFreeVars,PNFreeVars):-
	get_var_position(Eqs,(Pos,Var)),
	get_mvar(N,Var,PFreeVars,PNFreeVars0,N1,MVar),
	update_term(Eqs,Pos,MVar,Eqs0),!,
	replace_free_vars(N1,Eqs0,NEqs,PNFreeVars0,PNFreeVars).
replace_free_vars(_,E,E,F,F).

get_mvar(N,Var,PV,PV,N,MVar):-
	PV = [(VV,MVar)|_],
	VV == Var,!.
get_mvar(N,Var,[G|PV],[G|RR],N1,MVar):-
	get_mvar(N,Var,PV,RR,N1,MVar).
get_mvar(N,Var,[],[(Var,MVar)],N1,MVar):-
	N1 is N + 1,
	atom_number(Atom,N1),
	atom_concat('var',Atom,MVar).



get_max_var_number([],0).
get_max_var_number([equation(_,Vars,_,_)|R],N) :-
	get_max_var_number(R,NR),
	length(Vars,NV),
	max(NR,NV,N).


max(N,M,N) :- N >= M, !.
max(N,M,M) :- N < M.

get_fn_names([],[]).
get_fn_names([equation(F,_,_,_)|R],[F|FR]) :-
	get_fn_names(R,FR).

convert_to_mathematica_([],_,_,[]).
convert_to_mathematica_([equation(F,Vars,Expr,_)|R],N,Fns,[Eq|EqR]) :-
	complete_until(Vars,N,AllVars),
	Lhs =.. [F|AllVars],
	CompleteEq = (Lhs == Expr),
	prune_non_vars(AllVars,N,SubstL,SubstR),
	substitute_vars(CompleteEq,SubstL,SubstR,EqNotComplete),
	complete_fn_names(EqNotComplete,Fns,N,Eq),
	convert_to_mathematica_(R,N,Fns,EqR).

complete_until([],Vars,Vars).
complete_until([E|R],[V|RVars],[V|CR]) :-
	E == '?', complete_until(R,RVars,CR).
complete_until([E|R],[_|RVars],[E|CR]) :-
	E \== '?', complete_until(R,RVars,CR).

complete_until_zeros([],[],[]).
complete_until_zeros([],[_|R],[0|ZR]) :-
	complete_until_zeros([],R,ZR).
complete_until_zeros([E|R],[_|RVars],[E|CR]) :-
	complete_until_zeros(R,RVars,CR).

generate_mathematica_fn([],_,[]).
generate_mathematica_fn([F|R],Vars,[Expr|ER]) :-
	Expr =.. [F|Vars],
	generate_mathematica_fn(R,Vars,ER).

prune_non_vars([],[],[],[]).
prune_non_vars([V|R],[_|S],T,U) :-
	nonvar(V), !,
	prune_non_vars(R,S,T,U).
prune_non_vars([V|R],[E|S],[V|T],[E|U]) :-
	var(V), !,
	prune_non_vars(R,S,T,U).

complete_fn_names((Head==Body),Fns,Vars,(CHead==CBody)) :-
	Head =.. [H|HArgs],
	complete_until(HArgs,Vars,CHArgs),
	CHead =.. [H|CHArgs],
	complete_fn_names_expr(Body,Fns,Vars,CBody).

complete_fn_names_expr(Expr,_,_,Expr) :-
%	var(Expr), !, fail.
	var(Expr), !.
complete_fn_names_expr(Expr,Fns,Vars,Eq) :-
	Expr =.. [F|Args],
	complete_fn_names_each(Args,Fns,Vars,CArgs),
	( member(F,Fns)
	-> complete_until_zeros(CArgs,Vars,CCArgs)
	; CCArgs = CArgs
	),
	Eq =.. [F|CCArgs].

complete_fn_names_each([],_,_,[]).
complete_fn_names_each([E|R],Fns,Vars,[CE|CR]) :-
	complete_fn_names_expr(E,Fns,Vars,CE),
	complete_fn_names_each(R,Fns,Vars,CR).


% :- pred generate_mathematica_text(MEqs,MFns,MVars,MText).
% :- doc(generate_mathematica_text/4,
% 	"Generates the Mathematica code to solve the equations
%          @var{MEqs} on functions @var{MFns} and variables
%          @var{MVars}. These elements can be generated from the
%          equations with @tt{convert_to_mathematica}.").
generate_mathematica_text(MEqs,MFns,MVars,MText) :-
	generate_eqs_text(MEqs,EqsText),
	generate_rest_text(MFns,FnsText),
	generate_rest_text(MVars,VarsText),
	FormatText = "FullForm[FullSimplify[RSolve[~s,~s,~s], Element[~s,Integers] ]]",
	format_to_string(FormatText,[EqsText,FnsText,VarsText,VarsText],MText).

generate_rest_text(Eqs,Text) :-
	generate_expr_text_each(Eqs,EqsText,false),
	join_by(EqsText,",",QText),
	append("{",QText,QParen),
	append(QParen,"}",Text).

generate_eqs_text(Eqs,Text) :-
	generate_eq_text_each(Eqs,EqsText),
	join_by(EqsText,",",QText),
	format_to_string("{~s}",[QText],Text).

generate_eq_text_each([],[]).
generate_eq_text_each([(Lhs==Rhs)|Rs],[EqText|RsText]) :-
	generate_expr_text(Lhs,LText,false),
	generate_expr_text(Rhs,RText,true),
	format_to_string("~s==~s",[LText,RText],EqText),
	generate_eq_text_each(Rs,RsText).

generate_gret_text(Eqs,Text) :-
	generate_gret_text_each(Eqs,EqsText),
	join_by(EqsText,"&&",Text).

generate_gret_text_each([],[]).
generate_gret_text_each([X|Rs],[EqText|RsText]) :-

	format_to_string("~q>=0",[X],EqText),
	generate_gret_text_each(Rs,RsText).



:- pred simplify_in_mathematica(Expression,Solution)
	: term * var
       => term * term.
:- doc(simplify_in_mathematica/2,
	"Tries to simplify a @var{Expression} by calling Mathematica.").
simplify_in_mathematica(Expr,Solution) :-
	catch(
		 (	
		     % Simplify in Mathematica
		     term_variables(Expr,FinalVars),
%		     note_message("Expr>>> ~q",[Expr]),
		     length(FinalVars,FinalVarsLength),
		     generate_n_vars(FinalVarsLength,VarsAsNumbers),
		     substitute_vars(Expr,FinalVars,VarsAsNumbers,MExpr),
		     generate_simplify_mathematica_text(MExpr,VarsAsNumbers,MText),
		     ( call_mathematica(MText,FinalVars,no,SolutionExpr)
		     -> Solution = SolutionExpr
		     ; Solution = Expr
		     )
		 ),
		 Ex,
		 (
		     (mathematica_exception(Ex) -> 
		      throw(backend_exception(simplify,mathematica,[Expr],Ex));throw(Ex))
		 )
	     ).

generate_simplify_mathematica_text(Expr,Vars,MText) :-
	generate_expr_text(Expr,ExprText,true),
	generate_integer_assumptions(Vars,AssText),
	FormatText = "FullForm[FullSimplify[~s,{~s}]]",
	format_to_string(FormatText,[ExprText,AssText],MText).

generate_integer_assumptions([],"") :- !.
generate_integer_assumptions([V],Ass) :-
	!, format_to_string("Element[~q,Integers]",[V],Ass).
generate_integer_assumptions([V|R],Ass) :-
	generate_integer_assumptions(R,RAss),
	format_to_string("Element[~q,Integers],~s",[V,RAss],Ass).

:- pred expand_in_mathematica(Expression,Solution)
	: term * var
       => term * term.
:- doc(expand_in_mathematica/2,
	"Expands @var{Expression} into sums and powers.").
expand_in_mathematica(Expr,Solution) :-
	catch(
		 (
		     % Simplify in Mathematica
%		     note_message("Expr expand>>> ~q",[Expr]),
		     term_variables(Expr,FinalVars),
		     length(FinalVars,FinalVarsLength),
		     generate_n_vars(FinalVarsLength,VarsAsNumbers),
		     substitute_vars(Expr,FinalVars,VarsAsNumbers,MExpr),
		     generate_expand_mathematica_text(MExpr,MText),
		     ( call_mathematica(MText,FinalVars,no,SolutionExpr)
		     -> Solution = SolutionExpr
		     ; Solution = Expr
		     )
		 ),
		 Ex,
		 (
		     (mathematica_exception(Ex) -> 
		      throw(backend_exception(expand,mathematica,[Expr],Ex));throw(Ex))
		 )
	     ).
generate_expand_mathematica_text(Expr,MText) :-
	generate_expr_text(Expr,ExprText,true),
	FormatText = "FullForm[Expand[~s]]",
	format_to_string(FormatText,[ExprText],MText).

expand_functions_in_mathematica(Expr,Solution) :-
%	note_message("Expr>>> ~q",[Expr]),
	term_variables(Expr,FinalVars),
	length(FinalVars,FinalVarsLength),
	generate_n_vars(FinalVarsLength,VarsAsNumbers),
	substitute_vars(Expr,FinalVars,VarsAsNumbers,MExpr),
	generate_expand_functions_mathematica_text(MExpr,MText),
	( call_mathematica(MText,FinalVars,no,SolutionExpr)
	-> Solution = SolutionExpr
	 ; Solution = Expr
	).

generate_expand_functions_mathematica_text(Expr,MText) :-
	generate_expr_text(Expr,ExprText,true),
	FormatText = "FullForm[FullSimplify[FunctionExpand[~s]]]",
	format_to_string(FormatText,[ExprText],MText).


greater_than_mathematica(X,Y):-
	catch(
		 simplify_in_mathematica('Greater'(X,Y),'True'),
		 backend_exception(simplify,Solver,Args,Ex),
		 throw(backend_exception(greater_than,Solver,Args,Ex))
	     ).


equal_than_mathematica(X,Y):-
	catch(
		 simplify_in_mathematica('Equal'(X,Y),'True'),
		 backend_exception(simplify,Solver,Args,Ex),
		 throw(backend_exception(greater_than,Solver,Args,Ex))
	     ).



:- pred greater_or_equal_than_mathematica(X,Y)
	: mathematica_expression * mathematica_expression.

greater_or_equal_than_mathematica(X,Y):-
	catch(
		 (	
		     % Simplify in Mathematica
%		     note_message("GreaterEqual>>> ~q",[X]),
		     term_variables((X,Y),FinalVars),
		     length(FinalVars,FinalVarsLength),
		     generate_n_vars(FinalVarsLength,VarsAsNumbers),
		     substitute_vars(X,FinalVars,VarsAsNumbers,XExpr),
		     substitute_vars(Y,FinalVars,VarsAsNumbers,YExpr),
		     generate_gret_mathematica_text(XExpr,YExpr,VarsAsNumbers,MText),!,
		     call_mathematica(MText,FinalVars,no,Sol),Sol='True'
		     
		 ),
		 Ex,
		 (
		     (mathematica_exception(Ex) -> 
		      throw(backend_exception(greater_or_equal_than,mathematica,[X,Y],Ex));throw(Ex))
		 )
	     ).

generate_gret_mathematica_text(X,Y,VarsAsNumbers,MText) :-
	generate_expr_text(X,XText,true),
	generate_expr_text(Y,YText,true),
	generate_rest_text(VarsAsNumbers,NVars),
	generate_gret_text(VarsAsNumbers,PVars),
	FormatText = "Reduce[ForAll[~s,~s,~s >= ~s],Integers]",
	format_to_string(FormatText,[NVars,PVars,XText,YText],MText).




:- pred solve_in_mathematica(Equation,Var,Solution).
:- doc(solve_in_mathematica/2,
	"Solves the @var{Equation} for the @var{Var} into @var{Solution}.").

%Special trivial cases
solve_in_mathematica(A=B,Y,A):-
	Y==B,!.
solve_in_mathematica(A=B,Y,B):-
	Y==A,!.


solve_in_mathematica(Equation,V,Solution) :-
	catch(
		 (
		     % Generate equation without variables
%		     note_message("EQUATION>>> ~q",[Equation]),
		     term_variables(Equation,FinalVars),
		     length(FinalVars,FinalVarsLength),
		     generate_n_vars(FinalVarsLength,VarsAsNumbers),
		     substitute_vars(Equation,FinalVars,VarsAsNumbers,MExpr),
		     find_variable_number(FinalVars,VarsAsNumbers,V,MVar),
		     generate_solve_mathematica_text(MExpr,MVar,MText),
		     ( call_mathematica(MText,FinalVars,yes,SolutionExpr)
		     -> Solution = SolutionExpr
		     ; Solution = Equation
		     )
		 ),
		 Ex,
		 (
		     (mathematica_exception(Ex) -> 
		     throw(backend_exception(solve,mathematica,[Equation,V],Ex));throw(Ex))
		 )
	     ).

:- export(find_variable_number/4).
find_variable_number([V|_],[N|_],W,N) :- V == W, !.
find_variable_number([_|R],[_|S],W,N) :- find_variable_number(R,S,W,N).

generate_solve_mathematica_text(Lhs=Rhs,Var,MText) :-
	generate_expr_text(Lhs,LhsText,false),
	generate_expr_text(Rhs,RhsText,false),
	generate_expr_text(Var,VarText,false),
	FormatText = "FullForm[ Solve[~s==~s,~s] ]",
	format_to_string(FormatText,[LhsText,RhsText,VarText],MText).

generate_n_vars(0,[]) :- !.
generate_n_vars(N,Vars) :-
	N > 0, !,
	N1 is N - 1,
	generate_n_vars(N1,Vars1),
	atom_number(Atom,N),
	atom_concat('var',Atom,ThisVar),
	append(Vars1,[ThisVar],Vars).

generate_expr_text(Expr,Text,_) :-
	num(Expr), !,
	number_codes(Expr,Text).
generate_expr_text(Expr,Text,Paren) :-
	Expr =.. [F|Args],
	atom_codes(F,FText),
	generate_expr_text_each(Args,ArgsText,Paren),
	( ArgsText = []
	-> QText = FText
	 ; ( not_bracketed(F)
	   -> ( ArgsText = [OneArg]
	      -> format_to_string("~s~s",[FText,OneArg],QText) % Unary
	       ; ArgsText = [OneArg,TwoArg],
		 format_to_string("~s~s~s",[OneArg,FText,TwoArg],QText) % Binary
	      )
	   ; join_by(ArgsText,",",Inner),
	     format_to_string("~s[~s]",[FText,Inner],QText)
	   )
	),
	( Paren = true
	-> format_to_string("(~s)",[QText],Text)
	 ; Text = QText
	).

generate_expr_text_each([],[],_).
generate_expr_text_each([E|R],[EText|RText],Paren) :-
	generate_expr_text(E,EText,Paren),
	generate_expr_text_each(R,RText,Paren).

not_bracketed(X) :- member(X,['+','-','*','/','==']).

join_by([E],_,E).
join_by([E|R],Sep,Text) :-
	join_by(R,Sep,RText),
	append(E,Sep,ESep),
	append(ESep,RText,Text).






% BEGIN Parser for Mathematica expressions
elementfv(_,_,N)    --> number(N), !.
elementfv(_,_,N)    --> negative_number(N), !.
elementfv(Vars,FreeVars,V) --> variablefv(Vars,FreeVars,V), !.
elementfv(Vars,FreeVars,I) -->
	identifier(F),
	spaces,
	element_listfv(Vars,FreeVars,Args),
	{
	    process_element(F,Args,I)
	}.


element_listfv(_,_,[]) -->
	lbrace,
	spaces,
	rbrace,
	!.
element_listfv(Vars,FV,L)  -->
	lbrace,
	!,
	element_list_fv(Vars,FV,L).
element_listfv(_,_,[]) --> [].

element_list_fv(Vars,FV,[E|R]) -->
	elementfv(Vars,FV,E),
	spaces,
	comma,
	spaces,
	element_list_fv(Vars,FV,R).
element_list_fv(Vars,FV,[E])   -->
	elementfv(Vars,FV,E),
	spaces,
	rbrace.





element(_,N)    --> number(N), !.
element(_,N)    --> negative_number(N), !.
element(Vars,V) --> variable(Vars,V), !.
element(Vars,I) -->
	identifier(F),
	spaces,
	element_list(Vars,Args),
	{
	    process_element(F,Args,I)
	}.

element_list(_,[]) -->
	lbrace,
	spaces,
	rbrace,
	!.
element_list(Vars,L)  -->
	lbrace,
	!,
	element_list_(Vars,L).
element_list(_,[]) --> [].

element_list_(Vars,[E|R]) -->
	element(Vars,E),
	spaces,
	comma,
	spaces,
	element_list_(Vars,R).
element_list_(Vars,[E])   -->
	element(Vars,E),
	spaces,
	rbrace.

variable(Vars,Var) -->
	"var",!,
	number(N),
	{ nth(N,Vars,Var) }.

variablefv(Vars,FreeVars,Var) -->
	"var",!,
	number(N),
	{ nth(N,Vars,Var);
	    atom_number(Atom,N),
	    atom_concat('var',Atom,MVar),
	    member((Var,MVar),FreeVars)
	}.


identifier('+') --> "Plus", !.
identifier('*') --> "Times", !.
identifier('/') --> "Rational", !.

identifier(A) -->
	chars_no_space(X),
	{ atom_codes(A,X) }.

process_element(_,Args,'$Failed') :-
	varmember('$Failed',Args), !.
process_element('C',_,'$Failed') :- !.
%% process_element('*',[N,Arg],I) :-
%% 	N == -1, !,
%% 	( ground(Arg), num(Arg)
%% 	-> I is -Arg
%% 	 ; I = -Arg
%% 	).
%% process_element('*',[N,Arg],I) :-
%% 	ground(N), num(N), N < 0, !,
%% 	MinN is -N,
%% 	process_element('*',[MinN,Arg],IIn),
%% 	( ground(IIn), num(IIn)
%% 	-> I is -IIn
%% 	 ; I = -IIn
%% 	).
process_element('*',Args,I) :-
	!, generate_two_by_two('*',Args,I).
process_element('+',Args,I) :-
	!, generate_sum(Args,I).
process_element('Power',[A,B],'Exp'(B)) :-
	A == 'E', !.
%process_element('Power',[A,B],A ^ B) :- !.
process_element('Gamma',[A+B],'Factorial'(Var)) :-
	A == 1, !, Var = B.
process_element('Gamma',[A+B],'Factorial'(Var)) :-
	B == 1, !, Var = A.
process_element('Pochhammer',[1,Var],'Factorial'(Var)) :- !.


process_element(F,Args,I) :-
	I =.. [F|Args].

generate_two_by_two(F,[A,B],Solution) :-
	!, Solution =.. [F,A,B].
generate_two_by_two(F,[C|R],Solution) :-
	generate_two_by_two(F,R,Inner),
	Solution =.. [F,C,Inner].

generate_sum([A],A).
generate_sum([E|R],B) :-
	generate_sum(R,RSum),
	( ground(E), num(E), E < 0
	-> Em is -E, Expr = RSum - Em
	 ; Expr = RSum + E
	),
	( ground(Expr), arithexpression(Expr)
	-> B is Expr
	 ; B = Expr
	).


spaces --> space, spaces.
spaces --> [].

negative_number(N) -->
	"-",
	number(M),
	{ N is -M }.

number(N) -->
	number_parts(X),
	{ number_codes(N,X) }.

number_parts(N) -->
	digits(A),
	dot,
	digits(B),
	{ append(A,[0'.|B],N) }.
number_parts(A) -->
	digits(A).
number_parts([0'0,0'.|B]) -->
	dot,
	digits(B).
	
digits([D|R]) --> digit(D), digits(R).
digits([])    --> [].

chars_no_space([H|T]) --> char_no_space(H), chars_no_space(T).
chars_no_space([])    --> [].

char_no_space(X) --> [X], 
	{\+ comma([X],[]), \+ space([X],[]),
         \+ lbrace([X],[]), \+ rbrace([X],[])}.

dot      --> [0'.].
comma    --> [0',].
lbrace   --> [0'[].
rbrace   --> [0']].
digit(X) --> [X], {X > 47, X<58}.
space    --> [X], {X=<32}. % covers EOF, ^I, ^G, ^J, ^K, ^L, ^M, ^Z, ESC, ^_
%eol      --> [0'
%].
% END Parser for Mathematica expressions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%END CODE ORIGINALLY IN eq_solver_mathematica MODULE%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%TRANSLATION TO/FROM ALGEBRAIC EXPRESSION TO/FROM MATHEMATICA SYNTAX%%%%%%%
% TODO: Consider all basic Mathematica functions (like 'Factorial', 'Power', etc.).
% TODO: Define recurrence_relation_mathematica type



:- use_module(domain(resources/eq_solver/eq_solver_types)).
:- use_module(domain(resources/plai2infer),[replace_term/3]).






:- pred translate_from_algebraic(Exp,MExp):(algebraic_expression(Exp),var(MExp)) => (mathematica_expression(MExp)).
:- pred translate_from_algebraic(Exp,MExp):(recurrence_relation(Exp),var(MExp)) => (list(MExp)).
translate_from_algebraic(Exp,MExp):-
	\+ \+ algebraic_expression(Exp),!,
	replace_term(Exp,[('**','Power'),(sqrt,'Sqrt'),(exp,'Exp'),(fact,'Factorial'),(lucasl,'LucasL'),(fibo,'Fibonacci'),(log,'Log'),(inf,'Infinity'),(max,'Max'),(min,'Min')],MExp0),
	unwrap_fun(MExp0,MExp).
translate_from_algebraic(Rec,MRec):-
	\+ \+ recurrence_relation(Rec),
	equations2mathequations(Rec,MRec).

% unwrap_fun(Exp,Exp):-
% 	var(Exp),!.
% unwrap_fun(fun(Name,Args),Res):-
% 	replace_fun_args(Args,NArgs),
% 	Res =.. [Name|NArgs],!.
% unwrap_fun(Exp,Res):-
% 	Exp =.. [ExpHead|Args],
% 	replace_fun_args(Args,NArgs),
% 	Res =.. [ExpHead|NArgs].

% replace_fun_args([],[]).
% replace_fun_args([Arg|Args],[NArg|NArgs]):-
% 	unwrap_fun(Arg,NArg),
% 	replace_fun_args(Args,NArgs).

:- pred translate_to_algebraic(MExp,Exp):(mathematica_expression(MExp),var(Exp)) => (algebraic_expression(MExp)).
:- pred translate_to_algebraic(MExp,Exp):(list(MExp),var(Exp)) => (recurrence_relation(MExp)).
translate_to_algebraic(MRec,ARec):-
	%\+ \+ list(MRec),!,
	\+ var(MRec),
	list(MRec),!,
	mathequations2equations(MRec,ARec).
translate_to_algebraic(MExp,Exp):-
	\+ \+ mathematica_expression(MExp),!,
	replace_term(MExp,
	[('Power','**'),
	 ('Sqrt',sqrt),
	 ('Exp',exp), 
	 ('Factorial',fact),
	 ('LucasL',lucasl),
	 ('Fibonacci',fibo),
	 ('Log',log),
	 ('Infinity',inf),
	 ('DirectedInfinity'(1),inf),
	 ('Max',max),
	 ('Min',min)
	],Exp0),
	max_min_arity_2(Exp0,Exp1),
	wrap_math_fun(Exp1,Exp),!.

equations2mathequations([],[]).
equations2mathequations([equation(F,Args,Exp,Cons)|R],[equation(F,Args,NExp,Cons)|RR]):-
	translate_from_algebraic(Exp,NExp),
	equations2mathequations(R,RR).

mathequations2equations([],[]).
mathequations2equations([equation(F,Args,Exp,Cons)|R],[equation(F,Args,NExp,Cons)|RR]):-
	translate_to_algebraic(Exp,NExp),
	mathequations2equations(R,RR).

%:- export(max_min_arity_2/2).
:- pred max_min_arity_2(QExp,Exp): (term(QExp),var(Exp)) => (algebraic_expression(Exp))
# "Returns an expression by transforming the subterms max/n and min/n from @var{QExp} with n>2, in a nested max(x1,max(...)) and min(y1,min(...)), with n-1 max/2 and min/2 nested terms.".
max_min_arity_2(V,V):-
	var(V),!.
max_min_arity_2(V,VS):-
	V =.. [MM|Args],
	(MM == max; MM == min),
	 length(Args,N),
	 N > 2,!,
	 Args = [X,Y|T],
	 Sub =.. [MM|[X,Y]],
	 NewV =.. [MM|[Sub|T]],
	 max_min_arity_2(NewV,VS),!.
max_min_arity_2(V,V).

