:- module(solver_utils_translations,
	[translate/4],
	 [assertions,
	  nativeprops,
	  regtypes]).

:- use_module(domain(resources/term_utilities), [replace_subterm_cond/4]).
:- use_module(domain(resources/eq_solver/eq_solver_types),
	[
	    algebraic_expression/1,
	    backend_solver/1
	]).

% TODO: Fix translation to/from languages without explicitly translate
% to algebraic form

:- pred translate(From,To,Term,Result)
	:(backend_solver(From),backend_solver(To),term(Term),var(Result))
 =>(backend_solver(From),backend_solver(To),term(Term),term(Result)) #
 "Translates the expression @var{Term} from the syntax of the backend
 solver @var{From} to the syntax of the backend solver @var{To}, if
 possible.".

translate(I,O,From, To) :-
	nonvar(From),
	From = prod(Var,Ini,Fin,Exp),
	O == mathematica,
	I == algebraic,
	To = 'Prod'(Exp,Var,Ini,Fin),!.
translate(I,O,From, To) :-
	nonvar(From),
	From = sum(Var,Ini,Fin,Exp),
	O == mathematica,
	I == algebraic,
	To = 'Sum'(Exp,Var,Ini,Fin),!.
translate(I,O,Input,Output):-
	O == algebraic,
	replace_subterm_cond(Input,to_be_translated(I,algebraic),translate_functor(I,algebraic),Output0),
	wrap_functions(Output0,Output),!.
translate(From,To,Input,Output):-
	replace_subterm_cond(Input,to_be_translated(From,To),translate_functor(From,To),Output),!.
translate(X,Y,Input,Output):-
	translate(X,Z,Input,IntOutput),
	translate(Z,Y,IntOutput,Output).
	
wrap_functions(A,A):-
	var(A),!.
wrap_functions(A,A):-
	num(A),!.
wrap_functions(A,A):-
	A =.. [fun|_],!.
wrap_functions(inf,inf):-!.
wrap_functions(A,fun(A,[])):-
	A =.. [_Op],!.
wrap_functions(Com, Res):-
	Com =.. [Op|Args],
	wrap_functions_args(Args,NArgs),
	Res =.. [Op|NArgs],!.
wrap_functions_args([],[]).
wrap_functions_args([Arg|T],[NArg|R]):-
	wrap_functions(Arg,NArg),!,
	wrap_functions_args(T,R).

	

:- pred translate_functor(Pos,From,To,Term,Result)
	:(list(Pos,num),backend_solver(From),backend_solver(To),term(Term),var(Result))
 =>(list(Pos,num),backend_solver(From),backend_solver(To),term(Term),term(Result)).
translate_functor(_,mathematica,_,A,B):-
	var(A),var(B),!,fail.
translate_functor(_,algebraic,mathematica,A,B):-
	translate_functor(_,mathematica,algebraic,B,A).
translate_functor(_,mathematica,algebraic,'Power'(A,B),A**B):-!.
translate_functor(_,mathematica,algebraic,'Gamma'(A),B):- var(B),B = fact(A-1),!.
translate_functor(_,mathematica,algebraic,'Prod'(Exp,Var,Ini,Fin),B):- var(B),B = prod(Var,Ini,Fin,Exp),!.
translate_functor(_,mathematica,algebraic,'Sum'(Exp,Var,Ini,Fin),B):- var(B),B = sum(Var,Ini,Fin,Exp),!.
translate_functor(_,mathematica,algebraic,'Sqrt'(A),sqrt(A)):-!.
translate_functor(_,mathematica,algebraic,'Exp'(A),exp(A)):-!.
translate_functor(_,mathematica,algebraic,'Factorial'(A),fact(A)):-!.
translate_functor(_,mathematica,algebraic,'LucasL'(A),lucasl(A)):-!.
translate_functor(_,mathematica,algebraic,'Fibonacci'(A),fibo(A)):-!.
translate_functor(_,mathematica,algebraic,'Log'(A),log(A)):-!.
translate_functor(_,mathematica,algebraic,'Log'(A,B),log(A,B)):-!.
translate_functor(_,mathematica,algebraic,'Infinity',inf):-!.
translate_functor(_,mathematica,algebraic,'DirectedInfinity'(1),inf):-!.
translate_functor(_,mathematica,algebraic,'Max'(A,B),max(A,B)):-!.
translate_functor(_,mathematica,algebraic,'Min'(A,B),min(A,B)):-!.
translate_functor(_,mathematica,algebraic,MaxMinN,MaxMin2):-
	((Op = 'Max',OpA=max);(Op='Min',OpA=min)),
	functor(MaxMinN,Op,N),
	N > 2,
	MaxMinN =.. [Op|Args],
	generate_op_pairwise(OpA,Args,MaxMin2),!.
translate_functor(_,mathematica,algebraic,F,WF):-
	nonvar(F),
	F =.. [Op|_Args],
	functor(F,Op,N),
	functor(FE,Op,N),
	\+ algebraic_expression(FE),
	Op \= fun,
	N > 0,
	wrap_user_defined_function(F,WF),!.
translate_functor(_,mathematica,algebraic,F,WF):-
	var(F),nonvar(WF),
	wrap_user_defined_function(F,WF),!.

translate_functor(_,builtin,_,A,B):-
	var(A),var(B),!,fail.
translate_functor(_,algebraic,builtin,A,B):-
	translate_functor(_,builtin,algebraic,B,A).
translate_functor(_,builtin,algebraic,exp(A,B),A**B):-!.
%translate_functor(_,builtin,algebraic,log(A,B),log(A,B)):-!.
translate_functor(_,builtin,algebraic,F,WF):-
	nonvar(F),
	F =.. [Op|_Args],
	functor(F,Op,N),
	functor(FE,Op,N),
	\+ algebraic_expression(FE),
	Op \= fun,
	N > 0,
	wrap_user_defined_function(F,WF),!.
translate_functor(_,builtin,algebraic,F,WF):-
	var(F),nonvar(WF),
	wrap_user_defined_function(F,WF),!.

wrap_user_defined_function(F,fun(Op,Args)):-
	nonvar(F),
	F =.. [Op|Args].
wrap_user_defined_function(F,fun(Op,Args)):-
	var(F), 
	F =.. [Op|Args].

generate_op_pairwise(Op,[X,Y],OpXY):-
	OpXY =.. [Op|[X,Y]].
generate_op_pairwise(Op,[X,Y|T],OpXY):-
	OpXY0 =.. [Op|[X,Y]],
	generate_op_pairwise(Op,[OpXY0|T],OpXY).

to_be_translated0(Term,_From,_To):-
	nonvar(Term),
	\+ num(Term),
	Term \== inf,
	Term =.. [_OP],!.

to_be_translated(Term,From,To):-
	translate_functor(_,From,To,Term,_).
to_be_translated(Term,From,To):-
	translate_functor(_,From,Z,Term,R),
	translate_functor(_,Z,To,R,_).

