:- module(recurrence_processing,[
	assume/3,
	gather_and_solve_eqs/1,
	recompute/0,
	solve_eqs/1,
	show_eqs/0,
	show_unsolved_eqs/0,
	get_statistics/4,
	clean_eqs/0,
	closed_form/1, %MK: Temporary for using in eq_solver
	slice_args/3,  %MK: Temporary for using in eq_solver
	remove_elements/3, %MK: Temporary for using in eq_solver
	normalize_eqs_/2, %MK: Temporary for using in eq_solver
	unify_vars/1, %MK: Temporary for using in eq_solver
	non_recursive_eq/1, %MK: Temporary for using in eq_solver
	denormalize_tests/2, %MK: Temporary for using in eq_solver
	write_results/1
	], [assertions,regtypes,basicmodes,datafacts]).

:- doc(module,
	"Recurrence Equation Solver for res_plai and res_plai_stprf").
:- doc(author,"Maximiliano Klemen").

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- use_module(ciaopp(plai/plai_db)).
:- use_module(term_utilities,
	[update_term/4,
	 get_subterm_cond/3,
	 term_at_position/3
	]).
:- use_module(library(aggregates),[findall/3, setof/3]).
:- use_module(library(lists),[member/2, append/3, length/2,nth/3,delete/3]).
:- use_module(domain(s_eqs), [peel/4]).
:- use_module(library(terms_vars)).
:- use_module(.(eq_solver/solver/solver)).
:- use_module(ciaopp(preprocess_flags)).
:- use_module(library(sort)).
:- use_module(library(streams)).
:- use_module(library(write)).
:- use_module(library(hiordlib),[maplist/2,foldl/4]).
:- use_module(engine(runtime_control), [statistics/2]).
:- use_module(library(terms_io), [terms_to_file/2,file_to_terms/2]).
% TODO: This should be in the solver module.
:- use_module(.(eq_solver/solver/solver_special_eqs),
	[solve_special_equation/3]).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TODO: Constraint propagation                                           %
% TODO: Move min-max expression solver to the solver module              %
% TODO: Move free variables handling to solver.                          %
% TODO: Argument minimization and free variable identification need      %
%       revision                                                         %
%       f(X,Y) = f(X-Y,Y) + 1, f(0,_) = 1                                %
%       --> Y can be a free variable, but not being deleted.             %
%       f(X,Y,Z) = f(X+Z,Y-1,Z) + 1, f(_,0,_) = 1 --> X and Z            %
%       can be both removed.                                             %
%       (coupled example)                                                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- doc(bug, "The output of closed-forms to plai_db contains free
variables when the solution includes variables that were not
arguments of the equations.").



% :-export(solution/3).
% :-export(unsolved/1).
% :-export(eqs/4).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- data eqs/4.
:- data unsolved/1.
:- data solution/3.
:- data assumption/3.
:- data runtime/1.
% mapping_fn_names/1 stores a mapping from equation names to
% pretty-print names.
:- data mapping_fn_names/1.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Load the module in debug mode, and execute                         %%%
%%% spy debug_function_set, in order to trace the solving of           %%%
%%% the equations in the list bellow.                                  %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fdebugset([x]).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%--------------------------------------------------------------------------
%Collecting and writing equations from/to plai_db or from/to file
clean_eqs:-
	retractall_fact(eqs(_,_,_,_)),
	retractall_fact(runtime(_)),
	retractall_fact(unsolved(_)),
	retractall_fact(top_scc(_)),
%	retractall_fact(mapping_fn_names(_)),
%	retractall_fact(counter_function_name(_)),
	retractall_fact(solution(_,_,_)),
	retractall_fact(assumption(_,_,_)).
%	asserta_fact(mapping_fn_names([])),
%	asserta_fact(counter_function_name(0)).


:- export(load_eqs_from_file/1).
:- export(output_eqs_to_file/1).
load_eqs_from_file(File):-
	clean_eqs,
	file_to_terms(File,Eqs),
	load_eqs_from_terms(Eqs).

load_eqs_from_terms([]).
load_eqs_from_terms([Eq|T]):-
	assertz_fact(Eq),!,
	load_eqs_from_terms(T).

output_eqs_to_file(File):-
	findall(eqs(A,B,C,D),eqs(A,B,C,D),LEqs),
	findall(solution(A,B,C),solution(A,B,C),LSol),
	findall(unsolved(A),unsolved(A),LUns),
	append(LEqs,LSol,LOut0),
	append(LOut0,LUns,LOut),
	terms_to_file(LOut,File).


gather_eqs(An):-
	gather_eqs_(An).

gather_eqs_(res_plai):-
	complete(_,res_plai,_X3,_X4,[X5],_X6,_X7),
	X5 = res_plai(_,R),
	R=res_abs(_,_,_,Res,_,_,_,_,_,_,_,_,_,_),
	Res = resolution(_,_,_,Eqs),
	save_eqs(Eqs,lower),
	fail.
gather_eqs_(res_plai_stprf):-
	complete(_,res_plai_stprf,_X3,_X4,[X5],_X6,_X7),
	X5 = res_plai_stprf(_,R),
	R=res_abs_stprf(_,_,_,_,Res,_,_,_,_,_,_,_,_,_,_),
	Res = resolution(_,_,_,Eqs),
	save_eqs(Eqs,lower),
	fail.
gather_eqs_(_).


make_eqs_template(Eqs,TEqs,N):-
	Eqs = [equation(N,_,_,_)|_],
	findall(Pos, get_subterm_cond(Eqs,'=='(N),(Pos,_)), LPos),
	update_eqs(Eqs, LPos, _NVar,TEqs).

save_eqs([],_).
save_eqs([(_,unsolved_equations(_,Eq0))|T], B):-
	Eq0=Eq,
	%beautify_equations(Eq0,Eq),
	make_eqs_template(Eq, TEq,Name),
	save_template(TEq,Name, B),!,
	opposite_bound(B,B0),
	save_eqs(T, B0).
save_eqs([(_,closed_form(_))|T], B):-!,
	opposite_bound(B,B0),
	save_eqs(T,B0).

save_template(T,N,B):-
	current_fact(eqs(B,L,TS,Del)),
	\+ \+ (T = TS),!,
	retract_fact(eqs(B,L,TS,Del)),
	asserta_fact(eqs(B,[N|L],TS,Del)).
save_template(T,N,B):-
	T = [equation(N,_,_,_)|_],
	asserta_fact(eqs(B,[N],T,[])).

opposite_bound(lower, upper).
opposite_bound(upper,lower).


write_results(res_plai):-
	retract_fact(complete(GG,res_plai,X3,X4,[res_plai(Ty,res_abs(A1,A2,A3,Res,A4,A5,A6,A7,A8,A9,A10,A11,A13,A14))],X6,X7)),
	Res = resolution(B1,B2,B3,Eqs),
	write_results_(Eqs,NEqs),
	assertz_fact(complete(GG,res_plai,X3,X4,[res_plai(Ty,res_abs(A1,A2,A3,resolution(B1,B2,B3,NEqs),A4,A5,A6,A7,A8,A9,A10,A11,A13,A14))],X6,X7)),
	fail.
write_results(res_plai_stprf):-
	retract_fact(complete(GG,res_plai_stprf,X3,X4,[res_plai_stprf(Ty,res_abs_stprf(A1,A2,A3,C1,Res,A4,A5,A6,A7,A8,A9,A10,A11,A13,A14))],X6,X7)),
	Res = resolution(B1,B2,B3,Eqs),
	write_results_(Eqs,NEqs),
	assertz_fact(complete(GG,res_plai_stprf,X3,X4,[res_plai_stprf(Ty,res_abs_stprf(A1,A2,A3,C1,resolution(B1,B2,B3,NEqs),A4,A5,A6,A7,A8,A9,A10,A11,A13,A14))],X6,X7)),
	fail.
write_results(_).

write_results_([],[]).
write_results_([(A,unsolved_equations(B,[equation(FName,_,_,_)|_]))|T],[(A,closed_form(CSavedEq))|TT]):-
	current_pp_flag(output_resources,functions),!,
	eqs(_,Names,SEqs,Del), member(FName,Names),!, %mapping_fn_names(Map), member((FName,As),Map),member(As,Names),!,
	remove_elements(Del,B,NB),
	SEqs = [equation(F,_,_,_)|_],
	solution(F,Vars,SavedEq),
	copy_term((Vars,SavedEq), (CVars,CSavedEq)),
	CVars = NB,
	write_results_(T, TT).
write_results_([(A,unsolved_equations(B,[equation(FName,_,_,_)|_]))|T],[(A,unsolved_equations(NB,SEqs))|TT]):-
%	current_pp_flag(output_resources,equations),!,
	eqs(_,Names,SEqs,Del), member(FName,Names),!, %mapping_fn_names(Map), member((FName,As),Map),member(As,Names),!,
	remove_elements(Del,B,NB),
	SEqs = [equation(_F,NB,_,_)|_],
	write_results_(T, TT).
% write_results_([(A,unsolved_equations(B,[equation(FName,_,_,_)|_]))|T],[(A,closed_form(CSavedEq))|TT]):-
% 	current_pp_flag(output_resources,functions),!,
% 	eqs(_,Names,SEqs,Del), mapping_fn_names(Map), member((FName,As),Map),member(As,Names),!,
% 	remove_elements(Del,B,NB),
% 	SEqs = [equation(F,_,_,_)|_],
% 	solution(F,Vars,SavedEq),
% 	copy_term((Vars,SavedEq), (CVars,CSavedEq)),
% 	CVars = NB,
% 	write_results_(T, TT).
% write_results_([(A,unsolved_equations(B,[equation(FName,_,_,_)|_]))|T],[(A,unsolved_equations(NB,SEqs))|TT]):-
% 	current_pp_flag(output_resources,equations),!,
% 	eqs(_,Names,SEqs,Del), mapping_fn_names(Map), member((FName,As),Map),member(As,Names),!,
% 	remove_elements(Del,B,NB),
% 	SEqs = [equation(_F,NB,_,_)|_],
% 	write_results_(T, TT).
write_results_([(A,closed_form(C))|T], [(A,closed_form(C))|B]):-!,
	write_results_(T,B).

%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%Debugging information
show_unsolved_eqs :-
	(eqs(Bound,_,[equation(F,_,_,_)|_],_),
	 (Bound=upper,solution(F,_,inf))
	;
	    (Bound=lower,solution(F,_,0))
	),!,
	write('=======================( Unsolved Equations )======================='),nl,
	show_unsolved_eqs_,
	write('===================================================================='),nl.
show_unsolved_eqs.

show_unsolved_eqs_:-
	solution(F,Vs,Sol),
	eqs(Bound,_N,Eqs,_),Eqs=[equation(F,Vs,_,_)|_],
	((Bound=lower,Sol==0);Sol==inf),
	write(F),
	write(':'),
	write(Bound),nl,
	numbervars(Eqs,0,_),
	write(Eqs),nl,
	write('Closed form: '),write(Sol),nl,
	write('--------------------------------------------------------------------'),nl,
	fail.
show_unsolved_eqs_.

get_statistics(TotalEqs,TotalReduced,TotalReducedSolved,TimeSolver):-
	current_fact(runtime(TimeSolver)),
	findall(N, (eqs(_,L,_,_),length(L,N)),LN),
	foldl((''(X,Y,Z):- Z is X+Y),LN,0, TotalEqs),
	length(LN,TotalReduced),
	findall(S,(solution(F,_,S),eqs(B,N,_,_),member(F,N),((B=upper,S\==inf);(B=lower,S\==0))),LS),
	length(LS,TotalReducedSolved).

show_eqs :-
	eqs(_,_,_,_),!,
	write('============================( Equations )==========================='),nl,
	show_eqs_,
	write('===================================================================='),nl.
show_eqs.

show_eqs_:-
	eqs(Bound,_,Eqs,_),
	Eqs = [equation(F,Vs,_,_)|_],
	write(F),
	write(':'),
	write(Bound),nl,
	numbervars(Eqs,0,_),
	write(Eqs),nl,
	(solution(F,Vs,Sol) -> write('Closed form: '),write(Sol),nl; true),
	write('--------------------------------------------------------------------'),nl,
	fail.
show_eqs_.

%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%Normalization and minimization
unify_calls:-
	current_fact(eqs(_,L,[equation(N,_,_,_)|_],_)),
	replace_calls(L,N),
	fail.
unify_calls.

normalize_eqs:-
	retract_fact(eqs(B,L,Eqs,Del)),
	normalize_eqs_(Eqs,NEqs),
	unify_vars(NEqs),
	assertz_fact(eqs(B,L,NEqs,Del)),
	fail.
normalize_eqs.

minimize_args:-
	topological_sccs(SCCS),!,
	minimize_args_(SCCS),
	retractall_fact(top_scc(_)).

minimize_args_([]).
minimize_args_([SCCS|T]):-
	minimize_args__(SCCS),!,
	minimize_args_(T).

minimize_args__(TopOrder):-
	member(F,TopOrder),
	retract_fact(eqs(B,L,[equation(F,Ar,Ex,Test)|T],_Del)),
	Eqs0 = [equation(F,Ar,Ex,Test)|T],
	slice_args(Eqs0,RemArgs,NEqs),
	slice_calls(F,RemArgs),
	assertz_fact(eqs(B,L,NEqs,RemArgs)),
	fail.
minimize_args__(_).

slice_calls(F,RemArgs):-
	retract_fact(eqs(B,L,Eqs,Del)),
	Eqs=[equation(FF,_,_,_)|_],
	FF \== F,
	findall(Pos, (term_at_position(Pos,Eqs,Term), nonvar(Term), (Term=fun(G,_Args),G == F)), LPos),
	slice_calls_(LPos,RemArgs,Eqs,NEqs),
	assertz_fact(eqs(B,L,NEqs,Del)),
	fail.
slice_calls(_,_).

slice_calls_([],_,E,E).
slice_calls_([Pos|T],RemArgs,Eqs,NEqs):-
	term_at_position(Pos,Eqs,fun(F,Args)),
	remove_elements(RemArgs,Args,NArgs),
	update_term(Eqs,Pos,fun(F,NArgs),Eqs0),
	slice_calls_(T,RemArgs,Eqs0,NEqs).

update_eqs(E,[],_,E).
update_eqs(E,[Pos|T],NVar,TEqs):-
	update_term(E,Pos,NVar,Eqs0),
	update_eqs(Eqs0,T,NVar,TEqs).

eqs_calls(Eq, Calls):-
	Eq = [equation(N,_,_,_)|_],
	findall(F, (term_at_position(_,Eq,Term), nonvar(Term), (Term=fun(F,_Args),F \= N)), Calls0),
	sort(Calls0,Calls),!.
eqs_calls(_Eq, []).


unify_vars([]).
unify_vars([equation(_N,Args,_Exp,_T)|R]):-
	unify_vars_(Args,R).


unify_vars_(_Args,[]).
unify_vars_(Args,[equation(_N,Args,_Exp,_T)|R]):-
	unify_vars_(Args,R).

normalize_eqs_([],[]).
normalize_eqs_([Eq|T],[NEq|TT]):-
	normalize_eq_(Eq,NEq),
	normalize_eqs_(T,TT).

normalize_eq_(equation(N,Args,Expr,Test),equation(N,NArgs,Expr,NTest)):-
	peel_(Args,NArgs,NTest,Test).


peel_([],[],E,E).
peel_([X|R],[X|RR],E,T):-
	var(X),!,
	peel_(R,RR,E,T).
peel_([X|R],[_|RR],E,T):-
	X == '?',!,
	peel_(R,RR,E,T).
peel_([X|R],[Y|RR],[Y=X|E],T):-
	peel_(R,RR,E,T).

%--------------------------------------------------------------------------

assume(F,Vars,Sol):-
	retractall_fact(assumption(F,_,_)),
	assertz_fact(assumption(F,Vars,Sol)).

gather_and_solve_eqs(An):-
	statistics(runtime,_),
	clean_eqs,
	gather_eqs(An),
	unify_calls,
	normalize_eqs,
	minimize_args,	
	recompute,
	statistics(runtime,[_,Time]),
	asserta_fact(runtime(Time)).

solve_eqs(An):-
	statistics(runtime,_),
	clean_eqs,
	gather_eqs(An),
	unify_calls,
	normalize_eqs,
	minimize_args,
	solve_eqs__,
	write_results(An),
	statistics(runtime,[_,Time]),
	asserta_fact(runtime(Time)).
	
%inc_solve_eqs(An,Preds):-
%	statistics(runtime,_),

recompute:-
	topological_sccs(TopOrder),
	retractall_fact(solution(_,_,_)),
	solve_eqs_recompute(TopOrder).
	

solve_eqs__:-
	topological_sccs(TopOrder),
	solve_eqs_(TopOrder).

replace_calls_(E,_L,_N,[],E).
replace_calls_(E,L,N,[Pos|T], NE):-
	term_at_position(Pos,E,fun(F,Args)),
	(member(F,L) ->
	    update_term(E,Pos,fun(N,Args),NE0)
	;
	    NE0=E
	),
	 replace_calls_(NE0,L,N,T,NE).

replace_calls(L,N):-
	retract_fact(eqs(B,U,Eqs,Del)),
 	findall(Pos, (term_at_position(Pos,Eqs,Term), nonvar(Term), (Term=fun(F,_Args),F \= N)), LPos),
	replace_calls_(Eqs,L,N,LPos,NEqs),
	assertz_fact(eqs(B,U,NEqs,Del)),
	fail.
replace_calls(_,_).

% % --------------------------------------------
% % Predicates for pretty-print of equations 
% % --------------------------------------------
% % equation_name/1 stores the generated names of equations. This is
% % used after in res_abs to distinguish which functors need to be
% % replaced for pretty-print versions during the output.
% % :- export(equation_name/1).
% % :- data equation_name/1.
% % counter_functions_name/1 is a counter used for generation of new
% % equation names.
% :- data counter_function_name/1.

% % :- pred beautify_equations(Eq,EqB): (list(Eq,equation),var(EqB)) =>
% % 	list(EqB) # "It receives a list of terms
% % 	equation(Name,Args,Expr) and returns a list of pretty-print
% % 	versions of this equations (NewName(Args) == NewExpr),
% % 	consistently replacing equations names with shorter (more
% % 	readable) ones.".
% beautify_equations(Eq,EqB):-
% 	retract_fact(mapping_fn_names(Map)),
% 	beautify_equations_(Eq,Map,EqB,Map2),!,
% 	asserta_fact(mapping_fn_names(Map2)).

% beautify_equations_([],M,[],M).
% beautify_equations_([equation(F,Args,Expr,T)|R],Map,[equation(FName,Args,E,T)|BeautyR],Map2):-
% 	get_function_name(Map,F,FName,NMap),
% 	beautify_equations_expr(Expr,NMap,E,NNMap),
% 	beautify_equations_(R,NNMap,BeautyR,Map2).

% beautify_equations_expr(Expr,NMap,Expr,NMap):-
% 	var(Expr),!.
% beautify_equations_expr(Expr,NMap,fun(Name,NArgs),NNMap):-
% 	Expr = fun(F,Args),!,
% 	get_function_name(NMap,F,Name,Map1),
% 	beautify_equations_expr_args(Args,Map1,NArgs,NNMap).
% beautify_equations_expr(Expr,NMap,E,NNMap):-
% 	Expr =.. [F|Args],
% 	beautify_equations_expr_args(Args,NMap, NArgs,NNMap),
% 	E =.. [F|NArgs].

% beautify_equations_expr_args([],M,[],M).
% beautify_equations_expr_args([A|R],M,[NA|NR],NM):-
% 	beautify_equations_expr(A,M,NA,M0),!,
% 	beautify_equations_expr_args(R,M0,NR,NM).

% get_function_name(Map,F,FName,Map):-
% 	member((F,FName),Map),!.
% get_function_name(Map,F,FName,[(F,FName)|Map]):-
% 	get_new_function_name(FName),!.

% get_new_function_name(F):-
% 	retract_fact(counter_function_name(C)),
% 	C1 is C + 1,
% 	asserta_fact(counter_function_name(C1)),
% 	number_to_function(C,F).

% % Given a number, returns an atom that will be used as an
% % equation/function name.
% number_to_function(C,F):-
% 	C =< 20,!,
% 	C1 is C + 102,
% 	atom_codes(F,[C1]).
% number_to_function(C,F):-
% 	Suf is C - 20,
% 	number_codes(Suf,CSuf),
% 	atom_codes(F,[102|CSuf]).

% number_to_function_([],[]).
% number_to_function_([N|R],[F|T]):-
% 	F is N + 54,
% 	number_to_function_(R,T).

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

constant_var(ArgNumber, equation(F,Args,Expr,Test)):-
	findall(Pos, (term_at_position(Pos,Expr,STerm), nonvar(STerm), (STerm=fun(F,_Args))), RecCalls),
	nth(ArgNumber,Args,Arg),
	constant_var_(ArgNumber,RecCalls,Arg,equation(F,Args,Expr,Test)).

constant_var_(_,[],_,_).
constant_var_(ArgNumber,[Pos|T], Arg,equation(F,Args,Expr,Test)):-
	term_at_position(Pos,Expr,fun(F,CArgs)),
	nth(ArgNumber,CArgs,CArg),
	CArg == Arg,!,
	constant_var_(ArgNumber,T,Arg,equation(F,Args,Expr,Test)).
	
needed_arg1_(ArgNumber,equation(F,Args,Expr,Test)):-
	nth(ArgNumber,Args, Arg),
	findall(Pos, (term_at_position(Pos,Expr,STerm), nonvar(STerm), (STerm=fun(F,_Args))), RecCalls),
	remove_rec_calls(RecCalls,Expr,Expr0),
	varset(Expr0,VarsExpr),
	varset(Test,VarsTest),
	(varmember(Arg,VarsExpr);varmember(Arg,VarsTest)).

remove_rec_calls([],E,E).
remove_rec_calls([Pos|T],Expr,NExpr):-
	update_term(Expr,Pos,rc,Expr0),
	remove_rec_calls(T,Expr0,NExpr).

constant_arg(_,[]).
constant_arg(ArgNumber,[Eq|T]):-
	constant_var(ArgNumber,Eq),!,
	constant_arg(ArgNumber,T).

needed_arg(ArgNumber,[Eq|_T]):-
	needed_arg1_(ArgNumber,Eq),!.
needed_arg(ArgNumber,[_Eq|T]):-
	needed_arg(ArgNumber,T),!.

remove_elements(ArgNum,List,Result):-
	remove_elements_(1,ArgNum,List,Result).
remove_elements_(_,_,[],[]).
remove_elements_(I,ArgNum,[X|T],[X|TT]):-
	\+ member(I,ArgNum),!,
	I1 is I + 1,
	remove_elements_(I1,ArgNum,T,TT).
remove_elements_(I,ArgNum,[_X|T],TT):-
	I1 is I + 1,
	remove_elements_(I1,ArgNum,T,TT).

remove_member(_,[],[]).
remove_member(1,[_|L],L).
remove_member(X,[P|T],[P|TT]):-
	X>1,
	X1 is X -1,
	remove_member(X1,T,TT).

remove_args_test(_X,[],[]).
remove_args_test(ArgNum,[Cond|T],TT):-
	varset(Cond, F),
	varinclusion(F,ArgNum),!,
	remove_args_test(ArgNum,T,TT).
remove_args_test(X,[Cond|T],[Cond|TT]):-
	remove_args_test(X,T,TT).

varinclusion([],_).
varinclusion([X|T], TT):-
	varmember(X,TT),!,
	varinclusion(T,TT).

varmember(X,[Y|_T]):-
	X==Y,!.
varmember(X,[_Y|T]):-
	varmember(X,T).

remove_arg_eqs([],E,E).
remove_arg_eqs(_ArgNumber,[],[]).
remove_arg_eqs(ArgNumbers,[Eq|Eqs],[NEq|NEqs]):-
	remove_arg_eq(ArgNumbers,Eq,NEq),
	remove_arg_eqs(ArgNumbers,Eqs,NEqs).

remove_arg_eq(ArgNumbers, equation(N,Args,Expr,Test), equation(N,NArgs,NExpr,Test)):-
	remove_elements(ArgNumbers,Args,NArgs),
	remove_arg_exp(ArgNumbers,N,Expr,NExpr).

remove_arg_exp(ArgNumbers,Fn,Expr,NExpr):-
	findall(Pos, (term_at_position(Pos,Expr,Term), nonvar(Term), Term=fun(Fn,_Args)), LPos),
	remove_arg_exp_(ArgNumbers,LPos,Expr,NExpr).

remove_arg_exp_(_,[],E,E).
remove_arg_exp_(N,[Pos|T], Expr,NExpr):-
	term_at_position(Pos,Expr,Call),
	Call = fun(A,Args),
	remove_elements(N,Args,NArgs),
	update_term(Expr,Pos,fun(A,NArgs),Expr0),
	remove_arg_exp_(N,T,Expr0,NExpr).

slice_args(Eqs,Args, NEqs):-
	select_needless_args(Eqs,Args),!,
	remove_arg_eqs(Args,Eqs,NEqs).

select_needless_args(Eqs,NArgs):-
	Eqs = [equation(_,Args,_,_)|_],
	length(Args,N),
	select_needless_args_(1,N,Eqs,NArgs).

select_needless_args_(Y,N,_,[]):-
	Y > N.
select_needless_args_(I,N,Eqs,[I|T]):-
	\+ needed_arg(I,Eqs),!,
	I1 is I + 1,
	select_needless_args_(I1, N, Eqs, T).
select_needless_args_(I,N,Eqs,T):-
	I1 is I + 1,
	select_needless_args_(I1, N, Eqs, T).

select_constant_args(Eqs,NArgs):-
	Eqs = [equation(_,Args,_,_)|_],
	length(Args,N),
	select_constant_args_(1,N,Eqs,NArgs).

select_constant_args_(Y,N,_,[]):-
	Y > N,!.
select_constant_args_(I,N,Eqs,[I|T]):-
	constant_arg(I,Eqs),!,
	I1 is I + 1,
	select_constant_args_(I1, N, Eqs, T).
select_constant_args_(I,N,Eqs,T):-
	I1 is I + 1,
	select_constant_args_(I1, N, Eqs, T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% SSCs                                                         %%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% vertex(Id,index,lowlink,onstack)

:- export(top_scc/1).
:- data scc/1.
:- data top_scc/1.

topological_sccs(List):-
	current_fact(top_scc(List)),!.
topological_sccs(List):-
	tarjan,!,
	findall(SCC,scc(SCC),List),
	asserta_fact(top_scc(List)).
	

get_graph(List):-
	findall(vertex(A,0,0,false)-L,(eqs(_,_,[equation(A,B,Exp,T)|R],_), eqs_calls([equation(A,B,Exp,T)|R], L)), List).


tarjan:-
	retractall_fact(scc(_)),
	get_graph(Graph),
	tarjan_(Graph,1,[],_).


tarjan_(Graph,I,S,S1):-
	member(vertex(V,0,_,_)-_E, Graph),!,
	sc(V,Graph,I,S,I1,Graph1,S0),
	tarjan_(Graph1,I1,S0,S1).
tarjan_(_,_,_,_).

sc(V,Graph,I,S,I1,Graph1,SSC):-
	update_vertex(Graph,vertex(V,I,I,true),Graph0),
	I0 is I + 1,
	S0 = [V|S],
	member(vertex(V,_,_,_)-SuccV, Graph0),
	sc_succ(SuccV,V,Graph0,I0,S0,I1,Graph00,S00),
	(member(vertex(V,Low,Low,_)-_,Graph00) ->
	 generate_scc(V,S00,Graph00,Graph1,SSC)
	;
	    SSC=S00,Graph1=Graph00
	).

sc_succ([],_,G,I,S,I,G,S).
sc_succ([Succ|T],V, Graph, I,S,I1,Graph1,S1):-
	member(vertex(Succ,0,_,_)-_E, Graph),!,
	sc(Succ, Graph,I,S,I0,Graph0,SSC0),
	member(vertex(V,IP,LLV,On)-_, Graph0),
	member(vertex(Succ,_,LLSucc,_)-_, Graph0),
	min(LLSucc,LLV,NLLV),
	update_vertex(Graph0,vertex(V,IP,NLLV,On),Graph00),
	sc_succ(T,V,Graph00,I0,SSC0,I1,Graph1,S1).
sc_succ([Succ|T],V, Graph, I,S,I1,Graph1,S1):-
	member(vertex(Succ,_,LLSucc,true)-_, Graph),!,
	member(vertex(V,IP,LLV,On)-_, Graph),
	min(LLSucc,LLV,NLLV),
	update_vertex(Graph,vertex(V,IP,NLLV,On),Graph0),
	sc_succ(T,V,Graph0,I,S,I1,Graph1,S1).
sc_succ([_Succ|T],V, Graph, I,S,I1,Graph1,S1):-
	sc_succ(T,V,Graph,I,S,I1,Graph1,S1).

generate_scc(V,Stack,Graph,Graph1,Rest):-
	generate_scc_(V,Stack,Graph,[],SCC,Rest,Graph1),
	assertz_fact(scc(SCC)).


generate_scc_(V,[V|R],G,SCC,[V|SCC],R,G1):-
	member(vertex(V,I,LL,_)-_E,G),
	update_vertex(G,vertex(V,I,LL,false),G1),!.
generate_scc_(V, [W|T],G,SCCs,SCC,Rest,G1):-
	member(vertex(W,I,LL,_)-_E,G),
	update_vertex(G,vertex(W,I,LL,false),G0),!,
	generate_scc_(V,T,G0,[W|SCCs],SCC,Rest,G1).

update_vertex([vertex(Id,_I,_LL,_On)-Ed|R],vertex(Id,I,LL,On),[vertex(Id,I,LL,On)-Ed|R]):-!.
update_vertex([VX|R],Vertex,[VX|RR]):-
	update_vertex(R,Vertex,RR).

min(X,Y,X):-
	X<Y,!.
min(_,Y,Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% get_rels(List):-
% 	findall(A-L,(eqs(_,_,[equation(A,B,Exp,T)|R],_), eqs_calls([equation(A,B,Exp,T)|R], L)), List).

% topological_order(TopOrder):-
% 	current_fact(top_order(TopOrder)),!.
% topological_order(TopOrder):-
% 	get_rels(List),
% 	topological_order_(List,TopOrder),
% 	asserta_fact(top_order(TopOrder)).

% topological_order_([],[]).
% topological_order_(List,[F|R]):-
% 	member(F-[], List),
% 	delete(List,F-[],NList0),
% 	remove_rel(NList0,F,NList),
% 	topological_order_(NList, R).

% remove_rel([],_,[]).
% remove_rel([F-L|T], A, [F-L1|TT]):-
% 	delete(L,A,L1),
% 	remove_rel(T,A,TT).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%Incremental version %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% inc_solve_eqs_([],_,E,E).
% inc_solve_eqs_([SCC|T], before,Solved, NSolved):-
% 	member(F1,SCC),member(F1,Solved),
% 	inc_solve_eqs__(SCC, Solved,Solved0),!,
% 	inc_solve_eqs_(T,after,Solved0,NSolved).
% inc_solve_eqs_([_SCC|T], before,Solved, NSolved):-
% 	inc_solve_eqs_(T,before,Solved,NSolved).
% inc_solve_eqs_([SCC|T], after,Solved, NSolved):-
% 	inc_solve_eqs__(SCC, Solved,Solved0),!,
% 	inc_solve_eqs_(T,after,Solved0,NSolved).


% inc_solve_eqs__([],E,E).
% inc_solve_eqs__([F|T],Solved,NSolved):-
% 	(fdebug_(F);true),
% 	current_fact(eqs(Bound,_,Eqs,_)),
% 	Eqs = [equation(F,Vars,_,_)|_],!,
% 	needs_recomputation(F,Vars,Bound,Eqs,Solved),
% 	current_pp_flag(req_solver,Solver),
% 	rsolve_(Solver,Eqs,Vars,Bound,Solution0),
% 	simplify(Solution0,Solution),!,
% 	update_fact(solution(F,Vars,Solution)),
% 	inc_solve_eqs__(T,[F|Solved],NSolved).
	
	
% %	current_fact(solution(F,Vars,))
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

solve_eqs_recompute([]).
solve_eqs_recompute([SCC|T]):-
	solve_eqs_recompute_(SCC),!,
	solve_eqs_recompute(T).



solve_eqs_([]).
solve_eqs_([SCC|T]):-
	solve_eqs___(SCC),!,
	solve_eqs_(T).

solve_eqs___([]).
solve_eqs___([F|T]):-
	(fdebug_(F);true),
	current_fact(eqs(Bound,_,Eqs,_)),
	Eqs = [equation(F,Vars,_,_)|_],!,
	current_pp_flag(req_solver,Solver),
	rsolve_(Solver,Eqs,Vars,Bound,Solution0),
	simplify(Solution0,Solution),!,
	assertz_fact(solution(F,Vars,Solution)),
	solve_eqs___(T).

solve_eqs_recompute_([]).
solve_eqs_recompute_([F|T]):-
	(fdebug_(F);true),
	(
	    current_fact(assumption(F,Vars,Solution)) ->
	    retractall_fact(solution(F,_,_)),
	    assertz_fact(solution(F,Vars,Solution))
	;

	    current_fact(eqs(Bound,_,Eqs,_)),
	    Eqs = [equation(F,Vars,_,_)|_],!,
	    current_pp_flag(req_solver,Solver),
	    rsolve_(Solver,Eqs,Vars,Bound,Solution0),
	    simplify(Solution0,Solution),!,
	    assertz_fact(solution(F,Vars,Solution))
	    ),!, 
	solve_eqs_recompute_(T).



fdebug_(F):-
	fdebugset(LF),
	member(F,LF),!,
	debug_function_set.

debug_function_set.



:- pred simplify_equations(Equations,SimplifiedEquations).
:- export(simplify_equations/2).
simplify_equations([],[]).
simplify_equations([equation(I,O,Expr,T)|R],[equation(I,O,Simp,T)|SimpR]) :-
	simplify(Expr,Simp),!,
	simplify_equations(R,SimpR).


rsolve_(Solver,Eqs0,Vars,Bound,Solution):-
	denormalize_tests(Eqs0,Eqs00),
	inline_solutions(Eqs00,NEqs000),
	%unfold_if_equal(NEqs00,NEqs000),
	add_eqs(NEqs000,NEqs10),
	inline_ground_calls(NEqs10,NEqs11),
	simplify_equations(NEqs11,NEqs1),
	(already_solved(Bound,NEqs1,Solution)
	;
	    % (Solver == mathematica ->
	    %  select_constant_args(NEqs1,ConstantArgs),
	    %  remove_arg_eqs(ConstantArgs,NEqs1,NEqs),
	    %  remove_elements(ConstantArgs,Vars,NVars)
	    % ;
	    NEqs1 = NEqs,NVars=Vars,
	    % ),
	    (rsolve(Solver,NEqs,NVars,Bound,Solution);assertz_fact(unsolved(NEqs)),fail)
	),!.
rsolve_(_Solver,_Eqs,_Vars,lower,0):- !.
rsolve_(_Solver,_Eqs,_Vars,upper,inf):- !.


closed_form(Eqs):-
	term_at_position(_,Eqs,R),
	nonvar(R),
	R = fun(_,_),!,
	fail.
closed_form(_).



non_recursive_eq(equation(F,_,Exp,_)):-
	non_recursive_exp(F,Exp).

non_recursive_exp(_F,V):-
	var(V),!.
non_recursive_exp(_F,V):-
	number(V),!.
non_recursive_exp(F,V):-
	V = fun(F,_),!,
	fail.
non_recursive_exp(F,V):-
	functor(V,_,NArgs),
	non_recursive_exp_args(F,V,1,NArgs),!.

non_recursive_exp_args(_,_,I,NArgs):-
	I > NArgs,!.
non_recursive_exp_args(F,V,I,NArgs):-
	I =< NArgs,
	arg(I,V,ArgV),
	I1 is I + 1,
	non_recursive_exp(F,ArgV),!,
	non_recursive_exp_args(F,V,I1,NArgs).

already_solved(Bound, Eqs,Solution):-
	closed_form(Eqs),
	unconditional_eqs(Eqs),!,
	get_exps(Eqs,SetExp),
	set_min_max_exp(Bound,SetExp,Solution0),
	simplify(Solution0,Solution).
already_solved(Bound, Eqs,Solution):-
	solve_special_equation(Eqs,Bound,Solution).
	
unconditional_eqs([]).
unconditional_eqs([equation(_,_,_,[])|T]):-
	unconditional_eqs(T).

get_exps([],[]).
get_exps([equation(_,_,E,_)|T],[E|TT]):-
	get_exps(T,TT).

set_min_max_exp(upper,SetExp,inf):-
	member(Mem,SetExp), (Mem==inf;Mem=='?'),!.
set_min_max_exp(lower,SetExp,0):-
	member(Mem,SetExp), (Mem==inf;Mem=='?'),!.
set_min_max_exp(Bound,SetExp,Sol):-
	set_min_max_exp_(Bound,SetExp,Sol).
set_min_max_exp_(_,[E],E).
set_min_max_exp_(upper,[A|T],max(A,Sol0)):-
	set_min_max_exp_(upper,T,Sol0).
set_min_max_exp_(lower,[A|T],min(A,Sol0)):-
	set_min_max_exp_(lower,T,Sol0).

denormalize_tests([],[]).
denormalize_tests([equation(F,Args,Exp,Test)|T], [equation(F,NArgs,Exp,NTest)|TT]):-
	inline_test(Args,Test,NArgs,NTest),
	denormalize_tests(T,TT).

inline_test([],T,[],T).
inline_test([Var|T], Tests, [C|TT], NTests):-
	member(AVar=C,Tests),
	AVar == Var,
	nonvar(C),
	num(C),!,
	delete_var(Tests, AVar=C, NTests0),
	inline_test(T,NTests0,TT,NTests).
inline_test([Var|T], Tests, [Var|TT], NTests):-
	inline_test(T,Tests,TT,NTests).


delete_var([],_,[]).
delete_var([X|T], Var=C, Test):-
	X = (Var1=C1),
	Var1 == Var,
	C1 == C,!,
	delete_var(T,Var=C,Test).
delete_var([X|T], Con, [X|Test]):-
	delete_var(T,Con,Test).

% if Eqs = (f(x) = g(x), and g is not solved already,
% replace def of f with a copy of g)
% unfold_if_equal([equation(F,Args,Exp,[])], NEqs):-
% 	nonvar(Exp),
% 	Exp = fun(G,OArgs),
% 	OArgs == Args,
% 	current_fact(eqs(_,N,GEqs,_)),
% 	\+ current_fact(solution(G,_,_)),
% 	member(G,N),!,
% 	replace_functor(F,G,GEqs,NEqs),
% 	NEqs = [equation(_,Args,_,_)|_].
% unfold_if_equal(E,E).

replace_functor(F,G,GEqs,NEqs):-
	term_at_position(Pos,GEqs,Term),
	Term == G,!,
	update_term(GEqs, Pos, F, GExp0),
	replace_functor(F,G,GExp0,NEqs).
replace_functor(_,_,E,E).
 
inline_solutions(Eqs,NEqs):-
	inline_solutions_(Eqs,[],NEqs),!.
inline_solutions_(Eqs,Analyzed,NEqs):-
	Eqs = [equation(N,_,_,_)|_],
	term_at_position(Pos,Eqs,Term),
	nonvar(Term),
	Term=fun(G,Args),
	G \== N,
	\+ member(G,Analyzed),
	(current_fact(solution(G,Vars,Exp)) ->
	 Vars = Args,NSet = Analyzed,
	 update_term(Eqs,Pos,Exp,Eqs0)
	;
	    Eqs0 = Eqs,NSet = [G|Analyzed]
	),!,
	inline_solutions_(Eqs0,NSet,NEqs).
inline_solutions_(NEqs,_,NEqs).

add_eqs(Eqs,NEqs):-
	add_eqs_(Eqs,[],NEqs),!.
add_eqs_(Eqs,Replaced,NEqs):-
	Eqs = [equation(N,_,_,_)|_],
	term_at_position(_Pos,Eqs,Term),
	nonvar(Term),
	Term=fun(G,_Args),
	G \== N,
	\+ member(G,Replaced),!,
	current_fact(eqs(_,FF,CalledG0,_)),
	member(G,FF),
	denormalize_tests(CalledG0,CalledG),
	append(Eqs,CalledG,Eqs0),!,
	add_eqs_(Eqs0,[G|Replaced],NEqs).
add_eqs_(NEqs,_,NEqs).


inline_ground_calls(Eqs,REqs):-
	obtain_call_solution_pairs(Eqs,Map),
	inline_ground_calls_(Map,Eqs,RREqs),
	simplify_equations(RREqs,REqs).

inline_ground_calls_(T,E,NE):-
	term_at_position(Pos,E,Term),
	nonvar(Term),
	member((Term1,Val),T),Term1==Term,!,
	update_term(E,Pos,Val,E0),
	inline_ground_calls_(T,E0,NE).
inline_ground_calls_(_,E,E).

obtain_call_solution_pairs([],[]).
obtain_call_solution_pairs([equation(F,Args,Exp,_)|T],[(fun(F,Args),Exp)|R]):-
	\+ \+ maplist(num,Args),
	\+ \+ num(Exp),!,
	obtain_call_solution_pairs(T,R).
obtain_call_solution_pairs([_|T],R):-
	obtain_call_solution_pairs(T,R).