:- module(term_utilities,
	[
	    term_at_position/3,
	    get_subterm/3,
	    at/3,
	    get_subterm_cond/3,
	    get_var_position/2,
	    replace_subterm_cond/4,
	    update_term/4
	],
	[
	    hiord,
	    regexp,
	    fsyntax,
	    assertions,
	    nativeprops,
	    regtypes
	]).


:- doc(title, "Term handling utilities using subterm positions.").  
:- doc(author, "Maximiliano Klemen").  
:- doc(module, 
	"This module provides utilities for querying and updating
      terms using subterm positions. Considering a term as a tree
      structure, a subterm position is a list of numbers representing
      the path from the root to the subterm corresponding to that
      position, where the the children are numbered starting at 1.
      For example, in the term a(1,2,b(3,4)), the position of 4 is
      [3,2], and the position of 2 is [2]. Some of this code comes
      from the module eq_solver_term_position, authored by Alejandro
      Serrano. This intends to be a simpler implementation, with some
      additional operations, such as the use of regular
      expressions for representing set of subterm positions.").

:- use_module(library(sort)).
:- use_module(library(aggregates)).
:- use_module(library(lists)).
:- use_module(library(hiordlib)).

:- pred term_at_position(Position,Term,InnerTerm) 
	: list(num) * term  * var 
 => list(num) * term * term #
 "Gets the @var{InnerTerm} found at the specified @var{Position} in
 @var{Term}".

:- pred term_at_position(Term,Position,InnerTerm)
	: var * term * term
 =>  list(num) * term * term #
 "Gets the @var{Position} of the term @var{InnerTerm} in @var{Term},
if @var{InnerTerm} is a subterm of @var{Term}".

:- pred term_at_position(Term,Position,InnerTerm)
	:  var * term * var
 => list(num) * term * term #
 "Gets the subterm @var{InnerTerm} and its position @var{Position} in
 @var{Term}, being @var{InnerTerm} a subterm of @var{Term}".
 
term_at_position([],X,X):-nonvar(X).
term_at_position([],X,X):-var(X),!.
term_at_position([E|R],T,Inner) :-
	nonvar(T),
	T =..[_|Args],!,
	nth(E,Args,OneStepChild),
	term_at_position(R,OneStepChild,Inner).


:- pred get_subterm(RegExp,Term,SubTerm)
	: struct_regexp * term * var
 => struct_regexp * term * term #
"Gets the subterm @var{SubTerm} of @var{Term}, whose position matches
the regular expression @var{RegExp} (interpreting list of numbers as a
string of the corresponding atom codes, i.e. [1,2] is interpreted as
the string 12)".

get_subterm(Reg,Term,Elem):-
	term_at_position(Pos,Term,Elem),
	match_struct(Reg,_,Pos,[]).

get_subterm_(Reg,Term,(Pos,Elem)):-
	term_at_position(Pos,Term,Elem),
	match_struct(Reg,_,Pos,[]).


:- pred at(Term,RegExp,SubTerms)
	: term * struct_regexp * var
 =>  term * struct_regexp * term #
"Gets a list of all those subterms @var{SubTerms} of @var{Term}, whose
position matches the regular expression @var{RegExp} (interpreting
list of numbers as a string of the corresponding atom codes,
i.e. [1,2] is interpreted as the string 12)".

at(Term,Reg,List):-
	findall(Elem,get_subterm(Reg,Term,Elem),List).

:- meta_predicate get_subterm_cond(_,pred(1),_).
:- pred get_subterm_cond(Term,Condition,PositionAndSubTerm)
	: term * callable * var
 => term * callable * term #
"Gets a pair (position,subterm) in @var{SubTerm}, where subterm is a
subterm of @var{Term} that verifies the condition @var{Condition}, and
position is its position in @var{Term}. @var{Condition} is called with
the subterm as the first argument, and it can have additional
arguments".

get_subterm_cond(Term,Cond,(Pos,E)):-
	term_at_position(Pos,Term,E),
	\+ \+ Cond(E).

% get_subterm_cond2(RegExp,Term,Cond,E):-
% 	get_subterm(RegExp,Term,E),
% 	\+ \+ Cond(E).



:- pred get_var_position(Term,PositionAndVar)
	: term * var
 => term * term #
"Gets all the occurrences of variables in @var{Term} by backtraking,
as a pair (Pos,Var), where Var is the variable and Pos is the position
in @var{Term} where it appears".

get_var_position(Term,VarPosPair):-
	get_subterm_cond(Term,var,VarPosPair).


:- meta_predicate replace_subterm_cond(_,pred(1),pred(3),_).
:- pred replace_subterm_cond(Term,Condition,Transformation,NewTerm)
	: term * callable * callable * var
 => term * callable * callable * term #
"Replaces all those subterms of @var{Term} that made the predicate
@var{Condition} succeeded as a first argument, with the output of
@var{Transformation}, that receives, in that order, the position of
the term to be replaced, the term to be replaced, and the third
argument is a variable that is bound with the term that is going to be
in place.  Both predicates, @var{Condition} and @var{Transformation}
can have additional arguments after the ones mentioned. This
predicate might not terminate, if @var{Transformation} and
@var{Condition} causes an infinite succession of subterms
to be replaced.".

replace_subterm_cond(Term,Condition,Transformation,NewTerm):-
	get_subterm_cond(Term,Condition,(Pos,E)),!,
	Transformation(Pos,E,Result),
	update_term(Term,Pos,Result,Term0),!,
	replace_subterm_cond(Term0,Condition,Transformation,NewTerm).
replace_subterm_cond(Term,_,_,Term).


:- pred update_term(Term,Position,NewInnerTerm,NewTerm)
	: term * list(num) * term * var 
 => term * list(num) * term * term #
"@var{NewTerm} is the term that results from replacing the subterm of
@var{Term}, at position @var{Position}, with the term
@var{NewInnerTerm}".

update_term(_X,[],E,E).
update_term(T,[E|R],Elem,TNew) :- 
	T =..[Op|Args],!,
	nth(E,Args,OneStepChild),
	update_term(OneStepChild,R,Elem,Inner),
	update_list(E,Args,Inner,ArgsNew),
	TNew =..[Op|ArgsNew].

update_list(1,[_|T],Elem,[Elem|T]):-!.
update_list(I,[P|T],Elem,[P|NT]):-
	I > 1,!,
	I1 is I - 1,
	update_list(I1,T,Elem,NT).


% %%%%Main predicate%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% :- meta_predicate operates(_,_,_,pred(3),_).
% operates(Term,RegExp,Ops,Coll,NTerm):-
% 	findall(Sub,(gen(RegExp,Term,U),operates_(U,Ops,Sub)),LSub),
% 	Coll(Term,LSub,NTerm).

% operates_(U,[],U).
% operates_(U,[filtert(C)|Tail],R):-
% 	filtert(U,C),
% 	operates_(U,Tail,R).
% operates_(U,[filterx(C)|Tail],R):-
% 	filterx(U,C),
% 	operates_(U,Tail,R).
% operates_(U,[filteru(C)|Tail],R):-
% 	filteru(U,C),
% 	operates_(U,Tail,R).
% operates_(U,[mapt(C)|Tail],R):-
% 	mapt(U,C,NU),
% 	operates_(NU,Tail,R).
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% %%%%SubTerm Generator%%%%%%%%%%%%%%%%%%%%%%%%%%%
% gen(RegExp,Term,(Pos,SubTerm)):-
% 	get_subterm_(RegExp,Term,(Pos,SubTerm)).

% all := ['*'(any)].
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% %%%%Filters and map operations%%%%%
% :- meta_predicate mapt(_,pred(2),_).
% :- pred mapt/3
% 	: term * callable * term.
% mapt((Pos,SubTerm),Map,(Pos,NTerm)):-
% 	Map(SubTerm,NTerm).

% :- meta_predicate filtert(_,pred(1)).
% filtert((_Pos,SubTerm),Cond):-
% 	Cond(SubTerm).

% filterx((Pos,_SubTerm),RegExp):-
% 	match_struct(RegExp,_,Pos,[]).

% filteru((_Pos,SubTerm),Term):-
% 	Term=SubTerm.
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% %%%%%%Collectors%%%%%%%%%%%%%%%%%%%%
% update(T,[],T).
% update(T,[(P,M)|Tail],NT):-
% 	update_term(T,P,M,T0),
% 	update(T0,Tail,NT).

% as_list(_,[],[]).
% as_list(_,[(_,E)|T],[E|TT]):-
% 	as_list(_,T,TT).

% :- meta_predicate reduce(_,_,pred(3),_,_).
% reduce(_,V0,_,[],V0).
% reduce(T,V0,Op,[(_,S)|Tail],Res):-
% 	Op(V0,S,V1),
% 	reduce(T,V1,Op,Tail,Res).
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% %%%example mapping functions%%%
% add(X,Y,Z):-
% 	Z is X + Y.

% % num2list(N,L):-
% % 	length(L,N).

% one(_,1).

% changef(tree(L,E,R),t(E,L,R)).
% changef(nil,n).
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

