:- module(resdefs_tr, [resdefs_sentence_tr/3, resdefs_goal_tr/3], [assertions, datafacts]).

:- doc(author, "Edison Mera").
:- doc(author, "Teresa Trigo").

:- use_module(library(aggregates)).
:- use_module(library(messages)).
:- use_module(library(compiler/c_itf), [location/1]).
:- use_module(library(resdefs/rescostfunc)).

:- data resource_db/2, def_ru_db/2, def_gru_db/2.
:- data resource_default_db/2. % Keeps track of resource cost declarations issued, in order to issue
                               % suitable defaults at the end if needed. 
:- export(resource_db/2).

assert_resource(R, M) :- assertz_fact(resource_db(R, M)).
assert_resource_default(Decl, M) :- assertz_fact(resource_default_db(Decl, M)).
cleanup_resource_db(M) :- retractall_fact(resource_db(_, M)).
cleanup_resource_default_db(M) :- retractall_fact(resource_default_db(_, M)).

resdefs_sentence_tr(0, 0, M) :-
	!,
	retractall_fact(def_ru_db(_,  M)),
	retractall_fact(def_gru_db(_, M)),
	cleanup_resource_db(M),
	cleanup_resource_default_db(M).
resdefs_sentence_tr((:- resource(Resource)), (:- resource(Resource)), M) :-
	!,
	assert_resource(Resource, M).
% -- Keep track of cost declarations to put defaults at the end if needed
% (may still need to do something about global resources)
resdefs_sentence_tr((:- head_cost(App,Resource,Default)),
	            (:- head_cost(App,Resource,Default)),M) :-
	!,
	check_resource(Resource, M),
	assert_resource_default(head_cost(App,Resource,Default), M).
resdefs_sentence_tr((:- literal_cost(App,Resource,Default)),
	            (:- literal_cost(App,Resource,Default)),M) :-
	!,
	check_resource(Resource, M),
	assert_resource_default(literal_cost(App,Resource,Default), M).
resdefs_sentence_tr((:- default_cost(App,Resource,Default)),
	            (:- default_cost(App,Resource,Default)),M) :-
	!,
	check_resource(Resource, M),
	assert_resource_default(default_cost(App,Resource,Default), M).
resdefs_sentence_tr((:- trust_default(+cost(App,Resource,Default))),
	            (:- trust_default(+cost(App,Resource,Default))),M) :-
	!,
	check_resource(Resource, M),
	assert_resource_default(trust_default(+cost(App,Resource,Default)), M).
% -------------
resdefs_sentence_tr(
	    (resource_usage(Res, Value) :- Body),
	    (resource_usage(Res, Value) :- Body), M) :-
	!,
	check_resource(Res, M),
	assertz_fact(def_ru_db(Res, M)).
resdefs_sentence_tr(
	    (global_resource_usage(Res, Value) :- Body),
	    (global_resource_usage(Res, Value) :- Body), M) :-
	!,
	check_resource(Res, M),
	assertz_fact(def_gru_db(Res, M)).
resdefs_sentence_tr(end_of_file, DRUL, M) :-
	issue_missing_default_costs(DRUL,RUL,M), 
        !,
	findall('$def$ru'(Res), retract_fact(def_ru_db(Res, M)), RUL, GRUL),
	findall('$def$gru'(Res), retract_fact(def_gru_db(Res, M)), GRUL,
	        [end_of_file]).

check_resource(Res, M) :-
	\+ \+ resource_db(Res, M) ->
	true
    ;
	location(Loc),
	error_message(Loc, "Resource ~w not defined.", [Res]).

% -- Issue default cost declarations remaining
% (may still need to do something about global resources)
issue_missing_default_costs(Decls0,Decls1,M) :-
	setof(Res,resource_db(Res, M),Resources),
	gather_missing_default_costs(Resources,M,Decls0,Decls1),
	cleanup_resource_default_db(M).

gather_missing_default_costs([],_M,Decls,Decls).
gather_missing_default_costs([Resource|Resources],M,Decls0,Decls7) :-
	gather_missing_default_costs_apps([ub,lb],head_cost,Resource,0,M,Decls0,Decls1),
        gather_missing_default_costs_apps([ub,lb],literal_cost,Resource,0,M,Decls1,Decls2),

        gather_missing_default_costs_apps([ub],default_cost,Resource,inf,M,Decls2,Decls3),
        gather_missing_default_costs_apps([lb],default_cost,Resource,0,M,Decls3,Decls4),

        gather_missing_default_costs_apps([ub],trust_default,Resource,inf,M,Decls4,Decls5),
        gather_missing_default_costs_apps([lb],trust_default,Resource,0,M,Decls5,Decls6),

        gather_missing_default_costs(Resources,M,Decls6,Decls7).

gather_missing_default_costs_apps([],_Decl,_Resource,_Default,_M,Decls,Decls).
gather_missing_default_costs_apps([Approx|Approxs],Decl,Resource,Default,M,Decls0,Decls2) :- 
	gather_missing_default_costs_app(Decl,Approx,Resource,Default,M,Decls0,Decls1),
	gather_missing_default_costs_apps(Approxs,Decl,Resource,Default,M,Decls1,Decls2).

gather_missing_default_costs_app(head_cost,Approx,Resource,Default,M,Decls0,Decls1) :- 
	!,
	( resource_default_db(head_cost(Approx,Resource,_),M) -> Decls1 = Decls0
	; Decls0 = [ (:- head_cost(Approx,Resource,Default)) | Decls1 ] ).
gather_missing_default_costs_app(literal_cost,Approx,Resource,Default,M,Decls0,Decls1) :- 
	!,
	( resource_default_db(literal_cost(Approx,Resource,_),M) -> Decls1 = Decls0
	; Decls0 = [ (:- literal_cost(Approx,Resource,Default)) | Decls1 ] ).
gather_missing_default_costs_app(default_cost,Approx,Resource,Default,M,Decls0,Decls1) :- 
	!,
	( resource_default_db(default_cost(Approx,Resource,_),M) -> Decls1 = Decls0
	; Decls0 = [ (:- default_cost(Approx,Resource,Default)) | Decls1 ] ).
gather_missing_default_costs_app(trust_default,Approx,Resource,Default,M,Decls0,Decls1) :- 
	( resource_default_db(trust_default(+cost(Approx,Resource,_)),M) -> Decls1 = Decls0
	; Decls0 = [ (:- trust_default(+cost(Approx,Resource,Default))) | Decls1 ] ).

% -------------

expand_cost(Goal, Rel, Ap, Type, Res, CostFunction,
	    cost(Goal, Rel, Ap, Type, Res, InputDataFuncs, CostFunctionN),
	    M) :-
	check_resource(Res, M),
	expand_cf(CostFunction, InputDataFuncs, CostFunctionN).

resdefs_goal_tr(end_of_file, _, M) :-
	!,
	cleanup_resource_db(M),
	cleanup_resource_default_db(M).
%
resdefs_goal_tr(cost(Goal, Ap, Res, CostFunction), Cost, M) :-
	expand_cost(Goal, abs, Ap, call, Res, CostFunction, Cost, M).
resdefs_goal_tr(cost(Goal, Ap, Type, Res, CostFunction), Cost, M) :-
	expand_cost(Goal, abs, Ap, Type, Res, CostFunction, Cost, M).
%
resdefs_goal_tr(rel_cost(Goal, Ap, Res, CostFunction), Cost, M) :-
	expand_cost(Goal, rel, Ap, call, Res, CostFunction, Cost, M).
resdefs_goal_tr(rel_cost(Goal, Ap, Type, Res, CostFunction), Cost, M) :-
	expand_cost(Goal, rel, Ap, Type, Res, CostFunction, Cost, M).
%
resdefs_goal_tr(intervals(Size, Intervals),
	    (intervals(Goal, V, Intervals)), _) :-
	compound_size(Size, V, Goal).
