:- module(ctchecks_plot,_,[assertions,regtypes,datafacts]).

:- doc(title, "Module for Plotting Cost Functions").
:- doc(author, "Luthfi Darmawan").
:- doc(summary, "Cost function plots to provide a graphical insight of resource 
	usage verification.").
:- doc(module, "
- main predicate for emacs interaction is assertion_current_line(File, LineNum).
  call it from the ciaopp driver by adding this code in the ciaopp driver
  gcheck(File, LineNum):-
	assertion_current_line(File, LineNum).
  Calling this code in ciaopp driver is mandatory in the case of
  verifying input file against analysis result database,
  thus  without using output file.
- main predicate for entra demo is entra_plot/2
- needs 'gnuplot' for plotting and 'imagemagick' for montaging images
- using assrt_lib there was a problem with assertion_read/9 in mac,
  then we switch to p_unit. Some artefacts are left for further examination.
- for interactive plot, terminal x11 breaks ciao in linux, use qt instead,
  meanwhile in mac x11 seems fine.
- interactive terminal can only hold one plot, user must choose plot_type,
  however we can have two terminal by running gnuplot in two separate executions
- p_unit:preprocessing_unit/3 takes major execution time,
  it was the same with assrt_lib:get_code_and_related_assertion/5
- ciao32 in linux32 is noticeably faster than in linux64
- use full file path, don't use abbreviation such as ~/
- simple rule about bound and skip for a cheap and smooth graphic:
  skip = max-min/100. Thus it consists only 100 points.

- to use this module in emacs interface to plot an input file
  against analysis result:
  1. load a module
  2. do the analysis
  3. call gcheck(File, LineNumber) from ciaopp top level
  4. plot_assrt_input predicate will do the work

- to use this module in emacs interface to plot output file of
  analysis result:
  1. call gcheck(File, LineNumber) from ciaopp top level
  2. plot_assrt_output predicate will do the work
").

:- use_package(ciaopp(p_unit/p_unit_argnames)).

:- use_module(engine(stream_basic)).
:- use_module(library(format)).
:- use_module(library(lists),[member/2, append/3]).
:- use_module(ciaopp(p_unit/itf_db), [curr_file/2]).
:- use_module(ciaopp(infer),   [get_info/5]). %get analysis result
:- use_module(ciaopp(p_unit), [preprocessing_unit/3]).
:- use_module(ciaopp(p_unit), [get_assertion/2]).
:- use_module(ciaopp(p_unit/assrt_db),[assertion_read/9, cleanup_assrt_db/0, assertion_body/7]).
:- use_module(library(aggregates), [setof/3, '^'/2]).
% :- use_module(library(assertions/assrt_lib),
% 	[cleanup_code_and_related_assertions/0, get_code_and_related_assertions/5, assertion_read/9,assertion_body/7]). %we encountered problem when using this library then we switch to p_unit
:- use_module(library(terms_vars)).
:- use_module(domain(resources/plai2infer), [replace_term/3]).
:- use_module(library(llists), [flatten/2]).
:- use_module(library(system), [get_tmp_dir/1]).
:- use_module(library(pathnames), [path_concat/3]).
%:- use_module(library(system), [system/1]). %dont use system/1, use process_call/3 instead
:- use_module(library(process), [process_call/3]).
:- use_module(library(glob)).
:- use_module(library(iso_char), [char_code/2]).

%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
%CONFIGURATIONS
:- data plot_id/1.

%plot_config(id, 'value'), set of default configurations
plot_config(min_x, 0).
plot_config(max_x, 100).
plot_config(skip, 1).
plot_config(out_file, 'ciaoplot').
plot_config(tmp_path, Tmp):- get_tmp_dir(Tmp).
plot_config(terminal, jpg).

%configuration default_value other_values
% interactive  no            yes
% plot_all     no            yes
% plot_type    comparison    percentage
%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
% main predicates
%------------------------------------------------------------------------------
% Plot cost function of an assertion on certain line
%------------------------------------------------------------------------------
assertion_current_line(File, Line):-
	init_plot_id,
	(
	    curr_file(File,_) -> 
	    sanitize_config([terminal(qt), interactive(yes), plot_type(comparison)], PlotConfig),
	    plot_assrt_input(Line, PlotConfig) %check from input file
	;
	    sanitize_config([terminal(qt), interactive(yes), plot_type(comparison)], PlotConfig),
	    plot_assrt_output(File,Line,PlotConfig) %check from output file
	).

%------------------------------------------------------------------------------
% Plot cost function of _all_ assertion on output of analysis
%------------------------------------------------------------------------------
entra_plot(File, PlotConfig):-
	init_plot_id,
	sanitize_config(PlotConfig, PlotConfig1),!,
%	cleanup_tmp_dir(PlotConfig1), %error
	entra_output_file_assertion_plot(File, PlotConfig1).

%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
%------------------------------------------------------------------------------
%Write a gnuplot template script file into temporary folder TmpPath 
% parameters are given as a list
%------------------------------------------------------------------------------
write_general_template(ScriptStream, Config):-
	member(min_x(Xmin), Config),
	member(max_x(Xmax), Config),
	member(out_file(OutputFile), Config),
	member(terminal(Terminal), Config),
%	truecolor option is mandatory for transparency
	(
	    Terminal = jpg ->
	    format(ScriptStream, "set terminal jpeg truecolor nocrop enhanced size 1024 font \"arial,12\"~n",[])
	;
	    Terminal = png ->
	    format(ScriptStream, "set terminal png truecolor nocrop enhanced size 1024 font \"arial,12\"~n",[]) 
	;
	    Terminal = qt ->
	    format(ScriptStream, "set terminal qt persist~n",[])
	;
	    format(ScriptStream, "set terminal x11 persist~n",[])
	),
	incr_plot_id, plot_id(PlotId),
	(
	    get_out_ext(Terminal, FileExt) ->
	    format(ScriptStream, "set output '~a.~d.~a'~n",[OutputFile, PlotId, FileExt])
	;
	    true
	),
	format(ScriptStream, "set key left~n",[]),
	format(ScriptStream, "set style data lines~n",[]),
	format(ScriptStream, "set xrange [ ~d : ~d ] noreverse nowriteback~n",[Xmin,Xmax]),
	format(ScriptStream, "set yrange [ 0 : * ] noreverse nowriteback~n",[]).

write_plot_template(Config):-
	member(tmp_path(Tmp), Config),
	member(pred(Pred0), Config),
	simplify_pred(Pred0,Pred),
	member(slb(Slb), Config),
	member(sub(Sub), Config),
	member(clb(Clb), Config),
	member(cub(Cub), Config),
	path_concat(Tmp, 'ciaoplot.dem', ScriptFile),
	open(ScriptFile, 'write', ScriptStream),
	write_general_template(ScriptStream, Config),	
	format(ScriptStream, "set title \"Comparison Between Specification and Cost of ~p\"~n",[Pred]),
	format(ScriptStream, "plot \\~n",[]),
	format(ScriptStream, "'spec.dat' u 1:2:3 \"%lf %lf %lf\" w filledcu clo fs trans soli 0.3 lc rgb \"green\" t \"Spec:~p; ~p\", \\~n",[Slb, Sub]),
	format(ScriptStream, "'cost.dat' u 1:2:3 \"%lf %lf %lf\" w filledcu clo fs trans soli 0.3 lc rgb \"red\" t \"Cost:~p; ~p\"~n",[Clb, Cub]),
	close(ScriptStream).

write_plot_template_percentage(Config):-
	member(tmp_path(Tmp), Config),
	member(pred(Pred0), Config),
	simplify_pred(Pred0,Pred),
	member(sub(Sub), Config),
	member(cub(Cub), Config),	
	path_concat(Tmp, 'percentage.dem', ScriptFile),
	open(ScriptFile, 'write', ScriptStream),
	write_general_template(ScriptStream, Config),
	format(ScriptStream, "set title \"Percentage Upperbound Difference of Specification and Cost of ~p\"~n",[Pred]),
	format(ScriptStream, "plot \\~n",[]),
	format(ScriptStream, "'percentage.dat' u 1:2:3 \"%lf %lf %lf\" w filledcu clo fs trans soli 0.3 lc rgb \"green\" t \"(~p)-(~p)\"~n",[Sub,Cub]),
	close(ScriptStream).

write_plot_template_true(Config):-
	member(tmp_path(Tmp), Config),
	member(pred(Pred0), Config),
	simplify_pred(Pred0,Pred),
	member(clb(Clb), Config),
	member(cub(Cub), Config),	
	path_concat(Tmp, 'true.dem', ScriptFile),
	open(ScriptFile, 'write', ScriptStream),
	write_general_template(ScriptStream, Config),
	format(ScriptStream, "set title \"Cost Lowerbound and Upperbound of ~p\"~n",[Pred]),
	format(ScriptStream, "plot \\~n",[]),
	format(ScriptStream, "'true.dat' u 1:2:3 \"%lf %lf %lf\" w filledcu clo fs trans soli 0.3 lc rgb \"green\" t \"lb:~p; ub:~p\"~n",[Clb, Cub]),
	close(ScriptStream).
 
%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
%                    PLOT FROM OUTPUT FILE
%-------------------------------------------------------------------------------
% processing an output of an analysis
%-------------------------------------------------------------------------------
plot_assrt_output(File,Line, PlotConfig):- %don't use ~, use full path w/o abbrv
	(
	    % avoid reread the file and keep adding same things in db
	    assertion_read(_,Module,_,_,_,_,File,_,_) ->
	    true
	;
	    % assrt_lib artefact
	    % get_code_and_related_assertions(File, Module,_Filename, _Suffix, _Base)
	    preprocessing_unit([File],[Module],_E)
	),!,
	process_check_assertion_cost(File, Line, Module, PlotConfig).

%-------------------------------------------------------------------------------
% processing an output of an analysis and force reload it even though it is
% already loaded
%-------------------------------------------------------------------------------
plot_assrt_output_reload(File, Line, PlotConfig):- %force delete previous database and reread
	% assrt_lib artefact
	% retractall_fact(assertion_read(_,_,_,_,_,_,File,_,_)),
	% cleanup_code_and_related_assertions,
	% get_code_and_related_assertions(File, Module,_Filename, _Suffix, _Base),
	cleanup_assrt_db,
	preprocessing_unit([File],[Module],_E),!,	
	process_check_assertion_cost(File, Line, Module, PlotConfig).

%-------------------------------------------------------------------------------
% reading file and process the assertion in it
%-------------------------------------------------------------------------------
entra_output_file_assertion_plot(File, PlotConfig):-
	cleanup_assrt_db,
	preprocessing_unit([File],[Module],_E),!,
	%delete previous images
	member(terminal(Terminal), PlotConfig),
	(
	    get_out_ext(Terminal, FileExt) ->
	    member(tmp_path(Tmp), PlotConfig),
	    member(out_file(OutFile), PlotConfig),
	    atom_concat(OutFile, '.*.', Ofb),
	    atom_concat(Ofb, FileExt, FileGlob), %eg ciaoplot.*.jpg
	    glob(Tmp, FileGlob, FilesDel),
	    (
		FilesDel = [] ->
		true
	    ;
		process_call(path(rm), ['-v'|FilesDel], [cwd(Tmp), status(_)]) %ignore status
	    )
	;
	    true % terminal output not to a file
	),
	(
	    process_check_assertion_cost(File, -1, Module, PlotConfig) ->
	    true
	;
	    format("Doesn't plot CHECK assertions.~n",[]),
	    (
		process_entry_assertion_cost(File, Module, PlotConfig) ->
		true
	    ;
		format("Doesn't plot ENTRY related assertions.~n",[])
	    )
	),
	(
	    member(plot_all(yes), PlotConfig),
	    process_true_assertion_cost(File, -1, Module, PlotConfig) ->
	    true
	;
	    format("Doesn't plot TRUE assertions.~n",[])
	),
	%montage images
	(
	    get_out_ext(Terminal, FileExt) ->
	    atom_concat(OutFile, '.', Ofo),
	    atom_concat(Ofo, FileExt, FileOutput),
	    process_call(path(montage), [FileGlob,'-tile', '1x', '-geometry', '1024', FileOutput], [cwd(Tmp), status(DrawStat)]),
	    (
		DrawStat = 0 ->
		format("All graphs in file ~a/~a.~n",[Tmp, FileOutput])
	    ;
		format("Montage graphs failed.",[])
	    )
	;
	    true % terminal output not to a file
	).

%-------------------------------------------------------------------------------
% show cost assertion of predicate related with entry assertion
%-------------------------------------------------------------------------------
process_entry_assertion_cost(File, Module, PlotConfig):-
	(
	    assertion_read(Pred, Module, true, entry, _Body,_Dic,File, _LB,_LE)
	;
	    format("Did not find entry assertion.~n",[]),
	    fail
	),
	(
	    process_entry_related_assertion_cost(Pred,File, Module, PlotConfig)
	;
	    format("Processing assertion of predicate related with entry assertion failed.~n",[]),
	    fail
	).

%-------------------------------------------------------------------------------
% collect all true assertion related with entry assertion
%-------------------------------------------------------------------------------
process_entry_related_assertion_cost(MPred, File, Module, PlotConfig):-
	setof((CType,CBody,CDic,CLB,CLE), MPred^assertion_read(MPred,Module,true,CType,CBody,CDic,File,CLB,CLE), CBodies),
	process_true_assertions(CBodies, CBodies, PlotConfig).

%-------------------------------------------------------------------------------
% compare check with true assertion 
%-------------------------------------------------------------------------------
process_check_assertion_cost(File, Line, Module, PlotConfig):-	
	%	assertion_read(Pred,Module,_Stat,_Type,Body,_Dic,File,LB,LE),
	%	LB=<Line, Line=<LE,
	(
	    Line = -1 -> %for ENTRA
	    setof((Pred,Body), Dic^LB^LE^(assertion_read(Pred, Module, Stat, Type,
	    Body,Dic,File, LB,LE), (Stat=check;Stat=checked;Stat=false),
	    (Type=pred; Type=comp)), Result) %creates choice point
	    ;	    
	    setof((Pred,Body), Dic^(assertion_read(Pred, Module, Stat, Type,
	    Body,Dic,File, LB,LE), (Type=pred; Type=comp),
	    LB=<Line, Line=<LE,(Stat=check;Stat=checked;Stat=false)), Result)
	),
	process_assertions(Result, File, Module, PlotConfig).

%-------------------------------------------------------------------------------
% compare each check assertion then draw the plot
%-------------------------------------------------------------------------------
process_assertions([],_F,_M,_C).
process_assertions([(MPred,Body)|NextAssertions], File, Module, PlotConfig):-
	assertion_body(MPred,_DP,_CP,_AP,GP,_CO,Body),
	format("The predicate: ~p~n", [MPred]),
	get_orig_pred(MPred, Pred), %it creates choice point either MPred or just Pred
	(  %from the assrt_lib the cost w/o resources_props quantifier
	    (member('resources_props:cost'(Pred,_,_UBLB,_,ResType, _, _, _Cost), GP); member(cost(Pred,_UBLB,ResType, _Cost), GP)) -> %old resources style assertion
	    (
		(member('resources_props:cost'(Pred,_,lb,_,ResType, RefAClb,_,AClb), GP); member(cost(Pred,lb,ResType, AClb), GP)) ->
		(
		    (member('resources_props:cost'(Pred,_,ub,_,ResType, RefACub,_,ACub), GP); member(cost(Pred,ub,ResType, ACub), GP)) ->
		     format("Cost on assertion: ~p ~p~n", [lb(AClb), ub(ACub)]),
		     (
			 cost2pred_var(Pred, AClb, RefAClb, Slb),
			 cost2pred_var(Pred, ACub, RefACub, Sub)
		     ;
			 Slb = AClb,
			 Sub = ACub
		     )
		;
		    format("Cost on assertion: ~p ~p~n", [lb(AClb), ub(none)]),
		    (
			cost2pred_var(Pred, AClb, RefAClb, Slb)
		    ;
			Slb = AClb
		    ),
		    Sub = none
		)
	    ; %lb doesn't exist
		(member('resources_props:cost'(Pred,_,ub,_,ResType, _,_,ACub), GP); member(cost(Pred,ub,ResType, ACub), GP)) ->
		format("Cost on assertion: ~p ~p~n", [lb(none), ub(ACub)]),
		Slb = none,
		(
		    cost2pred_var(Pred, ACub, RefACub, Sub)
		;
		    Sub = ACub
		)
            ;
	        format("ERROR:Unrecognize bound neither ub/lb~n",[])
            )
	;
	    member(costb(Pred, ResType, AClb, ACub), GP) -> %res_plai style assertion
	    format("Cost on assertion: ~p ~p~n", [AClb, ACub]),
	    Slb = AClb,
	    Sub = ACub
	;
	    member('native_props:costb'(Pred, ResType, AClb, ACub), GP) -> %res_plai style assertion
	    format("Cost on assertion: ~p ~p~n", [AClb, ACub]),
	    Slb = AClb,
	    Sub = ACub
	),
	%taking information from  true assertion
	%	assertion_read(Pred,Module,true,_Type,Body,_Dic,File,_LB,_LE).
	setof((CType,CBody,CDic,CLB,CLE), assertion_read(MPred,Module,true,CType,CBody,CDic,File,CLB,CLE), CBodies),
	(
	    get_cost_entra_rp(CBodies, CBodies, ResType, CCost) -> %res_plai style assertion for entra
	    format("Cost from  analysis: ~p~n", CCost),
	    (
		CCost = (lb(Clb), ub(Cub))
	    ;
		CCost = (Clb, Cub)
	    )	    
	;
	    get_cost_true_rp(CBodies, ResType, CCost) -> %res_plai style assertion
	    format("Cost from  analysis: ~p~n", CCost),
	    (
		CCost = (lb(Clb), ub(Cub))
	    ;
%		replace_term(CCost, [(ub(X), X), (lb(X),X)], NCCost),
		CCost = (Clb, Cub)
	    )
	;
	    get_cost_true_cl(CBodies, ResType, CCost) -> %old res style assertion
	    format("Cost from  analysis: ~p~n", CCost),
	    get_cub_clb(CCost, Clb, Cub)
	),
	QuadCost = quadcost(Slb, Sub, Clb, Cub), %quadcost(Slb, Sub, Clb, Cub)
	format("Conclusion: ~n Slb: ~p~n Sub: ~p~n Clb: ~p~n Cub: ~p~n", [Slb, Sub, Clb, Cub]),
	plot_data(MPred, QuadCost, PlotConfig),
	process_assertions(NextAssertions, File, Module, PlotConfig).


%-------------------------------------------------------------------------------
% make cost expression in term of input argument
%-------------------------------------------------------------------------------
cost2pred_var(Pred, OutCost, Ref, Cost):-
	varset(Pred, InputVars),
	varset(OutCost, OutputVars),
	out2in_vars_cl(OutputVars, InputVars, Ref, OutInVars),
	replace_term(OutCost, OutInVars, Cost).

%out2in_vars_cl(OutputVars, InputVars, Ref, OutInVars):- %old resources style
out2in_vars_cl([],_,_,[]).
out2in_vars_cl([OutV|OutputVars], InputVars, Ref, [OIV|OutInVars]):-
	in_input_var_cl(OutV, InputVars, Ref, OIV),
	out2in_vars_cl(OutputVars, InputVars, Ref, OutInVars).
	
 in_input_var_cl(OutV, [InV|InputVars], Ref, OutInPair):- 
	(
	    member('lists:length'(InV1, OutV), Ref),
	    InV == InV1->
	    OutInPair = (OutV, length(InV))
	;
	    member('term_basic:='(InV1, OutV), Ref),
	    InV == InV1->
	    OutInPair = (OutV, int(InV))
	;
	    in_input_var(OutV, InputVars, Ref, OutInPair)
	).

%-------------------------------------------------------------------------------
% collect all true assertion only 
%-------------------------------------------------------------------------------
process_true_assertion_cost(File, _Line, Module, PlotConfig):-
	setof((CType,CBody,CDic,CLB,CLE), MPred^assertion_read(MPred,Module,true,CType,CBody,CDic,File,CLB,CLE), CBodies),
	process_true_assertions(CBodies, CBodies, PlotConfig).

%-------------------------------------------------------------------------------
% process every comp assertion cost
% this code is very similar to get_cost_entra_rp/4, but this one continues
% after finds a cost function
%-------------------------------------------------------------------------------
process_true_assertions([],_,_). 
process_true_assertions([(comp,Body,Dic,LB,LE)|RestAssrt], Assrt, PlotConfig):-!,
	assertion_body(MPred,_DP,_CP,_AP,GP,_CO,Body),
	get_orig_pred(MPred, Pred),
	(	    
	    %	    (member('native_props:costb'(Pred, Res, AClb, ACub), GP); member(costb(Pred, Res, AClb, ACub), GP))->
	    member('native_props:costb'(Pred, _Res, AClb, ACub), GP) ->	    
	    varset((AClb,ACub), CompVars),
	    get_symbols(CompVars, Dic, Symbols, VSPairs),
	    member((success, BodyS, DicS, LB, LE), Assrt),
	    get_vars(Symbols, VSPairs, DicS, SucVars, VSVTriples), %VSV = (VarComp, Symbol, VarSucc)
	    assertion_body(MPred,_DPS,_CPS,APS,_GPS,_COS,BodyS),
	    varset(MPred, InputVars),
	    out2in_vars(SucVars, InputVars, APS, SucOutInVars),
	    succ_to_comp(SucOutInVars, VSVTriples, OutInVars), %connect variable from success assertion back to comp assertion
	    replace_term((AClb,ACub), OutInVars, CCost),
	    format("Cost from  analysis: ~p~n", CCost),
	    (
		CCost = (lb(Clb), ub(Cub))
	    ;
		CCost = (Clb, Cub)
	    ),
	    %plot
	    plot_data_true(Pred, quadcost(none,none,Clb,Cub), PlotConfig)
	;
	    true
	),
	process_true_assertions(RestAssrt, Assrt, PlotConfig).

process_true_assertions([_|RestAssrt], Assrt, PlotConfig):-
	process_true_assertions(RestAssrt, Assrt, PlotConfig).

get_cub_clb(CCost, Clb, Cub):-
	(
	    member(ub(Lb), CCost) ->
	    Clb = Lb
	;
	    Clb = none
	),
	(
	    member(lb(Ub), CCost) ->
	    Cub = Ub
	;
	    Cub = none
	).

cost_lb_ub_from_infer(FlatCost, Res, Clb, Cub):-
	(
	    member(cost(lb, Res, Clb), FlatCost)->
	    true
	;
	    Clb = none
	),
	(
	    member(cost(ub, Res, Cub), FlatCost)->
	    true
	;
	    Cub = none
	).

%old res style assertion
get_cost_true_cl([],_,[]).
get_cost_true_cl([(_CType,Body,_CDic,_CLB,_CLE)|Bs], Res, Cost):-
	assertion_body(Pred,_DP,_CP,_AP,GP,_CO,Body),
	(
	    member('resources_props:cost'(Pred,_,lb,_,ResType, RefAClb,_,AClb), GP)-> %using p_unit
	    cost2pred_var(Pred, AClb, RefAClb, Clb),
	    Cost = [lb(Clb)|Cost1]
	;
	    member('resources_props:cost'(Pred,_,ub,_,ResType, RefACub,_,ACub), GP)-> %using p_unit
	    cost2pred_var(Pred, ACub, RefACub, Cub),
	    Cost = [ub(Cub)|Cost1]
	;
	    GP = [GPe],
	    member(cost(Pred, lb, Res, AClb), GPe)-> % p_unit old res
	    Cost = [lb(AClb)|Cost1]
	;
	    GP = [GPe],
	    member(cost(Pred, ub, Res, ACub), GPe)-> %p_unit old res
	    Cost = [ub(ACub)|Cost1]
	;
	    member(cost(Pred, lb, Res, AClb), GP)-> %using assrt_lib
	    Cost = [lb(AClb)|Cost1]
	;
	    member(cost(Pred, ub, Res, ACub), GP)-> %using assrt_lib
	    Cost = [ub(ACub)|Cost1]
	;
	    Cost = Cost1
	),
	get_cost_true_cl(Bs, Res, Cost1).

%res_plai style assertion
%get_cost_true_rp([],_,[]). let it fails
get_cost_true_rp([(_CType,Body,_CDic,_CLB,_CLE)|Bs], Res, Cost):-
	assertion_body(MPred,_DP,_CP,AP,GP,_CO,Body),
	get_orig_pred(MPred, Pred), %remove module quantifier from MPred
	(
	    (member('native_props:costb'(Pred, Res, AClb, ACub), GP); member(costb(Pred, Res, AClb, ACub), GP))->
	    varset(Pred, InputVars),
	    varset((AClb,ACub), OutputVars),
	    out2in_vars(OutputVars, InputVars, AP, OutInVars),
	    replace_term((AClb,ACub), OutInVars, Cost)
	;
	    get_cost_true_rp(Bs, Res, Cost)
	).

%get_cost_entra_rp([],_,[]). let it fails
get_cost_entra_rp([(comp,Body,Dic,LB,LE)|RestAssrt], Assrt, Res, Cost):-!,
	assertion_body(MPred,_DP,_CP,_AP,GP,_CO,Body),
	get_orig_pred(MPred, Pred),
	(	    
	    %	    (member('native_props:costb'(Pred, Res, AClb, ACub), GP); member(costb(Pred, Res, AClb, ACub), GP))->
	    member('native_props:costb'(Pred, Res, AClb, ACub), GP) ->	    
	    varset((AClb,ACub), CompVars),
	    get_symbols(CompVars, Dic, Symbols, VSPairs),
	    member((success, BodyS, DicS, LB, LE), Assrt),
	    get_vars(Symbols, VSPairs, DicS, SucVars, VSVTriples), %VSV = (VarComp, Symbol, VarSucc)
	    assertion_body(MPredS,_DPS,_CPS,APS,_GPS,_COS,BodyS),
	    varset(MPredS, InputVars),
	    out2in_vars(SucVars, InputVars, APS, SucOutInVars),
	    succ_to_comp(SucOutInVars, VSVTriples, OutInVars), %connect variable from success assertion back to comp assertion
	    replace_term((AClb,ACub), OutInVars, Cost)
	;
	    get_cost_entra_rp(RestAssrt, Assrt, Res, Cost)
	).
get_cost_entra_rp([_|RestAssrt], Assrt, Res, Cost):-
	get_cost_entra_rp(RestAssrt, Assrt, Res, Cost).

%connect variable of success assertion (SOutV) back to variable of comp assertion (COutV)
succ_to_comp([],_,[]).
succ_to_comp([(SOutV, InV)|SucOutInVars], VSVTriples, [(COutV, InV)|OutInVars]):-
	triple_member((COutV, _Symb, SOutV), VSVTriples),
	succ_to_comp(SucOutInVars, VSVTriples, OutInVars).

% strong triple membership test that uses equality test '==' on 3rd part of triple
triple_member((X,_,Y),[(X,_,Y1)|_]):- Y == Y1.
triple_member(T,[_|Ls]):- triple_member(T, Ls).

get_symbols([],_,[],[]).
get_symbols([Var|Vars], Dic, [Symbol|Symbols], [(Var, Symbol)|Pairs]):-
	member(Symbol=Var1, Dic),
	Var == Var1,
	get_symbols(Vars, Dic, Symbols, Pairs).

get_vars([],_,_,[],[]).
get_vars([Symbol|Symbols], [(PVar, Symbol)|Pairs], Dic, [Var|Vars], [(PVar, Symbol, Var)|Triples]):-
	member(Symbol=Var, Dic),
	get_vars(Symbols, Pairs, Dic, Vars, Triples).

out2in_vars([],_,_,[]).
out2in_vars([OutV|OutputVars], InputVars, AP, [OIV|OutInVars]):-
	in_input_var(OutV, InputVars, AP, OIV),
	out2in_vars(OutputVars, InputVars, AP, OutInVars).

% the == comparison is mandatory to guarantee the variables are the same.
% using only '=' will performs unification with arbitrary variable
%in_input_var(OutV,[A,B,C],[list(A,num),num(B),num(C),rsize(A,basic_props:list(K,L,num(M,N))),rsize(B,num(O,P)),rsize(C,num(Q,OutV))],OIV)
in_input_var(OutV, [InV|InputVars], AP, OutInPair):- 
	(
	    (
		member('native_props:rsize'(InV1, Prop),AP)
	    ;
		member(rsize(InV1, Prop),AP)
	    ),
	    InV == InV1->
	    (
		Prop = 'basic_props:list'(OutV1,_,_), OutV == OutV1 ->
		OutInPair = (OutV, lb(length(InV)))
	    ;
		Prop = 'basic_props:list'(_,OutV1,_), OutV == OutV1 ->
		OutInPair = (OutV, ub(length(InV)))
	    ;
		(
		    Prop = num(_,OutV1);
		    Prop = num(OutV1,_)
		), OutV == OutV1 ->
		OutInPair = (OutV, int(InV))
	    ;
		in_input_var(OutV, InputVars, AP, OutInPair)
	    )
	;
	    in_input_var(OutV, InputVars, AP, OutInPair)
	).

%-------------------------------------------------------------------------------
% plot data from user specification and from analysis
%-------------------------------------------------------------------------------
plot_data(Pred, QuadCost, PlotConfig):-
	(
	    varset(QuadCost, [Var]) -> %single var only
	    get_arg_id(Pred, Var, Id),
	    simplify_cost(QuadCost, Var, $(Id), QuadCost1),
	    write_plot_data(Pred, QuadCost1, PlotConfig), %plot comparison between all spec vs cost
	    write_plot_percentage(Pred, QuadCost1, PlotConfig) %plot comparison percentage ub only
	;
	    format("More than one variable, no ND plot!~n",[])
	).

%-------------------------------------------------------------------------------
% plot data from analysis only
%-------------------------------------------------------------------------------
plot_data_true(Pred, QuadCost, PlotConfig):-
	(
	    varset(QuadCost, [Var]) -> %single var only
	    get_arg_id(Pred, Var, Id),
	    simplify_cost(QuadCost, Var, $(Id), QuadCost1),
	    write_plot_data_true(Pred, QuadCost1, PlotConfig) %plot analysis cost
	;
	    format("More than one variable, no ND plot!~n",[])
	).

%-------------------------------------------------------------------------------
%plot comparison between spec ub,lb vs cost ub,lb
%-------------------------------------------------------------------------------
%assume: Clb and Cub never none in res_plai and resources analysis
%handling non exists cost
write_plot_data(Pred, quadcost(none, Sub, Clb,Cub), PlotConfig):-
	Slb = 0,
	plot_comparison(Pred, Slb, Sub, Clb, Cub, PlotConfig).
%write_plot_data(Pred, quadcost(Slb, none, Clb,Cub)):-
write_plot_data(_Pred, quadcost(_Slb, none, _Clb,_Cub), _PlotConfig):-
	format("ERROR: Sub=none is not implemented", []).
% 	Sub = ?, ??? Let me think
%       It should be something like plot '+' u 1:($1**2) w filledcu above fs trans soli 0.3 lc rgb "red"   title "Spec:"
%       that plot whole area above Slb
write_plot_data(Pred, quadcost(Slb, Sub, Clb,Cub), PlotConfig):-
	plot_comparison(Pred, Slb, Sub, Clb, Cub, PlotConfig).

%-------------------------------------------------------------------------------
%plot percentage comparison between spec ub vs cost ub
%-------------------------------------------------------------------------------
write_plot_percentage(Pred, quadcost(_Slb, Sub, _Clb,Cub), PlotConfig):-
	Sub \= none, 
	Cub \= none,
	plot_percentage(Pred, Sub, Cub, PlotConfig).

%-------------------------------------------------------------------------------
% plot cost lb ub
%-------------------------------------------------------------------------------
write_plot_data_true(Pred, quadcost(_Slb, _Sub, Clb,Cub), PlotConfig):-
	Clb \= none, 
	Cub \= none,
	plot_true(Pred, Clb, Cub, PlotConfig).
	
%-------------------------------------------------------------------------------
% determine variable position in the Pred arguments, eg. $(2) -> 2nd argument
%-------------------------------------------------------------------------------
get_arg_id(Pred, Var, Id):-
	Pred =.. [_F|Args],
	loc_args(Args, Var, 0, Id).

loc_args([V|Args], Var, Counter, Id):-
	(
	    Var == V ->
	    Id is Counter + 1
	;
	    Counter1 is Counter + 1,
	    loc_args(Args, Var, Counter1, Id)
	).
 
%-------------------------------------------------------------------------------
% replace size variable with simple $(x) variable
%-------------------------------------------------------------------------------
simplify_cost(Cost, Var, IdVar, SimpCost):-
	replace_term(Cost, [(int(Var),IdVar), (length(Var), IdVar)],SimpCost).

%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
%                    PLOT FROM INPUT FILE

:- use_module(library(aggregates), [findall/3]).

plot_assrt_input(Line, PlotConfig):-
%	Line is Line1 - 1, %emacs line start from 1, meanwhile in ciao it seems start from 0
	curr_file(F,Module),
	format("Current Line Number is ~d ~a ~a~n", [Line,F,Module]),
	findall(As, get_assertionAt(F, Module, Line, As), LA),
	(
	    LA == [] ->
	    format("There is no valid assertion on line ~d~n", [Line])
	;
	    LA = [_Assertion]  -> %one element
	    process_from_infile_assertions(LA, PlotConfig)	
	;
	    format("There are more than one cost assertions on line ~d. We can not draw a plot of multiple cost assertions at once.~n", [Line]),
	    process_from_infile_assertions(LA, PlotConfig)	
	).

get_assertionAt(F, Module, Line, As) :-
	As = as${module=>Module,
	         status=>Stat,
		 locator=> loc(F,LB,LE)},
	get_assertion(_, As),
	LB =< Line, Line =< LE,
	plot_stat(Stat).

plot_stat(check).
plot_stat(checked).
plot_stat(false).

process_from_infile_assertions([], _).
process_from_infile_assertions([A|As], PlotConfig):-
	process_from_infile_one_assertion(A, PlotConfig),
	process_from_infile_assertions(As, PlotConfig).

process_from_infile_one_assertion(Assertion, PlotConfig):-
	format("~nThe assertion ~n~p", [Assertion]),
	Assertion = as${
	    head=>Pred,
	    locator=> loc(_,LB,LE)
	},
	format("Lies on line ~d to ~d~n", [LB,LE]),
	format("The related predicate ~p~n", [Pred]),
	Assertion = as${comp=>Comp},
	(
	    % ------------------ old resources style assertion -----------------
	    member('resources_props:cost'(Pred,_,_UBLB,_,ResType, _, _, _Cost), Comp)->
	     (
		 member('resources_props:cost'(Pred,_,lb,_,ResType, RefAClb,_,AClb), Comp)->
		 (
		     member('resources_props:cost'(Pred,_,ub,_,ResType, RefACub,_,ACub), Comp)->
		     format("Cost on assertion: ~p ~p~n", [lb(AClb), ub(ACub)]),
		     (
			 cost2pred_var(Pred, AClb, RefAClb, Slb),
			 cost2pred_var(Pred, ACub, RefACub, Sub)
		     ;
			 Slb = AClb,
			 Sub = ACub
		     )
		 ;
		     format("Cost on assertion: ~p ~p~n", [lb(AClb), ub(none)]),
		     (
			 cost2pred_var(Pred, AClb, RefAClb, Slb)
		     ;
			 
			 Slb = AClb
		     ),
		     Sub = none
		 )
	     ; %lb doesn't exist
		 member('resources_props:cost'(Pred,_,ub,_,ResType, _,_,ACub), Comp)->
		 format("Cost on assertion: ~p ~p~n", [lb(none), ub(ACub)]),
		 Slb = none,
		 (
		     cost2pred_var(Pred, ACub, RefACub, Sub)
		 ;
		     Sub = ACub
		 )
            ;
	        format("ERROR:Unrecognize bound neither ub/lb~n",[])
            )
        ;
	    % ------------- res_plai style user assertion ----------------------
	    member('native_props:costb'(_Pred,_Res,Slb, Sub), Comp) ->
	    format("The cost part of assertion ~p ~p~n", [Slb, Sub])
	;
	    format("The cost part of assertion does not exist~n",[])
	),
	% we've got the assertion from the file
	% now we get the cost function from analysis
	(
	    setof( LRes,Key^Succ^get_info(resources,pred,Key,Pred,(Succ,LRes)),ResCost) ->
	    flatten(ResCost, FlatCost),
	    cost_lb_ub_from_infer(FlatCost, ResType, Clb, Cub),
	    format("The cost from analysis ~p~n", [FlatCost])
	; %res_plai cost is already converted in this point
	    setof( LRes,Key^Succ^get_info(res_plai,pred,Key,Pred,(Succ,LRes)),ResCost) ->
	    flatten(ResCost, FlatCost),
	    cost_lb_ub_from_infer(FlatCost, ResType, Clb, Cub),
	    format("The cost from analysis ~p~n", [FlatCost])
	;
	    format("The cost from analysis does not exist~n", [])
	),
	QuadCost = quadcost(Slb, Sub, Clb, Cub), %quadcost(Slb, Sub, Clb, Cub)
	format("Conclusion: ~n Slb: ~p~n Sub: ~p~n Clb: ~p~n Cub: ~p~n", [Slb, Sub, Clb, Cub]),
	plot_data(Pred, QuadCost, PlotConfig).

show_information(Assertion, _PlotConfig):- %defunct, but stays here for further examination
	format("~nThe assertion ~n~p", [Assertion]),
	Assertion = as${
	    head=>Goal,
	    locator=> loc(_,LB,LE)
	},
	format("Lies on line ~d to ~d~n", [LB,LE]),
	format("The related predicate ~p~n", [Goal]),
	Assertion = as${comp=>Comp},
	(
	    % ------------------ old resources style assertion -----------------
	    member('resources_props:cost'(_,_,_,_,_,_,_Measure,_ResSpec), Comp) ->
	    (
		member('resources_props:cost'(_,_,ub,_,_,_,Mub,Sub), Comp) ->
		(
		    member('resources_props:cost'(_,_,lb,_,_,_,Mlb,Slb), Comp) ->
		    %ub and lb
		    normalize_cost_ascl(Goal, Mub, Sub, NSub),
		    normalize_cost_ascl(Goal, Mlb, Slb, NSlb)
		;
		    %only ub
		    NSlb = none,
		    normalize_cost_ascl(Goal, Mub, Sub, NSub)
		)
	    ;
		(
		    member('resources_props:cost'(_,_,lb,_,_,_,Mlb,Slb), Comp) ->
		    normalize_cost_ascl(Goal, Mlb, Slb, NSlb),
		    NSub = none
		;
		    format("ERROR:Unrecognize bound neither ub/lb~n",[])
		)		
	    ),
	    format("The normalized cost part of assertion ~p ~p~n", [lb(NSlb), ub(NSub)]),
	    % we've got the assertion from the file
	    % now we get the cost function from analysis
	    (
		setof( LRes,Key^Succ^get_info(resources,pred,Key,Goal,(Succ,LRes)),ResCost) ->
%		format("The cost from analysis ~p~n", [ResCost]),
		%normalize the cost function
		(
		    varset(ResCost, [Var]) -> %single var only
		    get_arg_id(Goal, Var, Id),		    
		    simplify_cost(ResCost, Var, $(Id), NormResCost),
%		    format("The normalized cost from analysis ~p~n", [NormResCost]),
		    flatten(NormResCost, ListCost),
		    (
			member(cost(lb,_, Clb), ListCost),
			member(cost(ub,_, Cub), ListCost) ->
			format("The normalized cost from analysis ~p ~p~n", [lb(Clb), ub(Cub)])
		    ;
			%resource analysis should have both ub lb
			format("It doesn't have resource analysis!~n",[])
		    )
		;
		    format("More than one variable, no ND plot!~n",[])
		)		
	    %compute the plot data
	    % ; %res_plai cost is already converted in this point
	    % 	setof( LRes,Key^Succ^get_info(res_plai,pred,Key,Goal,(Succ,LRes)),ResCost) ->
	    % 	format("The cost from analysis ~p~n", [ResCost])
	    ;		
		format("The cost from analysis does not exist~n", [])
	    )
        ;
	    % ------------- res_plai style user assertion ----------------------
	    member('native_props:costb'(_Pred,_Res,Alb, Aub), Comp) ->
	    format("The cost part of assertion ~p ~p~n", [Alb, Aub]),
	    % we've got the assertion from the file
	    % now we get the cost function from analysis
	    (
	    % 	setof( LRes,Key^Succ^get_info(resources,pred,Key,Goal,(Succ,LRes)),ResCost) ->
	    % 	format("The cost from analysis ~p~n", [ResCost])
	    % ; %res_plai cost is already converted in this point
		setof( LRes,Key^Succ^get_info(res_plai,pred,Key,Goal,(Succ,LRes)),ResCost) ->
		%		format("The cost from analysis ~p~n", [ResCost]),
		(
		    varset(ResCost, [Var]) -> %single var only
		    get_arg_id(Goal, Var, Id),		    
		    simplify_cost(ResCost, Var, $(Id), NormResCost),
		    
		    flatten(NormResCost, ListCost),
		    (
			member(cost(lb,_, Clb), ListCost),
			member(cost(ub,_, Cub), ListCost) ->
			format("The normalized cost from analysis ~p ~p~n", [lb(Clb), ub(Cub)])
		    ;
			%resource analysis should have both ub lb
			format("It doesn't have resource analysis!~n",[])
		    )
		;
		    %resource analysis should have both ub lb
		    format("It doesn't have resource analysis!~n",[])
		)
	    ;		
		format("The cost from analysis does not exist~n", [])
	    )
        ;
	    format("The cost part of assertion does not exist~n",[])
	).


normalize_cost_ascl(Pred, Measure, ResSpec, NormResSpec):-
	(
	    varset(ResSpec, [VarSize]) -> %single var only
	    get_head_var(Measure, VarSize, Var),
	    get_arg_id(Pred, Var, Id),
	    replace_term(ResSpec, [(VarSize, $(Id))], NormResSpec)
	;
	    format("More than one variable, no ND plot!~n",[])
	).

get_head_var([M|Ms], Varsize, Var):-
	(
	    M = length(Var, Vs), Vs == Varsize ->
	    true
	;
	    get_head_var(Ms, Varsize, Var)
	).
%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
%                     WRITING DATA OF FUNCTION INTO FILE
%------------------------------------------------------------------------------
% plot analysis cost
%------------------------------------------------------------------------------
plot_true(Pred, Clb, Cub, PlotConfig):- %direct to gnuplot
	%plot '+' u 1:(2549765):(16652087*$1+5445103) w filledcu clo fs trans soli 0.3 lc rgb "red"   title "Cost: 2.5e+6; 1.7e+7*C+5.4e+6";
	member(interactive(yes), PlotConfig),
	member(terminal(Terminal), PlotConfig),
	member(min_x(MinX), PlotConfig),
	atom_number(MinXAtom, MinX),
	(
	    member(max_x(MaxX), PlotConfig) ->
	    atom_number(MaxXAtom, MaxX)
	;
	    MaxXAtom = '*'
	),
	simplify_pred(Pred, PredAtom),
	ciao2gnuplot(Clb, GClb, GLClb, TClb),
	ciao2gnuplot(Cub, GCub, GLCub, TCub),
	%construct gnuplot command
	merge_atoms([
		       'set terminal ', Terminal, ';',
		       'set yrange [0:*];',
		       'set xrange [', MinXAtom,':',MaxXAtom,'];',
		       'set title "', PredAtom, '"; set key left; ', %config
		       'plot ''+'' u 1:(', GClb, '):(', GCub,
		       ') w filledcu clo fs trans soli 0.3 lc rgb "red"   title "Cost: ',
		       TClb, '; ', TCub, '",',
		       GLClb, ' lc rgb "red" notitle,', GLCub,' lc rgb "red" notitle', %cost
		       '; pause mouse close'
		   ], GnuplotCmd),
	format("Command: ~a~n", [GnuplotCmd]),
	process_call(path(gnuplot), ['-e', GnuplotCmd], [status(DrawStat)]),
	(
	    DrawStat = 0 ->
	    true
	;
	    format("Graph drawing failed.~n",[])
	).

plot_true(Pred, Clb, Cub, PlotConfig):-
	\+ member(interactive(yes), PlotConfig),
	member(min_x(A), PlotConfig),
	member(max_x(B), PlotConfig),
	member(tmp_path(Tmp), PlotConfig),
	member(out_file(FileOut), PlotConfig),
	member(skip(Skip), PlotConfig),
	%additional config for title,legends,etc
	AddPlotConfig = [pred(Pred), clb(Clb), cub(Cub)],
	append(PlotConfig, AddPlotConfig, PlotConfig1), 	
	write_plot_template_true(PlotConfig1),
	path_concat(Tmp, 'true.dat', TrueFile),
	open(TrueFile, 'write', TrueStream),
	format(TrueStream,"#~a ~a ~a~n", ['X', 'Sub', 'Cub']),
	evaluate_true(Clb, Cub, A, B, Skip, TrueStream),
	close(TrueStream),
	process_call(path(gnuplot), ['true.dem'], [cwd(Tmp), status(DrawStat)]),
	(
	    DrawStat = 0 ->
	    plot_id(PlID),
	    member(terminal(Terminal), PlotConfig),
	    (
		get_out_ext(Terminal, ExtF)->
		format("Graph is drawn in ~a/~a.~d.~a~n",[Tmp,FileOut,PlID, ExtF])
	    ;
		true
	    )
	;
	    format("Graph drawing failed.~n",[])
	).
plot_true(_,_,_,_):- %garbage collector, never fails
	format("GC:Doesn't plot true",[]).

%------------------------------------------------------------------------------
% plot percentage difference
%------------------------------------------------------------------------------
plot_percentage(Pred, Sub, Cub, PlotConfig):-
	\+ member(interactive(yes), PlotConfig), %just to prevent choice points
	member(min_x(A), PlotConfig),
	member(max_x(B), PlotConfig),
	member(tmp_path(Tmp), PlotConfig),
	member(out_file(FileOut), PlotConfig),
	member(skip(Skip), PlotConfig),
	%additional config for title,legends,etc
	AddPlotConfig = [pred(Pred), sub(Sub), cub(Cub)],
	append(PlotConfig, AddPlotConfig, PlotConfig1), 
	write_plot_template_percentage(PlotConfig1),
	path_concat(Tmp, 'percentage.dat', PercentageFile),
	open(PercentageFile, 'write', PercentageStream),
	format(PercentageStream,"#~a ~a ~a~n", ['X', 'Sub', 'Cub']),
	evaluate_percentage(Sub, Cub, A, B, Skip, PercentageStream),
	close(PercentageStream),
	process_call(path(gnuplot), ['percentage.dem'], [cwd(Tmp), status(DrawStat)]),
	(
	    DrawStat = 0 ->
	    plot_id(PlID),
	    member(terminal(Terminal), PlotConfig),
	    (
		get_out_ext(Terminal, ExtF)->
		format("Graph is drawn in ~a/~a.~d.~a~n",[Tmp,FileOut,PlID, ExtF])
	    ;
		true
	    )
	;
	    format("Graph drawing failed.~n",[])
	).
plot_percentage(_,_,_,_):- %garbage collector, never fails
	format("GC:Doesn't plot percentage.~n",[]).

%------------------------------------------------------------------------------
% Plot pair cost assertion and analysis as regions between ub and lb,
% When the functions is evaluable by gnuplot with interactive terminal, we send
% these function directly to gnuplot. Otherwise, we evaluate the functions and
% make a data series files. The data evaluation is represented using 2 files
% for spec and for cost
% plot_comparison(pred, $(1), exp(2,$(1)),  exp($(1),2), exp($(1),3),  [min_x(0),max_x(100),tmp_path(/tmp),out_file(ciaoplot.jpg)]).
% plot_comparison(pred, $(1), exp(2,$(1)),  exp($(1),2), exp($(1),3),  [interactive(yes), plot_type(comparison)]).
%------------------------------------------------------------------------------
plot_comparison(Pred, Slb, Sub, Clb, Cub, PlotConfig):- %direct to gnuplot
%	plot '+' u 1:($1**2):($1**3) w filledcu clo fs trans soli 0.3 lc rgb "green", x**2, x**3 notitle
	member(interactive(yes), PlotConfig),
	member(plot_type(comparison), PlotConfig),
	member(terminal(Terminal), PlotConfig),
	member(min_x(MinX), PlotConfig),
	atom_number(MinXAtom, MinX),
	(
	    member(max_x(MaxX), PlotConfig) ->
	    atom_number(MaxXAtom, MaxX)
	;
	    MaxXAtom = '*'
	),
	simplify_pred(Pred, PredAtom),
	ciao2gnuplot(Slb, GSlb, GLSlb, TSlb),
	ciao2gnuplot(Sub, GSub, GLSub, TSub),
	ciao2gnuplot(Clb, GClb, GLClb, TClb),
	ciao2gnuplot(Cub, GCub, GLCub, TCub),
	%construct gnuplot command
	merge_atoms([
		       'set terminal ', Terminal, ';',
		       'set yrange [0:*];',
		       'set xrange [', MinXAtom,':',MaxXAtom,'];',
		       'set title "', PredAtom, '"; set key left; ', %config
		       'plot ''+'' u 1:(', GSlb, '):(', GSub,
		       ') w filledcu clo fs trans soli 0.3 lc rgb "green" title "Spec: ',
		       TSlb, '; ', TSub, '",', %spec
		       ' ''+'' u 1:(', GClb, '):(', GCub,
		       ') w filledcu clo fs trans soli 0.3 lc rgb "red"   title "Cost: ',
		       TClb, '; ', TCub, '",', %cost
		       GLSlb, ' lc rgb "green" notitle,', GLSub, ' lc rgb "green" notitle,',
		       GLClb, ' lc rgb "red" notitle,', GLCub, ' lc rgb "red" notitle',%border of region
		       '; pause mouse close'
		   ], GnuplotCmd),
	format("Command: ~a~n", [GnuplotCmd]),
	process_call(path(gnuplot), ['-e', GnuplotCmd], [status(DrawStat)]),
	(
	    DrawStat = 0 ->
	    true
	;
	    format("Graph drawing failed.~n",[])
	).

plot_comparison(Pred, Slb, Sub, Clb, Cub, PlotConfig):- %using dataseries
	\+ member(interactive(yes), PlotConfig), %just to prevent choice points
	member(min_x(A), PlotConfig),
	member(max_x(B), PlotConfig),
	member(tmp_path(Tmp), PlotConfig),
	member(out_file(FileOut), PlotConfig),
	member(skip(Skip), PlotConfig),
	%additional config for title,legends,etc
	AddPlotConfig = [pred(Pred), slb(Slb), sub(Sub), clb(Clb), cub(Cub)],
	append(PlotConfig, AddPlotConfig, PlotConfig1), 	
	write_plot_template(PlotConfig1),
	path_concat(Tmp, 'spec.dat', SpecFile),
 	path_concat(Tmp, 'cost.dat', CostFile),
	open(SpecFile, 'write', SpecStream),
	open(CostFile, 'write', CostStream),
	format(SpecStream,"#~a ~a ~a~n", ['X', 'Slb', 'Sub']),
	format(CostStream,"#~a ~a ~a~n", ['X', 'Clb', 'Cub']),
	evaluate_pair_func(Slb, Sub, Clb, Cub, A, B, Skip, SpecStream, CostStream),
	close(SpecStream),
	close(CostStream),	
	process_call(path(gnuplot), ['ciaoplot.dem'], [cwd(Tmp), status(DrawStat)]),
	(
	    DrawStat = 0 ->
	    plot_id(PlID),
	    member(terminal(Terminal), PlotConfig),
	    (
		get_out_ext(Terminal, ExtF)->
		format("Graph is drawn in ~a/~a.~d.~a~n",[Tmp,FileOut,PlID, ExtF])
	    ;
		true
	    )
	;
	    format("Graph drawing failed.~n",[])
	).
plot_comparison(_,_,_,_,_,_):- %garbage collector, never fails
	format("GC: Doesn't plot comparison.~n",[]).

evaluate_pair_func(Slb, Sub, Clb, Cub, A, B, Skip, SpecS, CostS):-
	A < B,
	eval_arith(Slb, A, RSlb),
	eval_arith(Sub, A, RSub),
	eval_arith(Clb, A, RClb),
	eval_arith(Cub, A, RCub),
	format(SpecS, "~d ~g ~g~n", [A, RSlb, RSub]),
	format(CostS, "~d ~g ~g~n", [A, RClb, RCub]),
	A1 is A + Skip,
	evaluate_pair_func(Slb, Sub, Clb, Cub, A1, B, Skip, SpecS, CostS).

evaluate_pair_func(Slb, Sub, Clb, Cub, A, B, _Skip, SpecS, CostS):-
	A >= B,
	eval_arith(Slb, A, RSlb),
	eval_arith(Sub, A, RSub),
	eval_arith(Clb, A, RClb),
	eval_arith(Cub, A, RCub),
	format(SpecS, "~d ~g ~g~n", [A, RSlb, RSub]),
	format(CostS, "~d ~g ~g~n", [A, RClb, RCub]).

evaluate_percentage(S, C, A, B, Skip, OutS):-
	A < B,
	eval_arith(S, A, RS),
	eval_arith(C, A, RC),
	Res = (RS - RC) / RS * 100,  % 100 percent
	format(OutS, "~d 0 ~g~n", [A, Res]),
	A1 is A + Skip,
	evaluate_percentage(S, C, A1, B, Skip, OutS).

evaluate_percentage(S, C, A, B, _Skip, OutS):-
	A >= B,
	eval_arith(S, A, RS),
	eval_arith(C, A, RC),
	Res = (RS - RC) / RS * 100, % 100 percent
	format(OutS, "~d 0 ~g~n", [A, Res]).

evaluate_true(L, U, A, B, Skip, OutS):-
	A < B,
	eval_arith(L, A, RL),
	eval_arith(U, A, RU),
	format(OutS, "~d ~g ~g~n", [A, RL, RU]),
	A1 is A + 1,
	evaluate_true(L, U, A1, B, Skip, OutS).

evaluate_true(L, U, A, B, _Skip, OutS):-
	A >= B,
	eval_arith(L, A, RL),
	eval_arith(U, A, RU),
	format(OutS, "~d ~g ~g~n", [A, RL, RU]).

%------------------------------------------------------------------------------
% taken from ciaopp/infercost/algebraic/math_polynom.pl
% arithmetic function evaluation in prolog
%------------------------------------------------------------------------------
:- doc(eval_arith/3, "Evaluate arithmetic expression on certain x value").
:- pred eval_arith(Expr, X, Res)#"value of @var{Res} might be @em{NaN} when 
the @var{Exp} is undefined on @var{X}".
%base
eval_arith(Expr, Subx, Subx):-
	functor(Expr, $, _). %variable x

eval_arith(Expr, _Sub, Expr):-
	number(Expr).

%recc	
eval_arith(Expr, Subx, Res):-
	functor(Expr, F, 1), %unary func
	arg(1, Expr, Arg1),
	eval_arith(Arg1, Subx, Res1),
	(
	    F == '-' ->
	    Res is - Res1
	% ;
	%     F == 'fact' ->
	%     fac(Res1, Res)
	).

eval_arith(Expr, Subx, Res):-
	functor(Expr, F, 2), %binary func
	arg(1, Expr, Arg1),
	arg(2, Expr, Arg2),
	eval_arith(Arg1, Subx, Res1),
	eval_arith(Arg2, Subx, Res2),
	(
	    F == '+' ->
	    Res is Res1 + Res2
	;
	    F == '-' ->
	    Res is Res1 - Res2
	;
	    F == '*' ->
	    Res is Res1 * Res2
	;
	    F == '/' ->
	    Res is Res1 / Res2
	;
	    F == '**' ->
	    Res is Res1 ** Res2
	;
	    F == 'exp' ->
	    Res is Res1 ** Res2
	;
	    F == 'log' ->
	    %log(Res1,Res2) means 'Res1' is the base
	    Res is log(Res2) / log(Res1)
	% ;
	%     F == 'comb' ->
	%     combination(Res1, Res2,  Res)
	).

%
sanitize_config(Config, NewConfig):-
	append(Config, _, NConfig),!,
	(
	    member(min_x(A), Config) ->
	    true
	;
	    plot_config(min_x, A),
	    member(min_x(A), NConfig)
	),!,
	(
	    member(max_x(B), Config) ->
	    A < B
	;
	    (
		member(interactive(yes), Config) ->
		true
	    ;
		plot_config(max_x, B),
		member(max_x(B), NConfig),
		A < B
	    )
	),!,

	(
	    member(tmp_path(P), Config) ->
	    true
	;
	    plot_config(tmp_path, P),
	    member(tmp_path(P), NConfig)
	),!,
	(
	    member(out_file(F), Config) ->
	    true
	;
	    plot_config(out_file, F),
	    member(out_file(F), NConfig)
	),!,
	(
	    member(terminal(T), Config) ->
	    true
	;
	    plot_config(terminal, T),
	    member(terminal(T), NConfig)
	),!,
	(
	    member(skip(Skip), Config) ->
	    true
	;
	    plot_config(skip, Skip),
	    member(skip(Skip), NConfig)
	),!,
	(
	    member(plot_type(PType), Config) ->
	    member(plot_type(PType), NConfig)
	;
	    member(plot_type(comparison), NConfig)
	),!,
	(
	    member(plot_all(yes), Config) ->
	    true
	;
	    true
	),!,
	(
	    member(interactive(yes), Config) ->
	    \+ get_out_ext(T,_),
	    member(interactive(yes), NConfig)
	;
	    true
	),!,
	append(NConfig, [], NewConfig).

cleanup_tmp_dir(Config):-
	member(terminal(Terminal),Config),
	(
	    get_out_ext(Terminal, FileExt) ->
	    member(tmp_path(Tmp), Config),
	    member(out_file(OutFile), Config),
	    atom_concat(OutFile, '.*.', Ofb),
	    atom_concat(Ofb, FileExt, FileBlob),
	    process_call(path(rm), [FileBlob], [cwd(Tmp), status(_RmStat)])
	;
	    true
	).

get_orig_pred(FullPred, Pred):-
	%from mod:pred -> pred, 58 -> :	
	FullPred =.. [MP_functor|Args],
	atom_codes(MP_functor,AtmC),append(_,[58|Res],AtmC),atom_codes(P_functor,Res),
	Pred =.. [P_functor|Args].
get_orig_pred(Pred, Pred). %give the original, either w/o module or fullpred

init_plot_id:-
	retractall_fact(plot_id(_)),
	asserta_fact(plot_id(-1)).	
incr_plot_id:-
	plot_id(N),
	retractall_fact(plot_id(_)),
	N1 is N + 1,
	asserta_fact(plot_id(N1)).
initial_plot_id:-
	plot_id(-1).
%------------------------------------------------------------------------------
% get file extension when the terminal is a file output
%------------------------------------------------------------------------------
get_out_ext(Terminal, Ext):-
	(
	    Terminal = png;
	    Terminal = jpg
	),
	Ext = Terminal.

%------------------------------------------------------------------------------
% provide predicate representation for plot title
%------------------------------------------------------------------------------
simplify_pred_prolog(Pred, SPred):- %for prolog style title
	get_orig_pred(Pred, OPred),
	functor(OPred,F,N),
	atom_concat(F,'/', Ps),
	atom_number(AN,N),
	atom_concat(Ps, AN, PredArity),
	atom_codes(PredArity, PACode),
	treat_underscore(PACode,SPredCode),
	atom_codes(SPred, SPredCode).

 simplify_pred(Pred, SPred):- %for common style title
	get_orig_pred(Pred, OPred),
	functor(OPred,F,N),
	construct_arguments(0, N, Args),
	atom_concat(F, Args, PredArity),
	atom_codes(PredArity, PACode),
	treat_underscore(PACode,SPredCode),
	atom_codes(SPred, SPredCode).

%escape underscores with \
treat_underscore([],[]).
treat_underscore([A|P],PClean):-
	(
	    A = 95 ->
	    PClean = [92,95|Pc]
	;
	    PClean = [A|Pc]
	),
	treat_underscore(P, Pc).

construct_arguments(0,0,'()'):-!.
construct_arguments(0,N,Args):-!,
	construct_arguments(1, N, ArgR),
	atom_concat('(', ArgR, Args).
construct_arguments(N,N,Args):-!,
	num_to_char(N, ArgID),
	atom_concat(ArgID, ')',Args).
construct_arguments(I,N, Args):-
	I < N,
	num_to_char(I, ArgID),
	atom_concat(ArgID, ',', ArgI),
	Ii is I + 1,
	construct_arguments(Ii, N, ArgR),
	atom_concat(ArgI, ArgR, Args).
%------------------------------------------------------------------------------
% translating cost function into gnuplot format
%------------------------------------------------------------------------------
ciao2gnuplot(none, none, none, none):-!.
ciao2gnuplot(A, Func, FuncLine, Title):-
	functor(A, F, N),
	(
	    N == 2 ->
	    arg(1, A, Arg1),
	    arg(2, A, Arg2),
	    ciao2gnuplot(Arg1, FArg1, FLArg1, TArg1),
	    ciao2gnuplot(Arg2, FArg2, FLArg2, TArg2),
	    (
		F == exp,
		atom_concat(FArg1, **, FPart1),
		atom_concat(FLArg1, **, FLPart1),
		atom_concat(TArg1, '^', TPart1)
	    ;
		atom_concat(FArg1, F, FPart1),
		atom_concat(FLArg1, F, FLPart1),
		atom_concat(TArg1, F, TPart1)
	    ),
	    atom_concat(FPart1, FArg2, Func),
	    atom_concat(FLPart1, FLArg2, FuncLine),
	    atom_concat(TPart1, TArg2, Title)
	;
	    N == 1 ->
	    F == '$',
	    arg(1, A, Arg1),
%	    atom_number(ArgAtom, Arg1),
	    Func = '$1',
	    FuncLine ='x', %the only different with Func
%	    atom_concat('$(', ArgAtom, Title0),
%	    atom_concat(Title0, ')', Title)
	    num_to_char(Arg1, Title)
	;
	    N == 0 -> %it must be number
	    atom_number(Func, A),
	    FuncLine = Func,
	    format_to_string("~2g",A, SA),
	    atom_codes(Title, SA)
	;
	    format(user, "~n function ~p is unspecified~n",[A]),
	    fail
	).

merge_atoms([],'').
merge_atoms([A|LA], Atom):-
	merge_atoms(LA, Atom1),
	atom_concat(A, Atom1, Atom).

% provides nicer output for predicate arguments in plot title and legend
num_to_char(Num, Atom):-
	Num > 0,
	Num < 27,
	%Code is Num + 96, %lower case for common style
	Code is Num + 64, %upper case to distinguish w/ small e of expontiation
	char_code(Atom, Code).
